/****************************************************************************
 *
 *  Rohde & Schwarz ZVL Network Analyzer instrument driver include file
 *
 ****************************************************************************/

#ifndef __RSZVL_HEADER
#define __RSZVL_HEADER

#include "rsidr_core.h"
#include "rszvl_attributes.h"
#include "rszvl_utility.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#pragma warning (disable:4152) // warning C4152: nonstandard extension, function/data pointer conversion in expression
#pragma warning (disable:4100) // warning C4100: 'xxx' : unreferenced formal parameter

#if defined(__cplusplus) || defined(__cplusplus__)
extern "C" {
#endif

/*****************************************************************************
 *- Useful Macros & Constants -----------------------------------------------*
 *****************************************************************************/

#define RSZVL_SIMULATION_ID_QUERY      "Rohde&Schwarz,ZVL,123456/789,1.23"  /* Default definition of ID Query Response for simulation */
#define RSZVL_VALID_ID_RESPONSE_STRING "Rohde&Schwarz,ZVL"      /* Valid response for identification query */
#define RSZVL_SIMULATION_OPT_QUERY     ""      /* Simulated response for *OPT? command */

#define RSZVL_IO_BUFFER_SIZE           1024L   /* I/O buffer size */
#define RSZVL_OPC_TIMEOUT              5000L   /* Maximum time to wait for OPC in milliseconds */
//#define IEEE_488_2_ERROR_MASK             0x3C    /* 488.2 Event Status Register (ESR) Bits */
#define IEEE_488_2_ERROR_MASK             0x4    /* 488.2 STB Bits */   
#define IEEE_488_2_ESE_MASK               0x3D    /* 488.2 Event Status Register (ESE) Bits */
#define RSZVL_NAN                      9.91e+37

/****************************************************************************
 *----------------- Instrument Driver Revision Information -----------------*
 ****************************************************************************/

/* Class Driver Identification */

#define RSZVL_CLASS_DRIVER_DESCRIPTION           ""
#define RSZVL_CLASS_DRIVER_PREFIX                ""
#define RSZVL_CLASS_DRIVER_VENDOR                ""
#define RSZVL_CLASS_DRIVER_REVISION              ""
#define RSZVL_CLASS_SPEC_MAJOR_VERSION           1L    /* Class specification major version */
#define RSZVL_CLASS_SPEC_MINOR_VERSION           0L    /* Class specification minor version */

/* Driver Identification */

#define RSZVL_SPECIFIC_DRIVER_DESCRIPTION        ""
#define RSZVL_SPECIFIC_DRIVER_PREFIX             "RSZVL"
#define RSZVL_SPECIFIC_DRIVER_LOCATOR            ""
#define RSZVL_SPECIFIC_DRIVER_VENDOR             ""
#define RSZVL_MAJOR_VERSION                      3L    /* Instrument driver major version          */
#define RSZVL_MINOR_VERSION                      20L   /* Instrument driver minor version          */
#define RSZVL_MINOR_MINOR_VERSION                0L    /* Instrument driver minor minor version    */

/* Driver Capabilities */          

#define RSZVL_SUPPORTED_INSTRUMENT_MODELS        ""    /* Instrument driver supported model(s)     */
#define RSZVL_GROUP_CAPABILITIES                 ""
#define RSZVL_FUNCTION_CAPABILITIES              ""

/****************************************************************************
 *---------------------------- Attribute Defines ---------------------------*
 ****************************************************************************/

    /*- Inherent Instrument Specific Attributes ----------------------------*/

        /* User Options */

#define RSZVL_ATTR_RANGE_CHECK                                           RS_ATTR_RANGE_CHECK                                     /* ViBoolean */
#define RSZVL_ATTR_QUERY_INSTRUMENT_STATUS                               RS_ATTR_QUERY_INSTRUMENT_STATUS                         /* ViBoolean */
#define RSZVL_ATTR_CACHE                                                 RS_ATTR_CACHE                                           /* ViBoolean */
#define RSZVL_ATTR_SIMULATE                                              RS_ATTR_SIMULATE                                        /* ViBoolean */
#define RSZVL_ATTR_RECORD_COERCIONS                                      RS_ATTR_RECORD_COERCIONS                                /* ViBoolean */
#define RSZVL_ATTR_INTERCHANGE_CHECK                                     RS_ATTR_INTERCHANGE_CHECK                               /* ViBoolean */
#define RSZVL_ATTR_SPY                                                   RS_ATTR_SPY                                             /* ViBoolean */
#define RSZVL_ATTR_USE_SPECIFIC_SIMULATION                               RS_ATTR_USE_SPECIFIC_SIMULATION                         /* ViBoolean */

        /* Class Driver Identification */

#define RSZVL_ATTR_CLASS_DRIVER_DESCRIPTION                              RS_ATTR_CLASS_DRIVER_DESCRIPTION                        /* ViString, read-only */
#define RSZVL_ATTR_CLASS_DRIVER_PREFIX                                   RS_ATTR_CLASS_DRIVER_PREFIX                             /* ViString, read-only */
#define RSZVL_ATTR_CLASS_DRIVER_VENDOR                                   RS_ATTR_CLASS_DRIVER_VENDOR                             /* ViString, read-only */
#define RSZVL_ATTR_CLASS_DRIVER_REVISION                                 RS_ATTR_CLASS_DRIVER_REVISION                           /* ViString, read-only */
#define RSZVL_ATTR_CLASS_DRIVER_CLASS_SPEC_MAJOR_VERSION                 RS_ATTR_CLASS_DRIVER_CLASS_SPEC_MAJOR_VERSION           /* ViInt32,  read-only */
#define RSZVL_ATTR_CLASS_DRIVER_CLASS_SPEC_MINOR_VERSION                 RS_ATTR_CLASS_DRIVER_CLASS_SPEC_MINOR_VERSION           /* ViInt32,  read-only */

        /* Driver Identification */

#define RSZVL_ATTR_SPECIFIC_DRIVER_DESCRIPTION                           RS_ATTR_SPECIFIC_DRIVER_DESCRIPTION                     /* ViString, read-only */
#define RSZVL_ATTR_SPECIFIC_DRIVER_PREFIX                                RS_ATTR_SPECIFIC_DRIVER_PREFIX                          /* ViString, read-only */
#define RSZVL_ATTR_SPECIFIC_DRIVER_LOCATOR                               RS_ATTR_SPECIFIC_DRIVER_LOCATOR                         /* ViString, read-only */
#define RSZVL_ATTR_SPECIFIC_DRIVER_VENDOR                                RS_ATTR_SPECIFIC_DRIVER_VENDOR                          /* ViString, read-only */
#define RSZVL_ATTR_SPECIFIC_DRIVER_REVISION                              RS_ATTR_SPECIFIC_DRIVER_REVISION                        /* ViString, read-only */
#define RSZVL_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MAJOR_VERSION              RS_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MAJOR_VERSION        /* ViInt32,  read-only */
#define RSZVL_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MINOR_VERSION              RS_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MINOR_VERSION        /* ViInt32,  read-only */

        /* Driver Capabilities */

#define RSZVL_ATTR_SUPPORTED_INSTRUMENT_MODELS                           RS_ATTR_SUPPORTED_INSTRUMENT_MODELS                     /* ViString, read-only */
#define RSZVL_ATTR_GROUP_CAPABILITIES                                    RS_ATTR_GROUP_CAPABILITIES                              /* ViString, read-only */
#define RSZVL_ATTR_FUNCTION_CAPABILITIES                                 RS_ATTR_FUNCTION_CAPABILITIES                           /* ViString, read-only */
#define RSZVL_ATTR_CHANNEL_COUNT                                         RS_ATTR_CHANNEL_COUNT                                   /* ViInt32,  read-only */

        /* Driver Setup */

#define RSZVL_ATTR_DRIVER_SETUP                                          RS_ATTR_DRIVER_SETUP                                    /* ViString            */

        /* Instrument Identification */

#define RSZVL_ATTR_INSTRUMENT_MANUFACTURER                               RS_ATTR_INSTRUMENT_MANUFACTURER                         /* ViString, read-only */
#define RSZVL_ATTR_INSTRUMENT_MODEL                                      RS_ATTR_INSTRUMENT_MODEL                                /* ViString, read-only */
#define RSZVL_ATTR_INSTRUMENT_FIRMWARE_REVISION                          RS_ATTR_INSTRUMENT_FIRMWARE_REVISION                    /* ViString, read-only */

        /* Advanced Session Information */

#define RSZVL_ATTR_IO_RESOURCE_DESCRIPTOR                                RS_ATTR_IO_RESOURCE_DESCRIPTOR                          /* ViString, read-only */
#define RSZVL_ATTR_LOGICAL_NAME                                          RS_ATTR_LOGICAL_NAME                                    /* ViString, read-only */

    /*- Instrument-Specific Attributes -------------------------------------*/

//#define RSZVL_ATTR_ID_QUERY_RESPONSE                                     (RS_SPECIFIC_PUBLIC_ATTR_BASE + 1L)                     /* ViString (Read Only)              */

/* BEGIN GENERATE */
#define RSZVL_ATTR_AMPLITUDE_UNITS                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4096
#define RSZVL_ATTR_INSTR_MODE                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4097
#define RSZVL_ATTR_ATTENUATION                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4098
#define RSZVL_ATTR_ATTENUATION_AUTO                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4099
#define RSZVL_ATTR_FREQUENCY_START                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4100
#define RSZVL_ATTR_FREQUENCY_STOP                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4101
#define RSZVL_ATTR_FREQUENCY_OFFSET                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4102
#define RSZVL_ATTR_FREQUENCY_CENTER                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4103
#define RSZVL_ATTR_FREQUENCY_CENTER_STEP                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4104
#define RSZVL_ATTR_FREQUENCY_STEP_AUTO                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4105
#define RSZVL_ATTR_FREQUENCY_CENTER_LINK                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4106
#define RSZVL_ATTR_FREQUENCY_CENTER_LINK_FACTOR                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4107
#define RSZVL_ATTR_FREQUENCY_SPAN                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4108
#define RSZVL_ATTR_FREQUENCY_SPAN_FULL                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4109
#define RSZVL_ATTR_FREQUENCY_MODE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4110
#define RSZVL_ATTR_INPUT_IMPEDANCE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4111
#define RSZVL_ATTR_NUMBER_OF_SWEEPS                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4112
#define RSZVL_ATTR_REFERENCE_LEVEL                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4113
#define RSZVL_ATTR_REFERENCE_LEVEL_OFFSET                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4114
#define RSZVL_ATTR_RESOLUTION_BANDWIDTH                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4115
#define RSZVL_ATTR_RESOLUTION_BANDWIDTH_AUTO                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4116
#define RSZVL_ATTR_RESOLUTION_BANDWIDTH_RATIO                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4117
#define RSZVL_ATTR_SWEEP_MODE_CONTINUOUS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4118
#define RSZVL_ATTR_SWEEP_TIME                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4119
#define RSZVL_ATTR_SWEEP_TIME_AUTO                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4120
#define RSZVL_ATTR_SWEEP_POINTS                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4121
#define RSZVL_ATTR_SWEEP_COUNT_CURRENT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4122
#define RSZVL_ATTR_VERTICAL_SCALE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4123
#define RSZVL_ATTR_VIDEO_BANDWIDTH                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4124
#define RSZVL_ATTR_VIDEO_BANDWIDTH_AUTO                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4125
#define RSZVL_ATTR_VIDEO_BANDWIDTH_RATIO                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4126
#define RSZVL_ATTR_MEAS_ABORT                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4127
#define RSZVL_ATTR_INIT                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4128
#define RSZVL_ATTR_INIT_CONT                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4129
#define RSZVL_ATTR_INIT_SW_TRIGGER                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4130
#define RSZVL_ATTR_INIT_CONMEAS                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4131
#define RSZVL_ATTR_AVG_STATE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4132
#define RSZVL_ATTR_AVG_TYPE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4133
#define RSZVL_ATTR_AVG_COUNT                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4134
#define RSZVL_ATTR_AVG_COUNT_AUTO                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4135
#define RSZVL_ATTR_RESOLUTION_BANDWIDTH_FILTER_TYPE                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4136
#define RSZVL_ATTR_VIDEO_BANDWIDTH_TYPE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4137
#define RSZVL_ATTR_TRACE_STATE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4138
#define RSZVL_ATTR_TRACE_RESET_BEHAVIOR                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4139
#define RSZVL_ATTR_TRACE_TYPE                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4140
#define RSZVL_ATTR_TRACE_MATH_STATE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4141
#define RSZVL_ATTR_TRACE_MATH_POSITION                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4142
#define RSZVL_ATTR_DETECTOR_TYPE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4143
#define RSZVL_ATTR_DETECTOR_TYPE_AUTO                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4144
#define RSZVL_ATTR_CALIBRATION_CHECK                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4145
#define RSZVL_ATTR_CALIBRATION                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4146
#define RSZVL_ATTR_CALIBRATION_ABORT                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4147
#define RSZVL_ATTR_CALIBRATION_RESULT_QUERY                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4148
#define RSZVL_ATTR_CALIBRATION_STATE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4149
#define RSZVL_ATTR_INP_UPORT_STATE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4150
#define RSZVL_ATTR_INP_UPORT_VALUE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4151
#define RSZVL_ATTR_OUT_UPORT_VALUE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4152
#define RSZVL_ATTR_DISP_REF_VALUE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4153
#define RSZVL_ATTR_DISP_REF_POSITION                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4154
#define RSZVL_ATTR_DISP_FREQ_STATE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4155
#define RSZVL_ATTR_DISP_TIME_STATE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4156
#define RSZVL_ATTR_DISP_PWR_SAVE_STATE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4157
#define RSZVL_ATTR_DISP_PWR_SAVE_HOLDOFF                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4158
#define RSZVL_ATTR_DISP_WINDOW_SIZE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4159
#define RSZVL_ATTR_DISP_COL_PRESET                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4160
#define RSZVL_ATTR_DISP_COL_PREDEFINED                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4161
#define RSZVL_ATTR_DISP_LOG_RANGE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4162
#define RSZVL_ATTR_DISP_AMPLITUDE_GRID_MODE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4163
#define RSZVL_ATTR_DISP_SINGLE_SWEEP                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4164
#define RSZVL_ATTR_DISP_UPDATE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4165
#define RSZVL_ATTR_UNIT_POWER                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4166
#define RSZVL_ATTR_MARKER_AMPLITUDE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4167
#define RSZVL_ATTR_MARKER_ENABLED                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4168
#define RSZVL_ATTR_MARKER_POSITION                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4169
#define RSZVL_ATTR_MARKER_TRACE                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4170
#define RSZVL_ATTR_MARKER_AOFF                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4171
#define RSZVL_ATTR_MARKER_LOEX                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4172
#define RSZVL_ATTR_SIGNAL_TRACK_ENABLED                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4173
#define RSZVL_ATTR_SIGNAL_TRACK_BWID                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4174
#define RSZVL_ATTR_SIGNAL_TRACK_THRESHOLD                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4175
#define RSZVL_ATTR_SIGNAL_TRACK_TRACE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4176
#define RSZVL_ATTR_MARKER_STEP_SIZE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4177
#define RSZVL_ATTR_MARKER_PROBABILITY                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4178
#define RSZVL_ATTR_MARKER_SEARCH_MIN_RIGHT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4179
#define RSZVL_ATTR_MARKER_SEARCH_MIN_LEFT                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4180
#define RSZVL_ATTR_MARKER_SEARCH_PEAK                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4181
#define RSZVL_ATTR_MARKER_SEARCH_PEAK_NEXT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4182
#define RSZVL_ATTR_MARKER_SEARCH_PEAK_RIGHT                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4183
#define RSZVL_ATTR_MARKER_SEARCH_PEAK_LEFT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4184
#define RSZVL_ATTR_MARKER_SEARCH_MIN                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4185
#define RSZVL_ATTR_MARKER_SEARCH_MIN_NEXT                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4186
#define RSZVL_ATTR_MARKER_SEARCH_PEAK_AUTO                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4187
#define RSZVL_ATTR_MARKER_SEARCH_MIN_AUTO                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4188
#define RSZVL_ATTR_MARKER_THRESHOLD_STATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4189
#define RSZVL_ATTR_MARKER_THRESHOLD                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4190
#define RSZVL_ATTR_PEAK_EXCURSION                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4191
#define RSZVL_ATTR_MARKER_SEARCH_LIMITS_STATE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4192
#define RSZVL_ATTR_MARKER_SEARCH_LIMITS_LEFT                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4193
#define RSZVL_ATTR_MARKER_SEARCH_LIMITS_RIGHT                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4194
#define RSZVL_ATTR_MARKER_PEAK_LIST_COUNT                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4195
#define RSZVL_ATTR_MARKER_PEAK_LIST_FOUND                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4196
#define RSZVL_ATTR_MARKER_PEAK_LIST_SORT                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4197
#define RSZVL_ATTR_MARKER_ZOOM                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4198
#define RSZVL_ATTR_MARKER_DEMOD_STATE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4199
#define RSZVL_ATTR_MARKER_DEMOD_TYPE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4200
#define RSZVL_ATTR_MARKER_DEMOD_HOLDOFF                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4201
#define RSZVL_ATTR_MARKER_DEMOD_CONT                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4202
#define RSZVL_ATTR_MARKER_TO_CENTER                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4203
#define RSZVL_ATTR_MARKER_TO_STEP                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4204
#define RSZVL_ATTR_MARKER_TO_REFERENCE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4205
#define RSZVL_ATTR_MARKER_REF_STATE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4206
#define RSZVL_ATTR_MARKER_REF_LEVEL                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4207
#define RSZVL_ATTR_MARKER_REF_LEVEL_OFFSET                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4208
#define RSZVL_ATTR_MARKER_REF_FREQ                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4209
#define RSZVL_ATTR_MARKER_REF_PEAK                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4210
#define RSZVL_ATTR_REFERENCE_MARKER_STATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4211
#define RSZVL_ATTR_REFERENCE_MARKER_MODE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4212
#define RSZVL_ATTR_REFERENCE_MARKER_POSITION                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4213
#define RSZVL_ATTR_REFERENCE_MARKER_REL_POSITION                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4214
#define RSZVL_ATTR_REFERENCE_MARKER_AMPLITUDE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4215
#define RSZVL_ATTR_REFERENCE_MARKER_TRACE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4216
#define RSZVL_ATTR_REFERENCE_MARKER_AOFF                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4217
#define RSZVL_ATTR_REFERENCE_MARKER_PEAK                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4218
#define RSZVL_ATTR_REFERENCE_MARKER_PEAK_NEXT                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4219
#define RSZVL_ATTR_REFERENCE_MARKER_PEAK_NEXT_RIGHT                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4220
#define RSZVL_ATTR_REFERENCE_MARKER_PEAK_NEXT_LEFT                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4221
#define RSZVL_ATTR_REFERENCE_MARKER_MIN                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4222
#define RSZVL_ATTR_REFERENCE_MARKER_MIN_NEXT                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4223
#define RSZVL_ATTR_REFERENCE_MARKER_MIN_NEXT_RIGHT                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4224
#define RSZVL_ATTR_REFERENCE_MARKER_MIN_NEXT_LEFT                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4225
#define RSZVL_LIMIT_STATE                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4226
#define RSZVL_LIMIT_CLEAR                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4227
#define RSZVL_LIMIT_ASSIGN_TRACE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4228
#define RSZVL_LIMIT_UNITS                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4229
#define RSZVL_LIMIT_COMMENT                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4230
#define RSZVL_LIMIT_COPY                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4231
#define RSZVL_LIMIT_NAME                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4232
#define RSZVL_LIMIT_DELETE                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4233
#define RSZVL_LIMIT_CONTROL_DOMAIN                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4234
#define RSZVL_LIMIT_CONTROL_SPACING                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4235
#define RSZVL_LIMIT_CONTROL_SCALING                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4236
#define RSZVL_LIMIT_CONTROL_OFFSET                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4237
#define RSZVL_LIMIT_CONTROL_SHIFT                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4238
#define RSZVL_LIMIT_LOWER_STATE                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4239
#define RSZVL_LIMIT_LOWER_SCALING                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4240
#define RSZVL_LIMIT_LOWER_SPACING                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4241
#define RSZVL_LIMIT_LOWER_MARGIN                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4242
#define RSZVL_LIMIT_LOWER_THRESHOLD                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4243
#define RSZVL_LIMIT_LOWER_OFFSET                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4244
#define RSZVL_LIMIT_LOWER_SHIFT                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4245
#define RSZVL_LIMIT_UPPER_STATE                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4246
#define RSZVL_LIMIT_UPPER_SCALING                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4247
#define RSZVL_LIMIT_UPPER_SPACING                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4248
#define RSZVL_LIMIT_UPPER_MARGIN                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4249
#define RSZVL_LIMIT_UPPER_THRESHOLD                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4250
#define RSZVL_LIMIT_UPPER_OFFSET                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4251
#define RSZVL_LIMIT_UPPER_SHIFT                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4252
#define RSZVL_ATTR_THRLINE_STATE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4253
#define RSZVL_ATTR_THRLINE_POSITION                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4254
#define RSZVL_ATTR_DLINE_STATE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4255
#define RSZVL_ATTR_DLINE_POSITION                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4256
#define RSZVL_ATTR_FLINE_STATE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4257
#define RSZVL_ATTR_FLINE_POSITION                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4258
#define RSZVL_ATTR_TLINE_STATE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4259
#define RSZVL_ATTR_TLINE_POSITION                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4260
#define RSZVL_ATTR_TRIGGER_SOURCE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4261
#define RSZVL_ATTR_TRIGGER_DELAY                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4262
#define RSZVL_ATTR_TRIGGER_SLOPE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4263
#define RSZVL_ATTR_TRIGGER_EVENT_START                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4264
#define RSZVL_ATTR_EXTERNAL_GATE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4265
#define RSZVL_ATTR_EXTERNAL_GATE_HOLD                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4266
#define RSZVL_ATTR_EXTERNAL_GATE_TRIGGER_TYPE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4267
#define RSZVL_ATTR_EXTERNAL_GATE_POLARITY                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4268
#define RSZVL_ATTR_EXTERNAL_GATE_LENGTH                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4269
#define RSZVL_ATTR_EXTERNAL_GATE_SIGNAL_SOURCE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4270
#define RSZVL_ATTR_VIDEO_TRIGGER_LEVEL                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4271
#define RSZVL_ATTR_TRIGGER_IFP_LEVEL                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4272
#define RSZVL_ATTR_TRIGGER_IFP_OFFSET                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4273
#define RSZVL_ATTR_TRIGGER_IFP_HYSTERESIS                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4274
#define RSZVL_ATTR_TRIGGER_TIME_INTERVAL                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4275
#define RSZVL_ATTR_AMPL_PREAMPLIFIER                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4276
#define RSZVL_LIMIT_CHECK_RESULT                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4277
#define RSZVL_ATTR_MARKER_NOISE_STATE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4278
#define RSZVL_ATTR_MARKER_NOISE_RESULT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4279
#define RSZVL_ATTR_REFERENCE_MARKER_PNO_STATE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4280
#define RSZVL_ATTR_REFERENCE_MARKER_PNO_RESULT                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4281
#define RSZVL_ATTR_MARKER_FREQUENCY_COUNTER_ENABLED                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4282
#define RSZVL_ATTR_MARKER_FREQUENCY_COUNTER_RESOLUTION                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4283
#define RSZVL_ATTR_MARKER_COUNT                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4284
#define RSZVL_ATTR_MARKER_NDB_STATE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4285
#define RSZVL_ATTR_MARKER_NDB_VAL                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4286
#define RSZVL_ATTR_MARKER_NDB_RESULT                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4287
#define RSZVL_ATTR_LIST_POW_STATE_OFF                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4288
#define RSZVL_ATTR_MEAS_TDOM_STATE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4289
#define RSZVL_ATTR_MEAS_TDOM_PEAK                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4290
#define RSZVL_ATTR_MEAS_TDOM_RMS                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4291
#define RSZVL_ATTR_MEAS_TDOM_MEAN                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4292
#define RSZVL_ATTR_MEAS_TDOM_SDEV                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4293
#define RSZVL_ATTR_MEAS_TDOM_AVG                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4294
#define RSZVL_ATTR_MEAS_TDOM_MAX_HOLD                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4295
#define RSZVL_ATTR_MEAS_TDOM_MODE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4296
#define RSZVL_ATTR_MEAS_TDOM_AOFF                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4297
#define RSZVL_ATTR_MEAS_TDOM_SET_REFERENCE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4298
#define RSZVL_ATTR_MEAS_TDOM_PEAK_RESULT                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4299
#define RSZVL_ATTR_MEAS_TDOM_RMS_RESULT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4300
#define RSZVL_ATTR_MEAS_TDOM_MEAN_RESULT                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4301
#define RSZVL_ATTR_MEAS_TDOM_SDEV_RESULT                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4302
#define RSZVL_ATTR_MEAS_TDOM_AVG_PEAK_RESULT                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4303
#define RSZVL_ATTR_MEAS_TDOM_AVG_RMS_RESULT                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4304
#define RSZVL_ATTR_MEAS_TDOM_AVG_MEAN_RESULT                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4305
#define RSZVL_ATTR_MEAS_TDOM_AVG_SDEV_RESULT                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4306
#define RSZVL_ATTR_MEAS_TDOM_PEAK_HOLD_RESULT                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4307
#define RSZVL_ATTR_MEAS_TDOM_PEAK_HOLD_RMS_RESULT                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4308
#define RSZVL_ATTR_MEAS_TDOM_PEAK_HOLD_MEAN_RESULT                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4309
#define RSZVL_ATTR_MEAS_TDOM_PEAK_HOLD_SDEV_RESULT                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4310
#define RSZVL_ATTR_MEAS_POW_SELECT                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4311
#define RSZVL_ATTR_MEAS_POW_OFF                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4312
#define RSZVL_ATTR_MEAS_POW_ADJ_NUM                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4313
#define RSZVL_ATTR_MEAS_POW_ADJ_MODE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4314
#define RSZVL_ATTR_MEAS_POW_MODE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4315
#define RSZVL_ATTR_MEAS_POW_RESULT_MODE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4316
#define RSZVL_ATTR_MEAS_POW_CHANNEL_SPACING                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4317
#define RSZVL_ATTR_MEAS_POW_ADJ_CHANNEL_SPACING                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4318
#define RSZVL_ATTR_MEAS_POW_ALT_CHANNEL_SPACING                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4319
#define RSZVL_ATTR_MEAS_POW_CHANNEL_BANDWIDTH                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4320
#define RSZVL_ATTR_MEAS_POW_ADJ_CHANNEL_BANDWIDTH                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4321
#define RSZVL_ATTR_MEAS_POW_ALT_CHANNEL_BANDWIDTH                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4322
#define RSZVL_ATTR_MEAS_POW_TRACE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4323
#define RSZVL_ATTR_MEAS_POW_STANDARD                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4324
#define RSZVL_ATTR_MEAS_POW_REF_VALUE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4325
#define RSZVL_ATTR_MEAS_POW_ADJ_PRESET                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4326
#define RSZVL_ATTR_MEAS_ACP_HSP                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4327
#define RSZVL_ATTR_MEAS_ACP_PRESET_REF_LEVEL                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4328
#define RSZVL_ATTR_MEAS_ACP_LIMIT_STATE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4329
#define RSZVL_ATTR_MEAS_ACP_LIMIT_REL_STATE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4330
#define RSZVL_ATTR_MEAS_ACP_LIMIT_ABS_STATE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4331
#define RSZVL_ATTR_MEAS_ACP_ALT_LIMIT_REL_STATE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4332
#define RSZVL_ATTR_MEAS_ACP_ALT_LIMIT_ABS_STATE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4333
#define RSZVL_ATTR_MEAS_POW_CARR_SIG_NUM                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4334
#define RSZVL_ATTR_MEAS_POW_REF_CHAN_SEL_AUTO                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4335
#define RSZVL_ATTR_MEAS_POW_ADJ_REF_TXCHANNEL                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4336
#define RSZVL_ATTR_MEAS_POW_BANDWIDTH                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4337
#define RSZVL_ATTR_MEAS_STAT_APD_STATE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4338
#define RSZVL_ATTR_MEAS_STAT_CCDF_STATE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4339
#define RSZVL_ATTR_MEAS_STAT_SAMPLES                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4340
#define RSZVL_ATTR_MEAS_STAT_X_REF                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4341
#define RSZVL_ATTR_MEAS_STAT_X_RANGE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4342
#define RSZVL_ATTR_MEAS_STAT_Y_MIN                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4343
#define RSZVL_ATTR_MEAS_STAT_Y_MAX                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4344
#define RSZVL_ATTR_MEAS_STAT_ADJ                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4345
#define RSZVL_ATTR_MEAS_STAT_PRESET                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4346
#define RSZVL_ATTR_STAT_RESULT                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4347
#define RSZVL_ATTR_MEAS_MDEPTH_STATE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4348
#define RSZVL_ATTR_MEAS_MDEPTH_RESULT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4349
#define RSZVL_ATTR_MEAS_MDEPTH_SEARCH                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4350
#define RSZVL_ATTR_MEAS_TOI_STATE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4351
#define RSZVL_ATTR_MEAS_TOI_RESULT                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4352
#define RSZVL_ATTR_MEAS_TOI_SEARCH                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4353
#define RSZVL_ATTR_MEAS_HDIST_STATE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4354
#define RSZVL_ATTR_MEAS_HDIST_NOOFHARM                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4355
#define RSZVL_ATTR_MEAS_HDIST_RBWAUTO                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4356
#define RSZVL_ATTR_MEAS_HDIST_PRESET                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4357
#define RSZVL_ATTR_MPOW_MIN                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4358
#define RSZVL_ATTR_MPOW_FTYPE                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4359
#define RSZVL_ATTR_IQ_SAMPLE_RATE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4360
#define RSZVL_ATTR_IQ_DATA_STATE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4361
#define RSZVL_ATTR_IQ_DATA_AVER_STATE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4362
#define RSZVL_ATTR_IQ_DATA_AVER_COUNT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4363
#define RSZVL_ATTR_ID_QUERY_RESPONSE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4364
#define RSZVL_ATTR_SERVICE_INPUT_SOURCE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4365
#define RSZVL_ATTR_SERVICE_INPUT_COMB_FREQUENCY                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4366
#define RSZVL_ATTR_SERVICE_NOISE_SOURCE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4367
#define RSZVL_ATTR_SERVICE_STEST_RESULTS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4368
#define RSZVL_ATTR_SERVICE_STEST                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4369
#define RSZVL_ATTR_FILE_RECALL                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4370
#define RSZVL_ATTR_FILE_STARTUP_RECALL                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4371
#define RSZVL_ATTR_FILE_SAVE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4372
#define RSZVL_ATTR_FILE_DATA_CLEAR                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4373
#define RSZVL_ATTR_FILE_DATA_CLEAR_ALL                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4374
#define RSZVL_ATTR_FILE_DEC_SEPARATOR                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4375
#define RSZVL_ATTR_FILE_MANAGER_EDIT_PATH                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4376
#define RSZVL_ATTR_FILE_MANAGER_DELETE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4377
#define RSZVL_ATTR_FILE_MANAGER_MAKE_DIR                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4378
#define RSZVL_ATTR_FILE_MANAGER_EDIT_PATH_DEVICE                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4379
#define RSZVL_ATTR_FILE_MANAGER_DELETE_DIR                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4380
#define RSZVL_ATTR_FILE_ITEMS_SAVE_RECAL_SEL_SOURCE_CAL_DATA                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4381
#define RSZVL_ATTR_FILE_ITEMS_SAVE_RECAL_SEL_TRAN                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4382
#define RSZVL_ATTR_FILE_ITEMS_SAVE_RECAL_SEL_HWSETTINGS                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4383
#define RSZVL_ATTR_FILE_ITEMS_SAVE_RECAL_SEL_TRAC                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4384
#define RSZVL_ATTR_FILE_ITEMS_SAVE_RECAL_SEL_LINE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4385
#define RSZVL_ATTR_FILE_ITEMS_SAVE_RECAL_ALL                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4386
#define RSZVL_ATTR_FILE_ITEMS_SAVE_RECAL_NONE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4387
#define RSZVL_ATTR_FILE_ITEMS_SAVE_RECAL_DEFAULT                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4388
#define RSZVL_ATTR_HCOPY_ABORT                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4389
#define RSZVL_ATTR_HCOPY_COLOR_DEF                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4390
#define RSZVL_ATTR_HCOPY_COLOR                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4391
#define RSZVL_ATTR_HCOPY_PRINT                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4392
#define RSZVL_ATTR_HCOPY_PRINT_NEXT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4393
#define RSZVL_ATTR_HCOPY_DEVICE_DESTINATION                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4394
#define RSZVL_ATTR_HCOPY_COLOR_PREDEFINED                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4395
#define RSZVL_ATTR_HCOPY_DEVICE_LANG_OUT_FORM                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4396
#define RSZVL_ATTR_HCOPY_PRINT_SCREEN                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4397
#define RSZVL_ATTR_HCOPY_COMM_SCR                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4398
#define RSZVL_ATTR_HCOPY_PRINT_TRACE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4399
#define RSZVL_ATTR_HCOPY_DEVICE_ORIENTATION                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4400
#define RSZVL_ATTR_HCOPY_FILE_NAME                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4401
#define RSZVL_ATTR_HCOPY_PRINTER                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4402
#define RSZVL_ATTR_HCOPY_PRINTER_NEXT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4403
#define RSZVL_ATTR_HCOPY_PRINTER_FIRST                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4404
#define RSZVL_ATTR_ROSC_SOURCE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4405
#define RSZVL_ATTR_SYST_ERR                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4406
#define RSZVL_ATTR_SYST_ERR_LIST                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4407
#define RSZVL_ATTR_SYST_ERR_CLEAR                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4408
#define RSZVL_ATTR_SYST_PRESET                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4409
#define RSZVL_ATTR_SYST_VERSION                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4410
#define RSZVL_ATTR_SYST_SPEAKER                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4411
#define RSZVL_ATTR_NWA_NUMBER_OF_SWEEPS                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4412
#define RSZVL_ATTR_NWA_AVG_STATE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4413
#define RSZVL_ATTR_NWA_AVG_COUNT                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4414
#define RSZVL_ATTR_NWA_AVERAGE_CLEAR                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4415
#define RSZVL_ATTR_NWA_RESOLUTION_BANDWIDTH                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4416
#define RSZVL_ATTR_NWA_SWEEP_POINTS                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4417
#define RSZVL_ATTR_NWA_CHANNEL_STATE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4418
#define RSZVL_ATTR_NWA_CHANNEL_NAME                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4419
#define RSZVL_ATTR_NWA_ASSIGN_TRACE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4420
#define RSZVL_ATTR_NWA_SOURCE_PORT                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4421
#define RSZVL_ATTR_NWA_RF_OUTPUT                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4422
#define RSZVL_ATTR_NWA_SWEEP_TYPE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4423
#define RSZVL_ATTR_NWA_SWEEP_MEAS_DELAY                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4424
#define RSZVL_ATTR_NWA_SWEEP_STEP                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4425
#define RSZVL_ATTR_NWA_INSERT_NEW_SEGMENT                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4426
#define RSZVL_ATTR_NWA_SEGMENT_SELECT_SWEEP_TIME                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4427
#define RSZVL_ATTR_NWA_SEGMENT_STATE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4428
#define RSZVL_ATTR_NWA_SEGMENT_FREQ_START                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4429
#define RSZVL_ATTR_NWA_SEGMENT_FREQ_STOP                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4430
#define RSZVL_ATTR_NWA_SEGMENT_POWER_CONTROL                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4431
#define RSZVL_ATTR_NWA_SEGMENT_ATTEN                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4432
#define RSZVL_ATTR_NWA_SEGMENT_BAND_CONTROL                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4433
#define RSZVL_ATTR_NWA_SEGMENT_SELECT_CONTROL                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4434
#define RSZVL_ATTR_NWA_SEGMENT_TIME_CONTROL                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4435
#define RSZVL_ATTR_NWA_SEGMENT_MEAS_DELAY_CONTROL                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4436
#define RSZVL_ATTR_NWA_SEGMENT_RES_BAND                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4437
#define RSZVL_ATTR_NWA_SEGMENT_SELECTIVITY                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4438
#define RSZVL_ATTR_NWA_SEGMENT_SWEEP_TIME                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4439
#define RSZVL_ATTR_NWA_SEGMENT_MEAS_DELAY                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4440
#define RSZVL_ATTR_NWA_SEGMENT_SWEEP_POINTS                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4441
#define RSZVL_ATTR_NWA_SEGMENT_POWER                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4442
#define RSZVL_ATTR_NWA_SEGMENT_GET_COUNT                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4443
#define RSZVL_ATTR_NWA_SEGMENT_DELETE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4444
#define RSZVL_ATTR_NWA_SEGMENT_DELETE_ALL                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4445
#define RSZVL_ATTR_NWA_SEGMENT_GET_FREQ_CENTER                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4446
#define RSZVL_ATTR_NWA_SEGMENT_GET_FREQ_SPAN                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4447
#define RSZVL_ATTR_NWA_STORE_SEGMENT                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4448
#define RSZVL_ATTR_NWA_LOAD_SEGMENT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4449
#define RSZVL_ATTR_NWA_CORRECTION_STATE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4450
#define RSZVL_ATTR_NWA_CORRECTION_DELETE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4451
#define RSZVL_ATTR_NWA_CORRECTION_STORE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4452
#define RSZVL_ATTR_NWA_CORRECTION_LOAD                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4453
#define RSZVL_ATTR_NWA_CORRECTION_RESOLVE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4454
#define RSZVL_ATTR_NWA_CORR_RSAV                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4455
#define RSZVL_ATTR_NWA_CORR_RSAV_DEFAULT                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4456
#define RSZVL_ATTR_NWA_CORR_EXP_DATA_TOUCH_FILE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4457
#define RSZVL_ATTR_NWA_CORR_CONN_TYPE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4458
#define RSZVL_ATTR_NWA_CORR_CONN_PORTS                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4459
#define RSZVL_ATTR_NWA_CORR_DELETE_USER_CONN                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4460
#define RSZVL_ATTR_NWA_CORR_GET_CAL_STATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4461
#define RSZVL_ATTR_NWA_CORR_GET_CAL_DATA_PARAMS                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4462
#define RSZVL_ATTR_NWA_CORR_GET_CAL_DATE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4463
#define RSZVL_ATTR_NWA_CORR_FACT_CAL_STATE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4464
#define RSZVL_ATTR_NWA_GENERATE_CALIBRATION_KIT_FILE                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4465
#define RSZVL_ATTR_NWA_ACTIVE_CALIBRATION_UNIT                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4466
#define RSZVL_ATTR_NWA_GET_ALL_CALIBRATION_UNITS                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4467
#define RSZVL_ATTR_NWA_EXPORT_CALIBRATION_DATA_TO_TOUCHSTONE_FORMAT         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4468
#define RSZVL_ATTR_NWA_CORR_SAVE_CAL_DATA                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4469
#define RSZVL_ATTR_NWA_CORR_SAVE_DEFAULT_DATA                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4470
#define RSZVL_ATTR_NWA_CORR_DELETE_CAL_DATA                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4471
#define RSZVL_ATTR_NWA_CORR_DELETE_CAL_DATA_LAST                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4472
#define RSZVL_ATTR_NWA_LOAD_CAL_KIT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4473
#define RSZVL_ATTR_NWA_OFFSET_ELENGTH                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4474
#define RSZVL_ATTR_NWA_OFFSET_PERMITTIVITY                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4475
#define RSZVL_ATTR_NWA_OFFSET_DISTANCE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4476
#define RSZVL_ATTR_NWA_OFFSET_DELAY                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4477
#define RSZVL_ATTR_NWA_OFFSET_AUTO_LENGTH                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4478
#define RSZVL_ATTR_NWA_OFFSET_LOSS                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4479
#define RSZVL_ATTR_NWA_OFFSET_LOSS_OFFSET                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4480
#define RSZVL_ATTR_NWA_OFFSET_LOSS_FREQUENCY                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4481
#define RSZVL_ATTR_NWA_OFFSET_AUTO_LENGTH_AND_LOSS                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4482
#define RSZVL_ATTR_NWA_OFFSET_STATE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4483
#define RSZVL_ATTR_NWA_SOURCE_POWER                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4484
#define RSZVL_ATTR_NWA_TRACE_SELECT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4485
#define RSZVL_ATTR_NWA_ASSIGN_TRACE_BY_NAME                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4486
#define RSZVL_ATTR_NWA_TRACE_DELETE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4487
#define RSZVL_ATTR_NWA_TRACE_UNASSIGN_DIAGRAM_AREA                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4488
#define RSZVL_ATTR_NWA_TRACE_X_OFFSET                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4489
#define RSZVL_ATTR_NWA_TRACE_FORMAT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4490
#define RSZVL_ATTR_NWA_TRACE_APERTURE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4491
#define RSZVL_ATTR_NWA_TRACE_MATH_FUNC                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4492
#define RSZVL_ATTR_NWA_TRACE_SMO_STATE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4493
#define RSZVL_ATTR_NWA_TRACE_SMO_APER                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4494
#define RSZVL_ATTR_NWA_TRACE_MAX_HOLD                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4495
#define RSZVL_ATTR_NWA_TRACE_MATH_MEMORIZE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4496
#define RSZVL_ATTR_NWA_TRACE_DELETE_MEMORY                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4497
#define RSZVL_ATTR_NWA_TRACE_AUTOSCALE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4498
#define RSZVL_ATTR_NWA_TRACE_AUTOSCALE_BY_NAME                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4499
#define RSZVL_ATTR_NWA_TRACE_SCAL_BOTTOM                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4500
#define RSZVL_ATTR_NWA_TRACE_SCAL_TOP                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4501
#define RSZVL_ATTR_NWA_TRACE_SCAL_PDIV                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4502
#define RSZVL_ATTR_NWA_TRACE_REF_LEVEL                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4503
#define RSZVL_ATTR_NWA_TRACE_REF_POSITION                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4504
#define RSZVL_ATTR_NWA_TRACE_EVAL_RANGE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4505
#define RSZVL_ATTR_NWA_TRACE_EVAL_RANGE_START                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4506
#define RSZVL_ATTR_NWA_TRACE_EVAL_RANGE_STOP                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4507
#define RSZVL_ATTR_NWA_TRACE_EVAL_EPDELAY_STATE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4508
#define RSZVL_ATTR_NWA_TRACE_EVAL_MMPT_STATE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4509
#define RSZVL_ATTR_NWA_TRACE_EVAL_MSTD_STATE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4510
#define RSZVL_ATTR_NWA_TRACE_EVAL_RMS_STATE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4511
#define RSZVL_ATTR_NWA_TRACE_EVAL_ALL_STATE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4512
#define RSZVL_ATTR_NWA_CUSTOM_HEADER_STATE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4513
#define RSZVL_ATTR_NWA_CUSTOM_HEADER_INSTRUMENT_INFO                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4514
#define RSZVL_ATTR_NWA_CUSTOM_HEADER_DATE_AND_TIME_INFO                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4515
#define RSZVL_ATTR_NWA_CUSTOM_HEADER_TEXT_STATE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4516
#define RSZVL_ATTR_NWA_CUSTOM_HEADER_TEXT                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4517
#define RSZVL_ATTR_NWA_CUSTOM_HEADER_CALIBRATION_INFO                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4518
#define RSZVL_ATTR_NWA_MARKER_STATE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4519
#define RSZVL_ATTR_NWA_MARKER_X                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4520
#define RSZVL_ATTR_NWA_MARKER_REF_STATE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4521
#define RSZVL_ATTR_NWA_MARKER_REF_X                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4522
#define RSZVL_ATTR_NWA_MARKER_REF_MODE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4523
#define RSZVL_ATTR_NWA_MARKER_REF_FIXED                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4524
#define RSZVL_ATTR_NWA_DELTA_MARKER_STATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4525
#define RSZVL_ATTR_NWA_MARKER_COUPLED                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4526
#define RSZVL_ATTR_NWA_MARKER_MODE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4527
#define RSZVL_ATTR_NWA_MARKER_TYPE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4528
#define RSZVL_ATTR_NWA_MARKER_SEARCH_TRACKING                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4529
#define RSZVL_ATTR_NWA_MARKER_AOFF                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4530
#define RSZVL_ATTR_NWA_MARKER_TO_CENTER                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4531
#define RSZVL_ATTR_NWA_MARKER_TO_START                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4532
#define RSZVL_ATTR_NWA_MARKER_TO_STOP                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4533
#define RSZVL_ATTR_NWA_MARKER_FORMAT                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4534
#define RSZVL_ATTR_NWA_MARKER_SEARCH_MODE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4535
#define RSZVL_ATTR_NWA_MARKER_BFILTER_SEARCH_MODE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4536
#define RSZVL_ATTR_NWA_MARKER_SEARCH_RANGE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4537
#define RSZVL_ATTR_NWA_MARKER_SEARCH_RANGE_START                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4538
#define RSZVL_ATTR_NWA_MARKER_SEARCH_RANGE_STOP                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4539
#define RSZVL_ATTR_NWA_MARKER_TARGET_VALUE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4540
#define RSZVL_ATTR_NWA_MARKER_BANDWIDTH                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4541
#define RSZVL_ATTR_NWA_MARKER_BFILT_SEARCH_RESULTS                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4542
#define RSZVL_ATTR_NWA_SAVE_ALL_MARKERS                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4543
#define RSZVL_ATTR_NWA_LIMIT_DISPLAY_STATE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4544
#define RSZVL_ATTR_NWA_LIMIT_STATE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4545
#define RSZVL_ATTR_NWA_LIMIT_LOWER_STATE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4546
#define RSZVL_ATTR_NWA_LIMIT_UPPER_STATE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4547
#define RSZVL_ATTR_NWA_LIMIT_RDOM_COMPLEX                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4548
#define RSZVL_ATTR_NWA_LIMIT_RDOM_FORMAT                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4549
#define RSZVL_ATTR_NWA_LIMIT_RDOM_SPACING                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4550
#define RSZVL_ATTR_NWA_DISPLAY_LINE_STATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4551
#define RSZVL_ATTR_NWA_DISPLAY_LINE_POSITION                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4552
#define RSZVL_ATTR_NWA_LIMIT_SEGM_TYPE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4553
#define RSZVL_ATTR_NWA_LIMIT_SEGM_STIMUL_START                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4554
#define RSZVL_ATTR_NWA_LIMIT_SEGM_STIMUL_STOP                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4555
#define RSZVL_ATTR_NWA_LIMIT_SEGM_RESPONSE_START                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4556
#define RSZVL_ATTR_NWA_LIMIT_SEGM_RESPONSE_STOP                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4557
#define RSZVL_ATTR_NWA_LIMIT_SEGM_STIMUL_SHIFT                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4558
#define RSZVL_ATTR_NWA_LIMIT_SEGM_LOWER_SHIFT                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4559
#define RSZVL_ATTR_NWA_LIMIT_SEGM_UPPER_SHIFT                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4560
#define RSZVL_ATTR_NWA_LIMIT_SEGM_COUNT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4561
#define RSZVL_ATTR_NWA_LIMIT_DELETE_ALL_SEGM                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4562
#define RSZVL_ATTR_NWA_RIPPLE_CHECK_STATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4563
#define RSZVL_ATTR_NWA_GET_RIPPLE_GLOBAL_LIMIT_CHECK_RESULT                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4564
#define RSZVL_ATTR_NWA_RIPPLE_RANGE_SEGMENT_LIMIT_CHECK                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4565
#define RSZVL_ATTR_NWA_RIPPLE_SHOW_LIMITS                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4566
#define RSZVL_ATTR_NWA_RIPPLE_RANGE_SEGMENT_START                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4567
#define RSZVL_ATTR_NWA_RIPPLE_RANGE_SEGMENT_STOP                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4568
#define RSZVL_ATTR_NWA_RIPPLE_DELETE_ALL_LIMIT_RANGES                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4569
#define RSZVL_ATTR_NWA_RIPPLE_LIMIT_PHYSICAL_UNITS                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4570
#define RSZVL_ATTR_NWA_RIPPLE_LIMIT_RESPONSE_DOMAIN_FORMAT_UNITS            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4571
#define RSZVL_ATTR_NWA_RIPPLE_GET_LIMIT_RANGES                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4572
#define RSZVL_ATTR_NWA_RIPPLE_LIMIT_RANGE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4573
#define RSZVL_ATTR_NWA_LIMIT_CHECK_RESULT                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4574
#define RSZVL_ATTR_NWA_CLIMIT_CHECK_RESULT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4575
#define RSZVL_ATTR_NWA_REFER_MARKER_RESPONSE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4576
#define RSZVL_ATTR_NWA_MARKER_POSITION                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4577
#define RSZVL_ATTR_NWA_MARKER_AMPLITUDE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4578
#define RSZVL_ATTR_NWA_INIT                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4579
#define RSZVL_ATTR_NWA_PARAMETER_CONVERSION                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4580
#define RSZVL_ATTR_NWA_DFAULT_STATE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4581
#define RSZVL_ATTR_NWA_DFAULT_START                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4582
#define RSZVL_ATTR_NWA_DFAULT_STOP                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4583
#define RSZVL_ATTR_NWA_DFAULT_CENTER                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4584
#define RSZVL_ATTR_NWA_DFAULT_SPAN                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4585
#define RSZVL_ATTR_NWA_DFAULT_POINTS                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4586
#define RSZVL_ATTR_NWA_DFAULT_SELECT_CABLE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4587
#define RSZVL_ATTR_NWA_DFAULT_DELETE_CABLE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4588
#define RSZVL_ATTR_NWA_DFAULT_SAVE_CABLE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4589
#define RSZVL_ATTR_NWA_DFAULT_LOAD_CABLE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4590
#define RSZVL_ATTR_NWA_FAULT_LIMIT_CHECK                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4591
#define RSZVL_ATTR_NWA_FAULT_THRESHOLD                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4592
#define RSZVL_ATTR_NWA_FAULT_PEAK_COUNT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4593
#define RSZVL_ATTR_NWA_COLOR_SCHEME                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4594
#define RSZVL_ATTR_NWA_FREQ_INFO_STATE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4595
#define RSZVL_ATTR_NWA_TRACE_COLOR_STATE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4596
#define RSZVL_ATTR_NWA_MARKER_COLOR_STATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4597
#define RSZVL_ATTR_NWA_SAVE_COLOR_SCHEME                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4598
#define RSZVL_ATTR_NWA_LOAD_COLOR_SCHEME                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4599
#define RSZVL_ATTR_NWA_DISPLAY_UPDATE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4600
#define RSZVL_ATTR_NWA_ERROR_DISPLAY_STATE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4601
#define RSZVL_ATTR_NWA_REMOTE_DISPLAY_TITLE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4602
#define RSZVL_ATTR_NWA_DIAGRAM_AREA_STATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4603
#define RSZVL_ATTR_NWA_DIAGRAM_AREA_MAXIMIZE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4604
#define RSZVL_ATTR_NWA_DIAGRAM_AREA_TITLE_STATE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4605
#define RSZVL_ATTR_NWA_DIAGRAM_AREA_TITLE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4606
#define RSZVL_ATTR_NWA_CREATE_NEW_SETUP                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4607
#define RSZVL_ATTR_NWA_SELECT_SETUP                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4608
#define RSZVL_ATTR_NWA_CLOSE_SETUP                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4609
#define RSZVL_ATTR_NWA_LOAD_SETUP                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4610
#define RSZVL_ATTR_NWA_SAVE_SETUP                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4611
#define RSZVL_ATTR_NWA_SYSTEM_PRESET_SCOPE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4612
#define RSZVL_ATTR_NWA_USER_PRESET_STATE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4613
#define RSZVL_ATTR_NWA_USER_PRESET_NAME                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4614
#define RSZVL_ATTR_NWA_SYSTEM_SETTINGS_UPDATE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4615
#define RSZVL_ATTR_NWA_SYSTEM_KLOCK                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4616
#define RSZVL_ATTR_NWA_ALARM_SOUND_STATE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4617
#define RSZVL_ATTR_NWA_STATUS_SOUND_STATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4618
#define RSZVL_ATTR_NWA_UNIT_OF_LENGTH                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4619
#define RSZVL_ATTR_FMDEM_MODE                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4620
#define RSZVL_ATTR_FMDEM_TRIGGER_SOURCE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4621
#define RSZVL_ATTR_ADEM_TRIGGER_AM_LEVEL_ABS                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4622
#define RSZVL_ATTR_ADEM_TRIGGER_AM_LEVEL_REL                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4623
#define RSZVL_ATTR_ADEM_TRIGGER_FM_LEVEL                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4624
#define RSZVL_ATTR_ADEM_TRIGGER_PM_LEVEL                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4625
#define RSZVL_ATTR_FMDEM_DISPLAY_PDIV                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4626
#define RSZVL_ATTR_FMDEM_STATE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4627
#define RSZVL_ATTR_FMDEM_AF_COUP                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4628
#define RSZVL_ATTR_FMDEM_MTIM                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4629
#define RSZVL_ATTR_FMDEM_BAND_DEM                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4630
#define RSZVL_ATTR_FMDEM_RLEN                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4631
#define RSZVL_ATTR_FMDEM_SRATE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4632
#define RSZVL_ATTR_FMDEM_ZOOM                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4633
#define RSZVL_ATTR_FMDEM_ZOOM_START                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4634
#define RSZVL_ATTR_ADEM_PHASE_WRAP                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4635
#define RSZVL_ATTR_ADEM_UNIT_ANGLE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4636
#define RSZVL_ATTR_ADEM_PM_RPO_X                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4637
#define RSZVL_ATTR_FMDEM_AF_CENTER                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4638
#define RSZVL_ATTR_FMDEM_AF_SPAN                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4639
#define RSZVL_ATTR_FMDEM_AF_START                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4640
#define RSZVL_ATTR_FMDEM_AF_STOP                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4641
#define RSZVL_ATTR_FMDEM_AF_FULL_SPAN                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4642
#define RSZVL_ATTR_FMDEM_BAND_RES                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4643
#define RSZVL_ATTR_FMDEM_SPEC_SPAN                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4644
#define RSZVL_ATTR_FMDEM_SPEC_ZOOM                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4645
#define RSZVL_ATTR_FMDEM_FILT_HPAS_STAT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4646
#define RSZVL_ATTR_FMDEM_FILT_HPAS_FREQ                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4647
#define RSZVL_ATTR_FMDEM_FILT_LPAS_STAT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4648
#define RSZVL_ATTR_FMDEM_FILT_DEMP_STAT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4649
#define RSZVL_ATTR_FMDEM_FILT_LPAS_FREQ                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4650
#define RSZVL_ATTR_FMDEM_FILT_DEMP_TCON                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4651
#define RSZVL_ATTR_ADEM_FILT_LPAS_FREQ_REL                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4652
#define RSZVL_ATTR_FMDEM_AFR_RES                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4653
#define RSZVL_ATTR_FMDEM_FERR_RES                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4654
#define RSZVL_ATTR_FMDEM_SIN_RES                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4655
#define RSZVL_ATTR_FMDEM_THD_RES                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4656
#define RSZVL_ATTR_FMDEM_CARR_RES                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4657
#define RSZVL_ATTR_ADEM_SUMM_RES                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4658
#define RSZVL_ATTR_ADEM_FM_OFFSET                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4659
#define RSZVL_ATTR_BTO_PTYPE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4660
#define RSZVL_ATTR_BTO_PRAT                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4661
#define RSZVL_ATTR_BTO_PCL                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4662
#define RSZVL_ATTR_BTO_POW_AVER                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4663
#define RSZVL_ATTR_BTO_BWID_AUTO                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4664
#define RSZVL_ATTR_BTO_BWID                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4665
#define RSZVL_ATTR_BTO_VID_AUTO                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4666
#define RSZVL_ATTR_BTO_VID                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4667
#define RSZVL_ATTR_BTO_FILT_MEAS                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4668
#define RSZVL_ATTR_BTO_TRAC_MOD                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4669
#define RSZVL_ATTR_BTO_TRAC                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4670
#define RSZVL_ATTR_BTO_TRAC_DET                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4671
#define RSZVL_ATTR_BTO_SWE                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4672
#define RSZVL_ATTR_BTO_SWE_TIME_AUTO                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4673
#define RSZVL_ATTR_BTO_SWE_TIME                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4674
#define RSZVL_ATTR_BTO_MODE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4675
#define RSZVL_ATTR_BTO                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4676
#define RSZVL_ATTR_BTO_LAP                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4677
#define RSZVL_ATTR_BTO_CHANNEL                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4678
#define RSZVL_ATTR_BTO_ACLR_ACP                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4679
#define RSZVL_ATTR_BTO_EGA_INP                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4680
#define RSZVL_ATTR_BTO_PULS_OFFS                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4681
#define RSZVL_ATTR_BTO_SEARCH_PULS                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4682
#define RSZVL_ATTR_BTO_SYNC_OFFS                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4683
#define RSZVL_ATTR_BTO_SEARCH_SYNC                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4684
#define RSZVL_ATTR_BTO_SEARCH_TIME                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4685
#define RSZVL_ATTR_BTO_SEARCH_TIME_AUTO                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4686
#define RSZVL_ATTR_BTO_CFST_BLOCK_COUNT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4687
#define RSZVL_ATTR_BTO_IBS_ACP                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4688
#define RSZVL_ATTR_BTO_IBS_GATE_AUTO                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4689
#define RSZVL_ATTR_BTO_PBSC                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4690
#define RSZVL_ATTR_BTO_RTP_TIME                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4691
#define RSZVL_ATTR_BTO_PDIV                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4692
#define RSZVL_ATTR_BTO_ZOOM_STATE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4693
#define RSZVL_ATTR_BTO_ZOOM_START                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4694
#define RSZVL_ATTR_BTO_OPOW_PEAK                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4695
#define RSZVL_ATTR_BTO_OPOW_AVER                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4696
#define RSZVL_ATTR_BTO_STAT                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4697
#define RSZVL_ATTR_BTO_ACLR_EXC                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4698
#define RSZVL_ATTR_BTO_MCH_DF_AVER                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4699
#define RSZVL_ATTR_BTO_MCH_DF_MAX                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4700
#define RSZVL_ATTR_BTO_MCH_RAT                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4701
#define RSZVL_ATTR_BTO_MCH_DF2_PERC                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4702
#define RSZVL_ATTR_BTO_ICFT                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4703
#define RSZVL_ATTR_BTO_CFDR_MAX                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4704
#define RSZVL_ATTR_BTO_CFDR_RATE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4705
#define RSZVL_ATTR_BTO_PLEN                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4706
#define RSZVL_ATTR_BTO_PTYP                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4707
#define RSZVL_ATTR_BTO_CFST_BLOCK_COUNT_VALUE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4708
#define RSZVL_ATTR_BTO_CFST_DEVM_RMS                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4709
#define RSZVL_ATTR_BTO_CFST_DEVM_PEAK                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4710
#define RSZVL_ATTR_BTO_CFST_DEVM_99PCT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4711
#define RSZVL_ATTR_BTO_CFST_FERR_TOTAL                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4712
#define RSZVL_ATTR_BTO_CFST_FERR_BLOCK                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4713
#define RSZVL_ATTR_BTO_CFST_FERR_INITIAL                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4714
#define RSZVL_ATTR_BTO_DPEN_PACKETS                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4715
#define RSZVL_ATTR_BTO_DPEN_PACKETS_PASSED                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4716
#define RSZVL_ATTR_BTO_DPEN_BER                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4717
#define RSZVL_ATTR_BTO_IBS_EXC                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4718
#define RSZVL_ATTR_BTO_IBS_HADJ                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4719
#define RSZVL_ATTR_BTO_IBS_TXCHANNEL                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4720
#define RSZVL_ATTR_BTO_RTP                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4721
#define RSZVL_ATTR_BTO_RTP_RATIO                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4722
#define RSZVL_ATTR_PMET_REL                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4723
#define RSZVL_ATTR_PMET_REL_AUTO                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4724
#define RSZVL_ATTR_PMET_FREQ                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4725
#define RSZVL_ATTR_PMET_MEAS_TIME                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4726
#define RSZVL_ATTR_PMET_COUPLING                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4727
#define RSZVL_ATTR_PMET_ZERO                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4728
#define RSZVL_ATTR_PMET_UNIT_ABS                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4729
#define RSZVL_ATTR_PMET_UNIT_REL                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4730
#define RSZVL_ATTR_PMET_STATE                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4731
#define RSZVL_ATTR_PMET_AVERAGE_AUTO                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4732
#define RSZVL_ATTR_PMET_AVERAGE_COUNT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4733
#define RSZVL_ATTR_PMET_REF_LEVEL_OFFSET_STATE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4734
#define RSZVL_ATTR_PMET_RELATIVE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4735
#define RSZVL_ATTR_PMET_READ                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4736
#define RSZVL_ATTR_PMET_FETCH                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4737
#define RSZVL_ATTR_SPEM_STATE                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4738
#define RSZVL_ATTR_SPEM_COLOR                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4739
#define RSZVL_ATTR_SPEM_SIZE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4740
#define RSZVL_ATTR_SPEM_HDEP                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4741
#define RSZVL_ATTR_SPEM_TIME_STAMP_STATE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4742
#define RSZVL_ATTR_SPEM_FRAME_COUNT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4743
#define RSZVL_ATTR_SPEM_FRAME_SELECT                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4744
#define RSZVL_ATTR_SPEM_CONTINUOUS                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4745
#define RSZVL_ATTR_SPEM_CLEAR                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4746
#define RSZVL_ATTR_FILE_ITEMS_SAVE_RECAL_SPEM                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4747
#define RSZVL_ATTR_SPEM_MARK_FRAME_NUMBER                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4748
#define RSZVL_ATTR_SPEM_MARKER_SEARCH_AREA                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4749
#define RSZVL_ATTR_SPEM_MARKER_SEARCH_XY                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4750
#define RSZVL_ATTR_SPEM_MARKER_SEARCH_Y                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4751
#define RSZVL_ATTR_SPEM_DELT_FRAME_NUMBER                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4752
#define RSZVL_ATTR_SPEM_DELT_SEARCH_AREA                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4753
#define RSZVL_ATTR_SPEM_DELT_SEARCH_XY                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4754
#define RSZVL_ATTR_SPEM_DELT_SEARCH_Y                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4755
#define RSZVL_ATTR_CATV_MODE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4756
#define RSZVL_ATTR_CATV_Y_SCALE_AUTO                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4757
#define RSZVL_ATTR_CATV_EPAT_ZOOM_STATE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4758
#define RSZVL_ATTR_CATV_EPAT_ZOOM_FACTOR                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4759
#define RSZVL_ATTR_CATV_MERR_ZOOM                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4760
#define RSZVL_ATTR_CATV_OVER_ZOOM                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4761
#define RSZVL_ATTR_CATV_QUAD_ZOOM                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4762
#define RSZVL_ATTR_CATV_TRACE_MODE_FREEZE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4763
#define RSZVL_ATTR_CATV_CHANNEL                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4764
#define RSZVL_ATTR_CATV_RF_FREQUENCY                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4765
#define RSZVL_ATTR_CATV_RF_FREQUENCY_STEP_SIZE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4766
#define RSZVL_ATTR_CATV_SWEEP_SPACING                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4767
#define RSZVL_ATTR_CATV_SIDE_BAND                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4768
#define RSZVL_ATTR_CATV_MODUL_STANDARD                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4769
#define RSZVL_ATTR_CATV_MODUL_STANDARD_GDELAY                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4770
#define RSZVL_ATTR_CATV_MODUL_STANDARD_SIG_TYPE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4771
#define RSZVL_ATTR_CATV_ATV_MEAS                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4772
#define RSZVL_ATTR_CATV_ATV_CARR_MEAS                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4773
#define RSZVL_ATTR_CATV_ATV_MEAS_MODE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4774
#define RSZVL_ATTR_CATV_ATV_BWID                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4775
#define RSZVL_ATTR_CATV_ATV_FREQ_NEXT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4776
#define RSZVL_ATTR_CATV_ATV_NOISE_FLOOR_CORRECTION                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4777
#define RSZVL_ATTR_CATV_ATV_REF_CHANNEL_MANUAL                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4778
#define RSZVL_ATTR_CATV_ATV_REF_POWER_MANUAL                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4779
#define RSZVL_ATTR_CATV_ATV_REF_POWER_MODE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4780
#define RSZVL_ATTR_CATV_ATV_HUM_UNIT                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4781
#define RSZVL_ATTR_CATV_ATV_VCP_UNIT                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4782
#define RSZVL_ATTR_CATV_ATV_STANDARD                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4783
#define RSZVL_ATTR_CATV_ATV_SOUND_SYSTEM                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4784
#define RSZVL_ATTR_CATV_ATV_COLOR                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4785
#define RSZVL_ATTR_CATV_ATV_TRIGGER_BAR_FIELD                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4786
#define RSZVL_ATTR_CATV_ATV_TRIGGER_BAR_LINE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4787
#define RSZVL_ATTR_CATV_ATV_TRIGGER_BAR_LINE_TYPE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4788
#define RSZVL_ATTR_CATV_ATV_TRIGGER_VIDEO_SCOPE_FIELD                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4789
#define RSZVL_ATTR_CATV_ATV_TRIGGER_VIDEO_SCOPE_LINE                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4790
#define RSZVL_ATTR_CATV_ATV_TRIGGER_QLINE_FIELD                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4791
#define RSZVL_ATTR_CATV_ATV_TRIGGER_QUIET_LINE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4792
#define RSZVL_ATTR_CATV_ATV_CARR_LIM_FREQ_OFFSET                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4793
#define RSZVL_ATTR_CATV_ATV_CARR_LIM_REL_POWER                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4794
#define RSZVL_ATTR_CATV_ATV_CARR_LIM_VCF                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4795
#define RSZVL_ATTR_CATV_ATV_CARR_LIM_VCP                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4796
#define RSZVL_ATTR_CATV_ATV_LIM_CN                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4797
#define RSZVL_ATTR_CATV_ATV_LIM_CSO                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4798
#define RSZVL_ATTR_CATV_ATV_LIM_CTB                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4799
#define RSZVL_ATTR_CATV_ATV_LIM_HUM_LOW                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4800
#define RSZVL_ATTR_CATV_ATV_LIM_HUM_UPP                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4801
#define RSZVL_ATTR_CATV_ATV_LIM_VMOD_MDEP                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4802
#define RSZVL_ATTR_CATV_ATV_LIM_VMOD_RPC                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4803
#define RSZVL_ATTR_CATV_ATV_LIM_VMOD_VCP                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4804
#define RSZVL_ATTR_CATV_ATV_CARR_LIM_FREQ_OFFSET_RESULT                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4805
#define RSZVL_ATTR_CATV_ATV_CARR_LIM_REL_POWER_RESULT                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4806
#define RSZVL_ATTR_CATV_ATV_CARR_LIM_VCF_RESULT                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4807
#define RSZVL_ATTR_CATV_ATV_CARR_LIM_VCP_RESULT                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4808
#define RSZVL_ATTR_CATV_ATV_LIM_CN_RESULT                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4809
#define RSZVL_ATTR_CATV_ATV_LIM_CSO_RESULT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4810
#define RSZVL_ATTR_CATV_ATV_LIM_CTB_RESULT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4811
#define RSZVL_ATTR_CATV_ATV_LIM_HUM_RESULT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4812
#define RSZVL_ATTR_CATV_ATV_LIM_VMOD_MDEP_RESULT                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4813
#define RSZVL_ATTR_CATV_ATV_CARR_VCPA                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4814
#define RSZVL_ATTR_CATV_ATV_CARR_VCF_VALUE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4815
#define RSZVL_ATTR_CATV_ATV_CARR_REL_POWER_VALUE                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4816
#define RSZVL_ATTR_CATV_ATV_CARR_FREQ_OFFSET_VALUE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4817
#define RSZVL_ATTR_CATV_ATV_VALUE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4818
#define RSZVL_ATTR_CATV_ATV_CN_CN_VALUE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4819
#define RSZVL_ATTR_CATV_ATV_CSO_CSO_VALUE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4820
#define RSZVL_ATTR_CATV_ATV_CTB_CTB_VALUE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4821
#define RSZVL_ATTR_CATV_ATV_HUM_VALUE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4822
#define RSZVL_ATTR_CATV_ATV_VMOD_RESULT_VALUE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4823
#define RSZVL_ATTR_CATV_DTV_MEAS                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4824
#define RSZVL_ATTR_CATV_DTV_SATT                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4825
#define RSZVL_ATTR_CATV_DTV_EPAT_UNIT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4826
#define RSZVL_ATTR_CATV_DTV_EPAT_RVEL                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4827
#define RSZVL_ATTR_CATV_DTV_ERROR_UNIT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4828
#define RSZVL_ATTR_CATV_DTV_EQUALIZER_STATE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4829
#define RSZVL_ATTR_CATV_DTV_EQUALIZER_FREEZE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4830
#define RSZVL_ATTR_CATV_DTV_EQUALIZER_RESET                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4831
#define RSZVL_ATTR_CATV_DTV_FILTER_ALPHA                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4832
#define RSZVL_ATTR_CATV_DTV_SRATE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4833
#define RSZVL_ATTR_CATV_DTV_STANDARD                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4834
#define RSZVL_ATTR_CATV_DTV_FORMAT                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4835
#define RSZVL_ATTR_CATV_DTV_CARR_LIM_FREQ_OFFSET                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4836
#define RSZVL_ATTR_CATV_DTV_LIM_CHANNEL_POWER                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4837
#define RSZVL_ATTR_CATV_DTV_LIM_SUPP                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4838
#define RSZVL_ATTR_CATV_DTV_LIM_ERR_LOW                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4839
#define RSZVL_ATTR_CATV_DTV_LIM_ERR_UPP                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4840
#define RSZVL_ATTR_CATV_DTV_LIM_IMB                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4841
#define RSZVL_ATTR_CATV_DTV_LIM_PJIT                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4842
#define RSZVL_ATTR_CATV_DTV_LIM_QERR                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4843
#define RSZVL_ATTR_CATV_DTV_LIM_SYMB_RATE_OFFSET                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4844
#define RSZVL_ATTR_CATV_DTV_CARR_LIM_FREQ_OFFSET_RESULT                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4845
#define RSZVL_ATTR_CATV_DTV_LIM_CHANNEL_POWER_RESULT                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4846
#define RSZVL_ATTR_CATV_DTV_CARR_LIM_CARR_SUPP_RESULT                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4847
#define RSZVL_ATTR_CATV_DTV_LIM_ERROR_RESULT                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4848
#define RSZVL_ATTR_CATV_DTV_LIM_IMB_RESULT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 4849
#define RSZVL_ATTR_CATV_DTV_LIM_PJIT_RESULT                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4850
#define RSZVL_ATTR_CATV_DTV_LIM_QERR_RESULT                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4851
#define RSZVL_ATTR_CATV_DTV_LIM_SYMB_RATE_OFFSET_RESULT                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4852
#define RSZVL_ATTR_CATV_DTV_SPEC                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4853
#define RSZVL_ATTR_CATV_DTV_ERROR                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4854
#define RSZVL_ATTR_CATV_DTV_OVER_CARR_FREQ_OFFSET                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4855
#define RSZVL_ATTR_CATV_DTV_OVER_SRATE_OFFSET                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4856
#define RSZVL_ATTR_CATV_DTV_MERR_IMB                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4857
#define RSZVL_ATTR_CATV_DTV_MERR_QERR                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4858
#define RSZVL_ATTR_CATV_DTV_MERR_CSUP                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4859
#define RSZVL_ATTR_CATV_DTV_MERR_PJIT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4860
#define RSZVL_ATTR_CATV_DTV_CHANNEL_POWER_RESULT                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4861
#define RSZVL_ATTR_CATV_TV_MEAS                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4862
#define RSZVL_ATTR_CATV_TV_CHANNEL                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4863
#define RSZVL_ATTR_CATV_TV_MOD_STANDARD_TILT_ACTIVATE                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4864
#define RSZVL_ATTR_CATV_TV_MOD_STANDARD_TILT_DEACTIVATE                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4865
#define RSZVL_ATTR_NOISE_MODE                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4866
#define RSZVL_ATTR_NOISE_CALC_LIM_TRAC                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4867
#define RSZVL_ATTR_NOISE_CONF_CORR                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4868
#define RSZVL_ATTR_NOISE_CONF_LIST_CONT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4869
#define RSZVL_ATTR_NOISE_CONF_LIST_SING                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4870
#define RSZVL_ATTR_NOISE_CONF_SING                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4871
#define RSZVL_ATTR_NOISE_INP_GAIN_STAT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4872
#define RSZVL_ATTR_NOISE_CORR_ENR_MODE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4873
#define RSZVL_ATTR_NOISE_CORR_ENR_SPOT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4874
#define RSZVL_ATTR_NOISE_CORR_TEMP                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4875
#define RSZVL_ATTR_NOISE_CORR_LOSS_MODE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4876
#define RSZVL_ATTR_NOISE_CORR_LOSS_SPOT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4877
#define RSZVL_ATTR_NOISE_CORR_EREJ                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4878
#define RSZVL_ATTR_NOISE_CORR_STAT                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4879
#define RSZVL_ATTR_NOISE_FREQ_STEP                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4880
#define RSZVL_ATTR_NOISE_FREQ_FIXED                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4881
#define RSZVL_ATTR_NOISE_CONF_SYST_LOSC_FREQ                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4882
#define RSZVL_ATTR_NOISE_CONF_MODE_DUT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4883
#define RSZVL_ATTR_NOISE_CONF_DUT_GAIN                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4884
#define RSZVL_ATTR_NOISE_CONF_DUT_SETT_TIME                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4885
#define RSZVL_ATTR_NOISE_ARRAY_MEMORY                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4886
#define RSZVL_ATTR_NOISE_MARKER_TRACE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4887
#define RSZVL_ATTR_NOISE_DISP_DATA_TRAC                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4888
#define RSZVL_ATTR_NOISE_DISP_TRAC_STAT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4889
#define RSZVL_ATTR_NOISE_DISP_SYMB                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4890
#define RSZVL_ATTR_NOISE_DISP_AUTO_SCAL                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4891
#define RSZVL_ATTR_NOISE_DISP_TRAC_BOTTOM                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4892
#define RSZVL_ATTR_NOISE_DISP_TRAC_TOP                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4893
#define RSZVL_ATTR_NOISE_DISP_TRAC_REF_LEV_STAT                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4894
#define RSZVL_ATTR_NOISE_DISP_FORMAT                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4895
#define RSZVL_ATTR_NOISE_DISP_TABLE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4896
#define RSZVL_ATTR_NOISE_DISP_TRACE_MEM                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4897
#define RSZVL_ATTR_NOISE_DISPL_CURR_RESULTS                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4898
#define RSZVL_ATTR_NOISE_DISP_X_AXIS                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4899
#define RSZVL_ATTR_BWCD_MODE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4900
#define RSZVL_ATTR_3GBS_MEAS                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4901
#define RSZVL_ATTR_3GPP_LEVEL_ADJUST                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4902
#define RSZVL_ATTR_3GBS_WCDP_HSDPAUPA_STATE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4903
#define RSZVL_ATTR_3GBS_WCDP_SYNC_TYPE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4904
#define RSZVL_ATTR_3GBS_WCDP_NORMALIZE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4905
#define RSZVL_ATTR_3GBS_WCDP_INVERT_Q                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4906
#define RSZVL_ATTR_3GBS_WCDP_POW_REF                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4907
#define RSZVL_ATTR_3GBS_WCDP_INACT_CHAN_THR                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4908
#define RSZVL_ATTR_3GBS_ANTENNA                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4909
#define RSZVL_ATTR_3GBS_MARKER_TO_CHANNEL                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4910
#define RSZVL_ATTR_3GBS_REFERENCE_MARKER_TO_CHANNEL                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4911
#define RSZVL_ATTR_3GBS_SCODE                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4912
#define RSZVL_ATTR_3GPP_PDIV                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4913
#define RSZVL_ATTR_3GPP_ASEQ                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4914
#define RSZVL_ATTR_3GPP_SCODE_SEARCH                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4915
#define RSZVL_ATTR_3GPP_ESP_LIMIT_LINE_SELECT                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 4916
#define RSZVL_ATTR_3GPP_ESP_PREDEFINED_LIMIT_LINE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4917
#define RSZVL_ATTR_3GPP_ESP_TRANSITION                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4918
#define RSZVL_ATTR_3GPP_ESP_LIMIT_LINE_RESTORE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4919
#define RSZVL_ATTR_3GBS_RESULTS_SELECT_CHAN                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4920
#define RSZVL_ATTR_3GBS_RESULTS_SELECT_CPICH_SLOT                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4921
#define RSZVL_ATTR_3GBS_RESULT                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4922
#define RSZVL_ATTR_3GBS_RESULT_SUM_VIEW                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 4923
#define RSZVL_ATTR_WLAN_STAN                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4924
#define RSZVL_ATTR_WLAN_PVT_RPOW                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4925
#define RSZVL_ATTR_WLAN_PVT_SEL                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4926
#define RSZVL_ATTR_WLAN_PVT                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4927
#define RSZVL_ATTR_WLAN_SPEC_FLAT                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4928
#define RSZVL_ATTR_WLAN_BTYP                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4929
#define RSZVL_ATTR_WLAN_PVT_AVER                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4930
#define RSZVL_ATTR_WLAN_POW_EXP_RF                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 4931
#define RSZVL_ATTR_WLAN_POW_AUTO                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4932
#define RSZVL_ATTR_WLAN_POW_AUTO_ONCE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4933
#define RSZVL_ATTR_WLAN_CHAN                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4934
#define RSZVL_ATTR_WLAN_EVM_ECAR                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4935
#define RSZVL_ATTR_WLAN_EVM_ESYM                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4936
#define RSZVL_ATTR_WLAN_SPEC_MASK                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4937
#define RSZVL_ATTR_WLAN_SPEC_MASK_SEL                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4938
#define RSZVL_ATTR_WLAN_SPEC_FFT                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4939
#define RSZVL_ATTR_WLAN_SPEC_ACPR                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4940
#define RSZVL_ATTR_WLAN_CONS_CSYM                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4941
#define RSZVL_ATTR_WLAN_BURS_COUN_STAT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4942
#define RSZVL_ATTR_WLAN_CONS_CCAR                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4943
#define RSZVL_ATTR_WLAN_CONS_CARR_SEL                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4944
#define RSZVL_ATTR_WLAN_CONS_CARR_SEL_NR                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 4945
#define RSZVL_ATTR_WLAN_STAT_CCDF                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4946
#define RSZVL_ATTR_WLAN_STAT_BSTR                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4947
#define RSZVL_ATTR_WLAN_STAT_SFI                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4948
#define RSZVL_ATTR_WLAN_INIT                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4949
#define RSZVL_ATTR_WLAN_INIT_CONT                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4950
#define RSZVL_ATTR_WLAN_MODE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4951
#define RSZVL_ATTR_WLAN_BURS_COUN                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4952
#define RSZVL_ATTR_WLAN_SWAP_IQ                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4953
#define RSZVL_ATTR_WLAN_ACH_MODE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4954
#define RSZVL_ATTR_WLAN_UNIT_PRE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4955
#define RSZVL_ATTR_WLAN_UNIT_EVM                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4956
#define RSZVL_ATTR_WLAN_UNIT_GIMB                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4957
#define RSZVL_ATTR_WLAN_PRE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4958
#define RSZVL_ATTR_WLAN_SRAT                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4959
#define RSZVL_ATTR_WLAN_POW_AUTO_SWE_TIME                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 4960
#define RSZVL_ATTR_WLAN_IQ_LOAD                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4961
#define RSZVL_ATTR_WLAN_IQ_SAVE                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4962
#define RSZVL_ATTR_WLAN_CALC_BURS                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4963
#define RSZVL_ATTR_WLAN_PRE_SEL                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4964
#define RSZVL_ATTR_WLAN_DISP_TABL                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4965
#define RSZVL_ATTR_WLAN_SCREEN_SELECT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 4966
#define RSZVL_ATTR_WLAN_DISP_FORMAT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4967
#define RSZVL_ATTR_WLAN_DISP_AUTO_Y                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4968
#define RSZVL_ATTR_WLAN_Y_PDIV                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 4969
#define RSZVL_ATTR_WLAN_TRIS                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4970
#define RSZVL_ATTR_WLAN_TFAL                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4971
#define RSZVL_ATTR_WLAN_FERR                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4972
#define RSZVL_ATTR_WLAN_SYMB                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4973
#define RSZVL_ATTR_WLAN_IQOF                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4974
#define RSZVL_ATTR_WLAN_EVM                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 4975
#define RSZVL_ATTR_WLAN_EVM_ALL                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4976
#define RSZVL_ATTR_WLAN_EVM_DATA                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4977
#define RSZVL_ATTR_WLAN_EVM_PIL                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4978
#define RSZVL_ATTR_WLAN_TFAL_RES                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4979
#define RSZVL_ATTR_WLAN_TRIS_RES                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4980
#define RSZVL_ATTR_WLAN_FERR_RES                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4981
#define RSZVL_ATTR_WLAN_SYMB_RES                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4982
#define RSZVL_ATTR_WLAN_IQOF_RES                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 4983
#define RSZVL_ATTR_WLAN_EVM_RES                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4984
#define RSZVL_ATTR_WLAN_LIM_FAIL_RES                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4985
#define RSZVL_ATTR_WLAN_EVM_ALL_RES                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4986
#define RSZVL_ATTR_WLAN_EVM_DATA_RES                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4987
#define RSZVL_ATTR_WLAN_EVM_PIL_RES                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4988
#define RSZVL_ATTR_WLAN_ACP_ADJ                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4989
#define RSZVL_ATTR_WLAN_ACP_ADJ_RES                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4990
#define RSZVL_ATTR_WLAN_ACP_ALT                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 4991
#define RSZVL_ATTR_WLAN_ACP_ALT_RES                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4992
#define RSZVL_ATTR_WLAN_SPEC_MASK_X                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4993
#define RSZVL_ATTR_WLAN_SPEC_MASK_Y                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 4994
#define RSZVL_ATTR_WLAN_MARKER_ENABLED                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 4995
#define RSZVL_ATTR_WLAN_MARKERS_AOFF                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4996
#define RSZVL_ATTR_WLAN_MARKER_TRACE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 4997
#define RSZVL_ATTR_WLAN_MARK_SYMB                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4998
#define RSZVL_ATTR_WLAN_MARK_CARR                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 4999
#define RSZVL_ATTR_WLAN_MARK_Y                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 5000
#define RSZVL_ATTR_WLAN_MARKER_POSITION                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 5001
#define RSZVL_ATTR_WLAN_MARKER_ZOOM                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 5002
#define RSZVL_ATTR_WLAN_MARKER_SEARCH_MAX                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 5003
#define RSZVL_ATTR_WLAN_MARKER_SEARCH_MIN                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 5004
#define RSZVL_ATTR_WLAN_FETC_BURS_COUNT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 5005
#define RSZVL_ATTR_WLAN_FETC_PRE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5006
#define RSZVL_ATTR_WLAN_FETC_PAYL                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5007
#define RSZVL_ATTR_WLAN_FETC_RMS                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5008
#define RSZVL_ATTR_WLAN_FETC_PEAK                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5009
#define RSZVL_ATTR_WLAN_FETC_CRES                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5010
#define RSZVL_ATTR_WLAN_FETC_TRIS                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5011
#define RSZVL_ATTR_WLAN_FETC_TFAL                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5012
#define RSZVL_ATTR_WLAN_FETC_FERR                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5013
#define RSZVL_ATTR_WLAN_FETC_SYMB                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5014
#define RSZVL_ATTR_WLAN_FETC_IQOF                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5015
#define RSZVL_ATTR_WLAN_FETC_GIMB                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5016
#define RSZVL_ATTR_WLAN_FETC_QUAD                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5017
#define RSZVL_ATTR_WLAN_FETC_EVM                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5018
#define RSZVL_ATTR_WLAN_FETC_EVM_DIR                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 5019
#define RSZVL_ATTR_WLAN_FETC_EVM_ALL                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 5020
#define RSZVL_ATTR_WLAN_FETC_EVM_DATA                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 5021
#define RSZVL_ATTR_WLAN_FETC_EVM_PILOT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 5022
#define RSZVL_ATTR_WLAN_FETC_SYMB_COUNT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 5023
#define RSZVL_ATTR_WLAN_EGAT_HOLD_SAMP                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 5024
#define RSZVL_ATTR_WLAN_EGAT_HOLD_TIME                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 5025
#define RSZVL_ATTR_WLAN_EGAT_LENG_SAMP                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 5026
#define RSZVL_ATTR_WLAN_EGAT_LENG_TIME                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 5027
#define RSZVL_ATTR_WLAN_EGAT_LINK                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5028
#define RSZVL_ATTR_WLAN_EGAT                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 5029
#define RSZVL_ATTR_WLAN_TRIG_MODE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5030
#define RSZVL_ATTR_WLAN_TRIG_POW_LEV                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 5031
#define RSZVL_ATTR_WLAN_TRIG_POW_LEV_AUTO                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 5032
#define RSZVL_ATTR_WLAN_CEST                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 5033
#define RSZVL_ATTR_WLAN_SIGS                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 5034
#define RSZVL_ATTR_WLAN_BCON_AUTO                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5035
#define RSZVL_ATTR_WLAN_BAN                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 5036
#define RSZVL_ATTR_WLAN_SYM_EQU                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 5037
#define RSZVL_ATTR_WLAN_SYM                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 5038
#define RSZVL_ATTR_WLAN_DBYT_EQU                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5039
#define RSZVL_ATTR_WLAN_DBYT                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 5040
#define RSZVL_ATTR_WLAN_DUR_EQU                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 5041
#define RSZVL_ATTR_WLAN_DUR                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 5042
#define RSZVL_ATTR_WLAN_TRAC_LEV                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5043
#define RSZVL_ATTR_WLAN_TRAC_PHAS                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5044
#define RSZVL_ATTR_WLAN_TRAC_TIME                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5045
#define RSZVL_ATTR_WIMAX_STAN                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 5046
#define RSZVL_ATTR_WIMAX_MODE                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 5047
#define RSZVL_ATTR_WIMAX_BAND_CHAN                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5048
#define RSZVL_ATTR_WIMAX_BURS_COUN_STAT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 5049
#define RSZVL_ATTR_WIMAX_BURS_COUN                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5050
#define RSZVL_ATTR_WIMAX_FBAN                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 5051
#define RSZVL_ATTR_WIMAX_CHAN                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 5052
#define RSZVL_ATTR_WIMAX_IGR                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 5053
#define RSZVL_ATTR_WIMAX_POW_AUTO                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5054
#define RSZVL_ATTR_WIMAX_POW_AUTO_ONCE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 5055
#define RSZVL_ATTR_WIMAX_POW_AUTO_SWE_TIME                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 5056
#define RSZVL_ATTR_WIMAX_POW_EXP_RF                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 5057
#define RSZVL_ATTR_WIMAX_SIGNAL_INPUT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 5058
#define RSZVL_ATTR_WIMAX_RLEV_RF                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5059
#define RSZVL_ATTR_WIMAX_SRAT                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 5060
#define RSZVL_ATTR_WIMAX_IQ_LOAD                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5061
#define RSZVL_ATTR_WIMAX_IQ_SAVE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5062
#define RSZVL_ATTR_WIMAX_SWAP_IQ                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5063
#define RSZVL_ATTR_WIMAX_TABLE_UNIT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 5064
#define RSZVL_ATTR_WIMAX_RSSI_AVG_PARAM                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 5065
#define RSZVL_ATTR_WIMAX_CALC_BURS                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5066
#define RSZVL_ATTR_WIMAX_FORMAT_SELECT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 5067
#define RSZVL_ATTR_WIMAX_NFFT                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 5068
#define RSZVL_ATTR_WIMAX_FRAME_LOAD                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 5069
#define RSZVL_ATTR_WIMAX_FRAME_SAVE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 5070
#define RSZVL_ATTR_WIMAX_SMU_ADDRESS                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 5071
#define RSZVL_ATTR_WIMAX_STAT_BSTR                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5072
#define RSZVL_ATTR_WIMAX_BSTR_SEL                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5073
#define RSZVL_ATTR_WIMAX_BSTR_SEL_NR                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 5074
#define RSZVL_ATTR_WIMAX_STAT_CCDF                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5075
#define RSZVL_ATTR_WIMAX_STAT_BSUM                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5076
#define RSZVL_ATTR_WIMAX_PVT                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 5077
#define RSZVL_ATTR_WIMAX_BURS_PVT_BURS                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 5078
#define RSZVL_ATTR_WIMAX_PVT_SEL                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5079
#define RSZVL_ATTR_WIMAX_CONS_SEL                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5080
#define RSZVL_ATTR_WIMAX_CONS_SEL_NR                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 5081
#define RSZVL_ATTR_WIMAX_CONS_CCAR                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5082
#define RSZVL_ATTR_WIMAX_CONS_CSYM                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5083
#define RSZVL_ATTR_WIMAX_PRE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 5084
#define RSZVL_ATTR_WIMAX_PRE_SEL                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5085
#define RSZVL_ATTR_WIMAX_UNIT_PRE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5086
#define RSZVL_ATTR_WIMAX_EVM_ECAR                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5087
#define RSZVL_ATTR_WIMAX_EVM_ESYM                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5088
#define RSZVL_ATTR_WIMAX_UNIT_EVM                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5089
#define RSZVL_ATTR_WIMAX_SPEC_ACPR_SEL                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 5090
#define RSZVL_ATTR_WIMAX_ACH_MODE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5091
#define RSZVL_ATTR_WIMAX_ACH_ACP                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5092
#define RSZVL_ATTR_WIMAX_ACH_BW_ACH                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 5093
#define RSZVL_ATTR_WIMAX_ACH_BW_ALT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 5094
#define RSZVL_ATTR_WIMAX_ACH_BW_CHAN                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 5095
#define RSZVL_ATTR_WIMAX_ACH_SPAC_ACH                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 5096
#define RSZVL_ATTR_WIMAX_ACH_SPAC_ALT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 5097
#define RSZVL_ATTR_WIMAX_ACP_NOISE_CORR                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 5098
#define RSZVL_ATTR_WIMAX_SPEC_ACPR                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5099
#define RSZVL_ATTR_WIMAX_SPEC_FFT                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5100
#define RSZVL_ATTR_WIMAX_SPEC_FLAT                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5101
#define RSZVL_ATTR_WIMAX_SPEC_FLAT_SEL                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 5102
#define RSZVL_ATTR_WIMAX_SPEC_MASK                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5103
#define RSZVL_ATTR_WIMAX_SPEC_MASK_SEL                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 5104
#define RSZVL_ATTR_WIMAX_ACPR_TIME                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5105
#define RSZVL_ATTR_WIMAX_ACPR_TIME_AUTO                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 5106
#define RSZVL_ATTR_WIMAX_SEM_CLASS                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5107
#define RSZVL_ATTR_WIMAX_ESP_ANALYSIS                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 5108
#define RSZVL_ATTR_WIMAX_ESP_TTA                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5109
#define RSZVL_ATTR_WIMAX_DISP_AUTO_Y                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 5110
#define RSZVL_ATTR_WIMAX_DISP_TABL                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5111
#define RSZVL_ATTR_WIMAX_SCREEN_SELECT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 5112
#define RSZVL_ATTR_WIMAX_DISP_FORMAT                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 5113
#define RSZVL_ATTR_WIMAX_Y_PDIV                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 5114
#define RSZVL_ATTR_WIMAX_EVM_ALL                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5115
#define RSZVL_ATTR_WIMAX_EVM_DATA                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5116
#define RSZVL_ATTR_WIMAX_FERR                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 5117
#define RSZVL_ATTR_WIMAX_SST                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 5118
#define RSZVL_ATTR_WIMAX_SYMB                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 5119
#define RSZVL_ATTR_WIMAX_IQOF                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 5120
#define RSZVL_ATTR_WIMAX_BERP                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 5121
#define RSZVL_ATTR_WIMAX_EVM_ALL_RES                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 5122
#define RSZVL_ATTR_WIMAX_EVM_DATA_RES                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 5123
#define RSZVL_ATTR_WIMAX_FERR_RES                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5124
#define RSZVL_ATTR_WIMAX_LIM_FAIL_RES                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 5125
#define RSZVL_ATTR_WIMAX_SPEC_MASK_X                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 5126
#define RSZVL_ATTR_WIMAX_SPEC_MASK_Y                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 5127
#define RSZVL_ATTR_WIMAX_SST_RES                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5128
#define RSZVL_ATTR_WIMAX_SYMB_RES                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5129
#define RSZVL_ATTR_WIMAX_IQOF_RES                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5130
#define RSZVL_ATTR_WIMAX_BERP_RES                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5131
#define RSZVL_ATTR_WIMAX_MARKER_ENABLED                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 5132
#define RSZVL_ATTR_WIMAX_MARKER_AOFF                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 5133
#define RSZVL_ATTR_WIMAX_MARKER_TRACE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 5134
#define RSZVL_ATTR_WIMAX_MARK_BURST                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 5135
#define RSZVL_ATTR_WIMAX_MARK_CARR                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5136
#define RSZVL_ATTR_WIMAX_MARK_SYMB                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5137
#define RSZVL_ATTR_WIMAX_MARK_Y                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 5138
#define RSZVL_ATTR_WIMAX_MARKER_POSITION                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 5139
#define RSZVL_ATTR_WIMAX_MARKER_ZOOM                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 5140
#define RSZVL_ATTR_WIMAX_MARKER_SEARCH_MAX                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 5141
#define RSZVL_ATTR_WIMAX_MARKER_SEARCH_MIN                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 5142
#define RSZVL_ATTR_WIMAX_FETC_BURS_COUNT                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 5143
#define RSZVL_ATTR_WIMAX_FETC_CRES                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5144
#define RSZVL_ATTR_WIMAX_FETC_EVM_ALL                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 5145
#define RSZVL_ATTR_WIMAX_FETC_EVM_DATA                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 5146
#define RSZVL_ATTR_WIMAX_FETC_EVM_PILOT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 5147
#define RSZVL_ATTR_WIMAX_FETC_FERR                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5148
#define RSZVL_ATTR_WIMAX_FETC_GIMB                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5149
#define RSZVL_ATTR_WIMAX_FETC_IQOF                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5150
#define RSZVL_ATTR_WIMAX_FETC_QUAD                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5151
#define RSZVL_ATTR_WIMAX_FETC_RMS                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5152
#define RSZVL_ATTR_WIMAX_FETC_SYMB                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5153
#define RSZVL_ATTR_WIMAX_FETC_SYMB_COUNT                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 5154
#define RSZVL_ATTR_WIMAX_FETC_RMS_ALL                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 5155
#define RSZVL_ATTR_WIMAX_FETC_RMS_PILOT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 5156
#define RSZVL_ATTR_WIMAX_FETC_RMS_DATA                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 5157
#define RSZVL_ATTR_WIMAX_FETC_RMS_DLPR                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 5158
#define RSZVL_ATTR_WIMAX_FETC_ZONE_COUNT                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 5159
#define RSZVL_ATTR_WIMAX_FETC_TDOM_PREAMBLE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 5160
#define RSZVL_ATTR_WIMAX_FETC_TDOM_SUBFRAME                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 5161
#define RSZVL_ATTR_WIMAX_FETC_TDOM_ZONE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 5162
#define RSZVL_ATTR_WIMAX_EGAT                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 5163
#define RSZVL_ATTR_WIMAX_EGAT_HOLD_SAMP                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 5164
#define RSZVL_ATTR_WIMAX_EGAT_HOLD_TIME                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 5165
#define RSZVL_ATTR_WIMAX_EGAT_LENG_SAMP                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 5166
#define RSZVL_ATTR_WIMAX_EGAT_LENG_TIME                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 5167
#define RSZVL_ATTR_WIMAX_EGAT_LINK                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5168
#define RSZVL_ATTR_WIMAX_TRIG_MODE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5169
#define RSZVL_ATTR_WIMAX_TRIG_EXT_LEV                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 5170
#define RSZVL_ATTR_WIMAX_TRIG_POW_LEV                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 5171
#define RSZVL_ATTR_WIMAX_TRIG_POW_LEV_AUTO                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 5172
#define RSZVL_ATTR_WIMAX_BAN                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 5173
#define RSZVL_ATTR_WIMAX_BCON_AUTO                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5174
#define RSZVL_ATTR_WIMAX_CEST                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 5175
#define RSZVL_ATTR_WIMAX_LMODE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 5176
#define RSZVL_ATTR_WIMAX_SUBCHANNEL                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 5177
#define RSZVL_ATTR_WIMAX_SUBCHANNEL_STATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 5178
#define RSZVL_ATTR_WIMAX_SUBCHANNEL_ULPH                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 5179
#define RSZVL_ATTR_WIMAX_SYM_EQU                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5180
#define RSZVL_ATTR_WIMAX_SYM                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 5181
#define RSZVL_ATTR_WIMAX_TRAC_LEV                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5182
#define RSZVL_ATTR_WIMAX_TRAC_PHAS                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5183
#define RSZVL_ATTR_WIMAX_TRAC_TIME                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5184
#define RSZVL_ATTR_WIMAX_TDD_TTG                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5185
#define RSZVL_ATTR_WIMAX_FFT_OFFSET                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 5186
#define RSZVL_ATTR_WIMAX_CEST_OFDMA                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 5187
#define RSZVL_ATTR_WIMAX_TRACK_PILOTS                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 5188
#define RSZVL_ATTR_WIMAX_DL_IDCELL                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5189
#define RSZVL_ATTR_WIMAX_DL_PREAMBLE_MODE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 5190
#define RSZVL_ATTR_WIMAX_DL_PREAMBLE_INDEX                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 5191
#define RSZVL_ATTR_WIMAX_DL_SUBCHANNEL                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 5192
#define RSZVL_ATTR_WIMAX_UL_FRAME                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5193
#define RSZVL_ATTR_WIMAX_ZONE_COUNT_QUERY                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 5194
#define RSZVL_ATTR_WIMAX_ZONE_ANALYZE_STATE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 5195
#define RSZVL_ATTR_WIMAX_ZONE_TYPE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5196
#define RSZVL_ATTR_WIMAX_ZONE_SEGMENT_NUM                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 5197
#define RSZVL_ATTR_WIMAX_ZONE_PERM_BASE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 5198
#define RSZVL_ATTR_WIMAX_ZONE_PRBS                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5199
#define RSZVL_ATTR_WIMAX_ZONE_DELETE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 5200
#define RSZVL_ATTR_WIMAX_ZONE_COUNT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 5201
#define RSZVL_ATTR_WIMAX_ZONE_SYMBOL_COUNT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 5202
#define RSZVL_ATTR_WIMAX_ZONE_SYMBOL_OFFSET                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 5203
#define RSZVL_ATTR_WIMAX_ZONE_BURST_COUNT_QUERY                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 5204
#define RSZVL_ATTR_WIMAX_ZONE_BURST_FORMAT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 5205
#define RSZVL_ATTR_WIMAX_ZONE_BURST_OFFSET_AUTO                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 5206
#define RSZVL_ATTR_WIMAX_ZONE_BURST_SLOT_DURATION                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5207
#define RSZVL_ATTR_WIMAX_ZONE_BURST_SUBC_COUNT                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 5208
#define RSZVL_ATTR_WIMAX_ZONE_BURST_SUBC_OFFSET                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 5209
#define RSZVL_ATTR_WIMAX_ZONE_BURST_SYMBOL_COUNT                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5210
#define RSZVL_ATTR_WIMAX_ZONE_BURST_SYMBOL_OFFSET                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5211
#define RSZVL_ATTR_WIMAX_ZONE_BURST_POWER                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 5212
#define RSZVL_ATTR_WIMAX_ZONE_BURST_TYPE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 5213
#define RSZVL_ATTR_WIMAX_ZONE_BURST_DELETE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 5214
#define RSZVL_ATTR_WIMAX_REFRESH                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 5215
#define RSZVL_ATTR_WIMAX_QUERY_TTC                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 5216
#define RSZVL_ATTR_TRIGGER_SOURCE_TV                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 5217
#define RSZVL_ATTR_TRIGGER_TV_VID_CONT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 5218
#define RSZVL_ATTR_TRIGGER_TV_VERTICAL_SIGNAL                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 5219
#define RSZVL_ATTR_TRIGGER_TV_LINE_SYSTEM                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 5220
#define RSZVL_ATTR_TRIGGER_TV_HORIZONTAL_SIGNAL                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 5221
#define RSZVL_ATTR_TRIGGER_TV_VIDEO_POLARITY                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 5222

// rszvl_rngUnits
#define RSZVL_VAL_UNIT_DBM                                                  0
#define RSZVL_VAL_UNIT_DBPW                                                 1
#define RSZVL_VAL_UNIT_WATT                                                 2
#define RSZVL_VAL_UNIT_DBUV                                                 3
#define RSZVL_VAL_UNIT_DBMV                                                 4
#define RSZVL_VAL_UNIT_VOLT                                                 5
#define RSZVL_VAL_UNIT_DBUA                                                 6
#define RSZVL_VAL_UNIT_AMP                                                  7
#define RSZVL_VAL_UNIT_DB                                                   8
#define RSZVL_VAL_UNIT_DEG                                                  9
#define RSZVL_VAL_UNIT_RAD                                                  10
#define RSZVL_VAL_UNIT_S                                                    11
#define RSZVL_VAL_UNIT_HZ                                                   12
#define RSZVL_VAL_UNIT_PCT                                                  13
#define RSZVL_VAL_UNIT_UNITLESS                                             14
#define RSZVL_VAL_UNIT_DBPT                                                 15

// rszvl_rngTraceMode
#define RSZVL_TRAC_MOD_WRITE                                                0
#define RSZVL_TRAC_MOD_VIEW                                                 1
#define RSZVL_TRAC_MOD_AVER                                                 2
#define RSZVL_TRAC_MOD_MAXH                                                 3
#define RSZVL_TRAC_MOD_MINH                                                 4
#define RSZVL_TRAC_MOD_BLANK                                                5

// rszvl_rngAmplitudeUnitsRangeTable
#define RSZVL_VAL_AMPLITUDE_UNITS_DBM                                       0
#define RSZVL_VAL_AMPLITUDE_UNITS_DBMV                                      1
#define RSZVL_VAL_AMPLITUDE_UNITS_DBUV                                      2
#define RSZVL_VAL_AMPLITUDE_UNITS_VOLT                                      3
#define RSZVL_VAL_AMPLITUDE_UNITS_WATT                                      4
#define RSZVL_VAL_AMPLITUDE_UNITS_DBPW                                      5
#define RSZVL_VAL_AMPLITUDE_UNITS_DBUA                                      6
#define RSZVL_VAL_AMPLITUDE_UNITS_AMP                                       7

// rszvl_rngAbsRel
#define RSZVL_VAL_ABS                                                       0
#define RSZVL_VAL_REL                                                       1

// rszvl_rngESPMode
#define RSZVL_VAL_ESP_MODE_AUTO                                             0
#define RSZVL_VAL_ESP_MODE_MAN                                              1
#define RSZVL_VAL_ESP_MODE_USER                                             2

// rszvl_rngDetectorTypeRangeTable
#define RSZVL_VAL_DETECTOR_TYPE_AUTO_PEAK                                   0
#define RSZVL_VAL_DETECTOR_TYPE_AVERAGE                                     1
#define RSZVL_VAL_DETECTOR_TYPE_MAX_PEAK                                    2
#define RSZVL_VAL_DETECTOR_TYPE_MIN_PEAK                                    3
#define RSZVL_VAL_DETECTOR_TYPE_SAMPLE                                      4
#define RSZVL_VAL_DETECTOR_TYPE_RMS                                         5
#define RSZVL_VAL_DETECTOR_TYPE_QPK                                         6

// rszvl_rngDISPConfPreDefColour
#define RSZVL_VAL_DISP_COL_BLAC                                             0
#define RSZVL_VAL_DISP_COL_BLUE                                             1
#define RSZVL_VAL_DISP_COL_BROW                                             2
#define RSZVL_VAL_DISP_COL_GRE                                              3
#define RSZVL_VAL_DISP_COL_CYAN                                             4
#define RSZVL_VAL_DISP_COL_RED                                              5
#define RSZVL_VAL_DISP_COL_MAG                                              6
#define RSZVL_VAL_DISP_COL_YELL                                             7
#define RSZVL_VAL_DISP_COL_WHIT                                             8
#define RSZVL_VAL_DISP_COL_DGRA                                             9
#define RSZVL_VAL_DISP_COL_LGRA                                             10
#define RSZVL_VAL_DISP_COL_LBLU                                             11
#define RSZVL_VAL_DISP_COL_LGRE                                             12
#define RSZVL_VAL_DISP_COL_LCY                                              13
#define RSZVL_VAL_DISP_COL_LRED                                             14
#define RSZVL_VAL_DISP_COL_LMAG                                             15

// rszvl_rngHcopyDeviceLang
#define RSZVL_VAL_HCOPY_DEVICE_LANG_GDI                                     0
#define RSZVL_VAL_HCOPY_DEVICE_LANG_WMF                                     1
#define RSZVL_VAL_HCOPY_DEVICE_LANG_EWMF                                    2
#define RSZVL_VAL_HCOPY_DEVICE_LANG_BMP                                     3

// rszvl_rngHcopyDeviceOrient
#define RSZVL_VAL_HCOPY_DEVICE_ORIENT_LAND                                  0
#define RSZVL_VAL_HCOPY_DEVICE_ORIENT_PORT                                  1

// rszvl_rngDomain
#define RSZVL_VAL_FREQ                                                      0
#define RSZVL_VAL_TIME                                                      1

// rszvl_rngLinLog
#define RSZVL_VAL_VBW_FILT_LIN                                              0
#define RSZVL_VAL_VBW_FILT_LOG                                              1

// rszvl_rngServiceInput
#define RSZVL_VAL_INPUT_RF                                                  0
#define RSZVL_VAL_INPUT_CAL                                                 1
#define RSZVL_VAL_INPUT_TG                                                  2

// rszvl_rngAvgType
#define RSZVL_VAL_AVG_LIN                                                   0
#define RSZVL_VAL_AVG_VID                                                   1

// rszvl_rngDISPSizeWindow
#define RSZVL_VAL_DISP_SIZE_LARGE                                           0
#define RSZVL_VAL_DISP_SIZE_SMALL                                           1

// rszvl_rngMarkerPeakListSort
#define RSZVL_VAL_MARKER_SORT_X                                             0
#define RSZVL_VAL_MARKER_SORT_Y                                             1

// rszvl_rngMarkerDemodType
#define RSZVL_VAL_MARKER_DEMOD_AM                                           0
#define RSZVL_VAL_MARKER_DEMOD_FM                                           1

// rszvl_rngMeasPowerSelect
#define RSZVL_VAL_MEAS_POW_ACP                                              0
#define RSZVL_VAL_MEAS_POW_CPOW                                             1
#define RSZVL_VAL_MEAS_POW_MCAC                                             2
#define RSZVL_VAL_MEAS_POW_OBAN                                             3
#define RSZVL_VAL_MEAS_POW_CN                                               4
#define RSZVL_VAL_MEAS_POW_CN0                                              5

// rszvl_rngMeasPowerStandard
#define RSZVL_VAL_MEAS_POW_STD_NONE                                         0
#define RSZVL_VAL_MEAS_POW_STD_NADC                                         1
#define RSZVL_VAL_MEAS_POW_STD_TETRA                                        2
#define RSZVL_VAL_MEAS_POW_STD_PDC                                          3
#define RSZVL_VAL_MEAS_POW_STD_PHS                                          4
#define RSZVL_VAL_MEAS_POW_STD_CDPD                                         5
#define RSZVL_VAL_MEAS_POW_STD_FIS95A                                       6
#define RSZVL_VAL_MEAS_POW_STD_RIS95A                                       7
#define RSZVL_VAL_MEAS_POW_STD_FIS95C0                                      8
#define RSZVL_VAL_MEAS_POW_STD_RIS95C0                                      9
#define RSZVL_VAL_MEAS_POW_STD_FJ008                                        10
#define RSZVL_VAL_MEAS_POW_STD_RJ008                                        11
#define RSZVL_VAL_MEAS_POW_STD_FIS95C1                                      12
#define RSZVL_VAL_MEAS_POW_STD_RIS95C1                                      13
#define RSZVL_VAL_MEAS_POW_STD_FWCD                                         14
#define RSZVL_VAL_MEAS_POW_STD_RWCD                                         15
#define RSZVL_VAL_MEAS_POW_STD_FW3G                                         16
#define RSZVL_VAL_MEAS_POW_STD_RW3G                                         17
#define RSZVL_VAL_MEAS_POW_STD_D2CD                                         18
#define RSZVL_VAL_MEAS_POW_STD_S2CD                                         19
#define RSZVL_VAL_MEAS_POW_STD_M2CD                                         20
#define RSZVL_VAL_MEAS_POW_STD_TCDM                                         21
#define RSZVL_VAL_MEAS_POW_STD_AWL                                          22
#define RSZVL_VAL_MEAS_POW_STD_BWL                                          23
#define RSZVL_VAL_MEAS_POW_STD_WIM                                          24
#define RSZVL_VAL_MEAS_POW_STD_WIBRO                                        25

// rszvl_rngMeasPowerMode
#define RSZVL_VAL_MEAS_POW_MODE_WRITE                                       0
#define RSZVL_VAL_MEAS_POW_MODE_MAXH                                        1

// rszvl_rngBandFiltType
#define RSZVL_VAL_FILT_TYPE_NORM                                            0
#define RSZVL_VAL_FILT_TYPE_FFT                                             1
#define RSZVL_VAL_FILT_TYPE_CFIL                                            2
#define RSZVL_VAL_FILT_TYPE_RRC                                             3
#define RSZVL_VAL_FILT_TYPE_PULSE                                           4

// rszvl_rngTFacUnit
#define RSZVL_VAL_TFAC_UNIT_DB                                              0
#define RSZVL_VAL_TFAC_UNIT_DBM                                             1
#define RSZVL_VAL_TFAC_UNIT_DBMV                                            2
#define RSZVL_VAL_TFAC_UNIT_DBUV                                            3
#define RSZVL_VAL_TFAC_UNIT_DBUVM                                           4
#define RSZVL_VAL_TFAC_UNIT_DBUA                                            5
#define RSZVL_VAL_TFAC_UNIT_DBUAM                                           6
#define RSZVL_VAL_TFAC_UNIT_DBPW                                            7
#define RSZVL_VAL_TFAC_UNIT_DBPT                                            8

// rszvl_rngFreqCentLink
#define RSZVL_VAL_CENT_FREQ_LINK_OFF                                        0
#define RSZVL_VAL_CENT_FREQ_LINK_SPAN                                       1
#define RSZVL_VAL_CENT_FREQ_LINK_RBW                                        2

// rszvl_rngFreqMode
#define RSZVL_VAL_FREQ_MODE_CW                                              0
#define RSZVL_VAL_FREQ_MODE_FIX                                             1
#define RSZVL_VAL_FREQ_MODE_SWE                                             2

// rszvl_rngOffOnAuto
#define RSZVL_VAL_STATE_OFF                                                 0
#define RSZVL_VAL_STATE_ON                                                  1
#define RSZVL_VAL_STATE_AUTO                                                2

// rszvl_rngExtMixerHarmonicType
#define RSZVL_VAL_EXT_MIX_HARM_ODD                                          0
#define RSZVL_VAL_EXT_MIX_HARM_EVEN                                         1
#define RSZVL_VAL_EXT_MIX_HARM_EODD                                         2

// rszvl_rngExtMixerHarmBand
#define RSZVL_VAL_EXT_MIX_BAND_A                                            0
#define RSZVL_VAL_EXT_MIX_BAND_Q                                            1
#define RSZVL_VAL_EXT_MIX_BAND_U                                            2
#define RSZVL_VAL_EXT_MIX_BAND_V                                            3
#define RSZVL_VAL_EXT_MIX_BAND_E                                            4
#define RSZVL_VAL_EXT_MIX_BAND_W                                            5
#define RSZVL_VAL_EXT_MIX_BAND_F                                            6
#define RSZVL_VAL_EXT_MIX_BAND_D                                            7
#define RSZVL_VAL_EXT_MIX_BAND_G                                            8
#define RSZVL_VAL_EXT_MIX_BAND_Y                                            9
#define RSZVL_VAL_EXT_MIX_BAND_J                                            10
#define RSZVL_VAL_EXT_MIX_BAND_USER                                         11

// rszvl_rngAdjChAutoSel
#define RSZVL_VAL_ADJ_RCHAN_AUTO_MIN                                        0
#define RSZVL_VAL_ADJ_RCHAN_AUTO_MAX                                        1
#define RSZVL_VAL_ADJ_RCHAN_AUTO_LHIG                                       2

// rszvl_rngAdjPreset
#define RSZVL_VAL_ADJ_PRE_ACP                                               0
#define RSZVL_VAL_ADJ_PRE_CPOW                                              1
#define RSZVL_VAL_ADJ_PRE_MCAC                                              2
#define RSZVL_VAL_ADJ_PRE_OBAN                                              3
#define RSZVL_VAL_ADJ_PRE_CN                                                4
#define RSZVL_VAL_ADJ_PRE_CN0                                               5

// rszvl_rngSourceIntExt
#define RSZVL_VAL_SOUR_INT                                                  0
#define RSZVL_VAL_SOUR_EXT                                                  1

// rszvl_rngExtGateTrigType
#define RSZVL_VAL_EGAT_TRIG_LEV                                             0
#define RSZVL_VAL_EGAT_TRIG_EDGE                                            1

// rszvl_rngHcopyDevice
#define RSZVL_VAL_HCOPY_DEVICE_MEM                                          0
#define RSZVL_VAL_HCOPY_DEVICE_PRN                                          1
#define RSZVL_VAL_HCOPY_DEVICE_CLP                                          2

// rszvl_rngPolarity
#define RSZVL_VAL_NEG                                                       0
#define RSZVL_VAL_POS                                                       1

// rszvl_rngExtGateSource
#define RSZVL_VAL_EGAT_SOUR_EXT                                             0
#define RSZVL_VAL_EGAT_SOUR_IFP                                             1

// rszvl_rngSweepContMode
#define RSZVL_VAL_SWE_MODE_AUTO                                             0
#define RSZVL_VAL_SWE_MODE_LIST                                             1
#define RSZVL_VAL_SWE_MODE_ESYNC                                            2

// rszvl_rngTriggerSource
#define RSZVL_VAL_TRG_IMM                                                   0
#define RSZVL_VAL_TRG_EXT                                                   1
#define RSZVL_VAL_TRG_IFP                                                   2
#define RSZVL_VAL_TRG_VID                                                   3
#define RSZVL_VAL_TRG_TIME                                                  4

// rszvl_rngFileDecSeparator
#define RSZVL_VAL_DEC_SEP_POIN                                              0
#define RSZVL_VAL_DEC_SEP_COMMA                                             1

// rszvl_rngMarkerStepSize
#define RSZVL_VAL_MARK_STEP_SIZE_STAN                                       0
#define RSZVL_VAL_MARK_STEP_SIZE_POIN                                       1

// rszvl_rngMpowFiltType
#define RSZVL_VAL_MPOW_FTYPE_NORM                                           0
#define RSZVL_VAL_MPOW_FTYPE_CFIL                                           1
#define RSZVL_VAL_MPOW_FTYPE_RRC                                            2

// rszvl_rngServiceCombFrequency
#define RSZVL_VAL_SERV_COMB_FREQ_COMB1                                      0
#define RSZVL_VAL_SERV_COMB_FREQ_COMB64                                     1
#define RSZVL_VAL_SERV_COMB_FREQ_COMB65                                     2

// rszvl_rngTraceSpacing
#define RSZVL_VAL_TRACE_SPACE_LIN                                           0
#define RSZVL_VAL_TRACE_SPACE_LOG                                           1
#define RSZVL_VAL_TRACE_SPACE_PCT                                           2
#define RSZVL_VAL_TRACE_SPACE_LDB                                           3

// rszvl_rngNWASweepType
#define RSZVL_VAL_SWE_TYPE_LIN                                              0
#define RSZVL_VAL_SWE_TYPE_LOG                                              1
#define RSZVL_VAL_SWE_TYPE_SEG                                              2

// rszvl_rngNWASegmentSelectSweepTime
#define RSZVL_VAL_SEG_SWE_TIME_SWT                                          0
#define RSZVL_VAL_SEG_SWE_TIME_DWEL                                         1

// rszvl_rngNWASegmentSelectivity
#define RSZVL_VAL_SEG_SELECTIVITY_NORMAL                                    0
#define RSZVL_VAL_SEG_SELECTIVITY_FAST                                      1

// rszvl_rngNWAConnectorType
#define RSZVL_VAL_CONN_N50FEMALE                                            0
#define RSZVL_VAL_CONN_N50MALE                                              1
#define RSZVL_VAL_CONN_N75FEMALE                                            2
#define RSZVL_VAL_CONN_N75MALE                                              3
#define RSZVL_VAL_CONN_PC7                                                  4
#define RSZVL_VAL_CONN_PC35FEMALE                                           5
#define RSZVL_VAL_CONN_PC35MALE                                             6
#define RSZVL_VAL_CONN_PC292FEMALE                                          7
#define RSZVL_VAL_CONN_PC292MALE                                            8
#define RSZVL_VAL_CONN_UFEMALE1                                             9
#define RSZVL_VAL_CONN_UMALE1                                               10
#define RSZVL_VAL_CONN_UFEMALE2                                             11
#define RSZVL_VAL_CONN_UMALE2                                               12

// rszvl_rngNWASameConnectorsAtPorts
#define RSZVL_VAL_CONN_PORTS_SINGLE                                         0
#define RSZVL_VAL_CONN_PORTS_ALL                                            1

// rszvl_rngNWATraceFormat
#define RSZVL_VAL_TRAC_FORM_MLIN                                            0
#define RSZVL_VAL_TRAC_FORM_MLOG                                            1
#define RSZVL_VAL_TRAC_FORM_PHAS                                            2
#define RSZVL_VAL_TRAC_FORM_UPH                                             3
#define RSZVL_VAL_TRAC_FORM_POL                                             4
#define RSZVL_VAL_TRAC_FORM_SMIT                                            5
#define RSZVL_VAL_TRAC_FORM_ISM                                             6
#define RSZVL_VAL_TRAC_FORM_GDEL                                            7
#define RSZVL_VAL_TRAC_FORM_REAL                                            8
#define RSZVL_VAL_TRAC_FORM_IMAG                                            9
#define RSZVL_VAL_TRAC_FORM_SWR                                             10

// rszvl_rngNWATraceMathFunction
#define RSZVL_VAL_TRAC_MATH_FUNC_NORM                                       0
#define RSZVL_VAL_TRAC_MATH_FUNC_SUBT                                       1
#define RSZVL_VAL_TRAC_MATH_FUNC_DIV                                        2

// rszvl_rngNWATraceMaxHold
#define RSZVL_VAL_TRAC_PHOL_OFF                                             0
#define RSZVL_VAL_TRAC_PHOL_MAX                                             1

// rszvl_rngNWADeleteMemTrace
#define RSZVL_VAL_TRAC_DELETE_MDAT1                                         0
#define RSZVL_VAL_TRAC_DELETE_MDAT2                                         1
#define RSZVL_VAL_TRAC_DELETE_MDAT3                                         2
#define RSZVL_VAL_TRAC_DELETE_MDAT4                                         3
#define RSZVL_VAL_TRAC_DELETE_MDAT5                                         4
#define RSZVL_VAL_TRAC_DELETE_MDAT6                                         5
#define RSZVL_VAL_TRAC_DELETE_MDAT7                                         6
#define RSZVL_VAL_TRAC_DELETE_MDAT8                                         7

// rszvl_rngNWATraceEvalRange
#define RSZVL_VAL_TRAC_EVAL_FULL_SPAN                                       0
#define RSZVL_VAL_TRAC_EVAL_1                                               1
#define RSZVL_VAL_TRAC_EVAL_2                                               2
#define RSZVL_VAL_TRAC_EVAL_3                                               3
#define RSZVL_VAL_TRAC_EVAL_4                                               4
#define RSZVL_VAL_TRAC_EVAL_5                                               5
#define RSZVL_VAL_TRAC_EVAL_6                                               6
#define RSZVL_VAL_TRAC_EVAL_7                                               7
#define RSZVL_VAL_TRAC_EVAL_8                                               8
#define RSZVL_VAL_TRAC_EVAL_9                                               9
#define RSZVL_VAL_TRAC_EVAL_10                                              10

// rszvl_rngNWAMarkerMode
#define RSZVL_VAL_MARK_MODE_CONT                                            0
#define RSZVL_VAL_MARK_MODE_DISC                                            1

// rszvl_rngNWAMarkerType
#define RSZVL_VAL_MARKER_TYPE_NORM                                          0
#define RSZVL_VAL_MARKER_TYPE_FIX                                           1

// rszvl_rngNWAMarkerFormat
#define RSZVL_VAL_MARK_FORM_MLIN                                            0
#define RSZVL_VAL_MARK_FORM_MLOG                                            1
#define RSZVL_VAL_MARK_FORM_PHAS                                            2
#define RSZVL_VAL_MARK_FORM_POL                                             3
#define RSZVL_VAL_MARK_FORM_GDEL                                            4
#define RSZVL_VAL_MARK_FORM_REAL                                            5
#define RSZVL_VAL_MARK_FORM_IMAG                                            6
#define RSZVL_VAL_MARK_FORM_SWR                                             7
#define RSZVL_VAL_MARK_FORM_LINP                                            8
#define RSZVL_VAL_MARK_FORM_LOGP                                            9
#define RSZVL_VAL_MARK_FORM_IMP                                             10
#define RSZVL_VAL_MARK_FORM_ADM                                             11

// rszvl_rngNWAMarkerSearch
#define RSZVL_VAL_MARK_SEARCH_MAX                                           0
#define RSZVL_VAL_MARK_SEARCH_MIN                                           1
#define RSZVL_VAL_MARK_SEARCH_RPE                                           2
#define RSZVL_VAL_MARK_SEARCH_LPE                                           3
#define RSZVL_VAL_MARK_SEARCH_NPE                                           4
#define RSZVL_VAL_MARK_SEARCH_TARG                                          5
#define RSZVL_VAL_MARK_SEARCH_LTAR                                          6
#define RSZVL_VAL_MARK_SEARCH_RTAR                                          7
#define RSZVL_VAL_MARK_SEARCH_BFIL                                          8

// rszvl_rngNWAMarkerBFilterSearchMode
#define RSZVL_VAL_MARK_BFIL_SEARCH_BPAS                                     0
#define RSZVL_VAL_MARK_BFIL_SEARCH_BST                                      1
#define RSZVL_VAL_MARK_BFIL_SEARCH_BPRM                                     2
#define RSZVL_VAL_MARK_BFIL_SEARCH_BSRM                                     3

// rszvl_rngNWAMarkerSearchRange
#define RSZVL_VAL_MARK_FULL_SPAN                                            0
#define RSZVL_VAL_MARK_RANGE_1                                              1
#define RSZVL_VAL_MARK_RANGE_2                                              2
#define RSZVL_VAL_MARK_RANGE_3                                              3
#define RSZVL_VAL_MARK_RANGE_4                                              4
#define RSZVL_VAL_MARK_RANGE_5                                              5
#define RSZVL_VAL_MARK_RANGE_6                                              6
#define RSZVL_VAL_MARK_RANGE_7                                              7
#define RSZVL_VAL_MARK_RANGE_8                                              8
#define RSZVL_VAL_MARK_RANGE_9                                              9
#define RSZVL_VAL_MARK_RANGE_10                                             10

// rszvl_rngNWALimitRDOMComplex
#define RSZVL_VAL_LIM_RDOM_COMP_S                                           0
#define RSZVL_VAL_LIM_RDOM_COMP_SINV                                        1
#define RSZVL_VAL_LIM_RDOM_COMP_Y                                           2
#define RSZVL_VAL_LIM_RDOM_COMP_Z                                           3
#define RSZVL_VAL_LIM_RDOM_COMP_YREL                                        4
#define RSZVL_VAL_LIM_RDOM_COMP_ZREL                                        5

// rszvl_rngNWALimitRDOMFormat
#define RSZVL_VAL_LIM_RDOM_FORM_COMP                                        0
#define RSZVL_VAL_LIM_RDOM_FORM_MAGN                                        1
#define RSZVL_VAL_LIM_RDOM_FORM_PHAS                                        2
#define RSZVL_VAL_LIM_RDOM_FORM_REAL                                        3
#define RSZVL_VAL_LIM_RDOM_FORM_IMAG                                        4
#define RSZVL_VAL_LIM_RDOM_FORM_SWR                                         5
#define RSZVL_VAL_LIM_RDOM_FORM_GDEL                                        6
#define RSZVL_VAL_LIM_RDOM_FORM_L                                           7
#define RSZVL_VAL_LIM_RDOM_FORM_C                                           8

// rszvl_rngNWALimitRDOMSpacing
#define RSZVL_VAL_LIM_RDOM_SPAC_LIN                                         0
#define RSZVL_VAL_LIM_RDOM_SPAC_LOG                                         1
#define RSZVL_VAL_LIM_RDOM_SPAC_DB                                          2
#define RSZVL_VAL_LIM_RDOM_SPAC_SIC                                         3

// rszvl_rngNWALimitSegmType
#define RSZVL_VAL_LIMIT_OFF                                                 0
#define RSZVL_VAL_LIMIT_UPPER                                               1
#define RSZVL_VAL_LIMIT_LOWER                                               2

// rszvl_rngNWAParamConversion
#define RSZVL_VAL_CONV_S                                                    0
#define RSZVL_VAL_CONV_Y                                                    1
#define RSZVL_VAL_CONV_Z                                                    2

// rszvl_rngNWAColorScheme
#define RSZVL_VAL_COLOR_DBACKGROUND                                         0
#define RSZVL_VAL_COLOR_LBACKGROUND                                         1
#define RSZVL_VAL_COLOR_BW_LSTYLES                                          2
#define RSZVL_VAL_COLOR_BW_SOLID                                            3

// rszvl_rngNWAPresetScope
#define RSZVL_VAL_PRESET_SINGLE                                             0
#define RSZVL_VAL_PRESET_ALL                                                1

// rszvl_rngInstrMode
#define RSZVL_VAL_SAN                                                       0
#define RSZVL_VAL_NWA                                                       1
#define RSZVL_VAL_NOISE                                                     2
#define RSZVL_VAL_BWCD                                                      3
#define RSZVL_VAL_WLAN                                                      4
#define RSZVL_VAL_WIMAX                                                     5
#define RSZVL_VAL_ADEM                                                      6
#define RSZVL_VAL_BTO                                                       7
#define RSZVL_VAL_CATV                                                      8

// rszvl_rngHcopyPageSettings
#define RSZVL_VAL_HCOPY_ACT                                                 0
#define RSZVL_VAL_HCOPY_ALL                                                 1
#define RSZVL_VAL_HCOPY_SING                                                2

// rszvl_rngNWAGetCalState
#define RSZVL_CALSTATE_CAL                                                  0
#define RSZVL_CALSTATE_CAI                                                  1
#define RSZVL_CALSTATE_CA                                                   2
#define RSZVL_CALSTATE_CAV                                                  3
#define RSZVL_CALSTATE_CALOFF                                               4

// rszvl_rngFmTriggerSource
#define RSZVL_VAL_TRG_FM                                                    3
#define RSZVL_VAL_TRG_AM                                                    4
#define RSZVL_VAL_TRG_PM                                                    5
#define RSZVL_VAL_TRG_AMR                                                   6

// rszvl_rngCoupling
#define RSZVL_VAL_DC                                                        0
#define RSZVL_VAL_AC                                                        1

// rszvl_rngAdemPhaseWrap
#define RSZVL_VAL_ADEM_UPHAS                                                0
#define RSZVL_VAL_ADEM_PHAS                                                 1

// rszvl_rngUnitAngle
#define RSZVL_VAL_UNIT_ANGLE_DEG                                            0
#define RSZVL_VAL_UNIT_ANGLE_RAD                                            1

// rszvl_rngBtoWritePtype
#define RSZVL_VAL_BTO_PTYPE_AUTO                                            0
#define RSZVL_VAL_BTO_PTYPE_DH1                                             1
#define RSZVL_VAL_BTO_PTYPE_DH3                                             2
#define RSZVL_VAL_BTO_PTYPE_DH5                                             3

// rszvl_rngBtoMeasFilt
#define RSZVL_BTO_MEASFILT_OFF                                              0
#define RSZVL_BTO_MEASFILT_BTO                                              1

// rszvl_rngBtoTraceDetector
#define RSZVL_BTO_TRAC_DET_APEAK                                            0
#define RSZVL_BTO_TRAC_DET_NEG                                              1
#define RSZVL_BTO_TRAC_DET_POS                                              2
#define RSZVL_BTO_TRAC_DET_SAMPLE                                           3
#define RSZVL_BTO_TRAC_DET_RMS                                              4
#define RSZVL_BTO_TRAC_DET_AVER                                             5

// rszvl_rngBtoMeasMode
#define RSZVL_BTO_MEAS_MODE_OPOW                                            0
#define RSZVL_BTO_MEAS_MODE_ACLR                                            1
#define RSZVL_BTO_MEAS_MODE_MCH                                             2
#define RSZVL_BTO_MEAS_MODE_ICFT                                            3
#define RSZVL_BTO_MEAS_MODE_CFDR                                            4
#define RSZVL_BTO_MEAS_MODE_RTP                                             5
#define RSZVL_BTO_MEAS_MODE_IBS                                             6
#define RSZVL_BTO_MEAS_MODE_DPEN                                            7
#define RSZVL_BTO_MEAS_MODE_CFST                                            8

// rszvl_rngBtoPtype
#define RSZVL_BTO_PTYPE_DM1                                                 4
#define RSZVL_BTO_PTYPE_DM3                                                 5
#define RSZVL_BTO_PTYPE_DM5                                                 6
#define RSZVL_BTO_PTYPE_FHS                                                 7
#define RSZVL_BTO_PTYPE_HV1                                                 8
#define RSZVL_BTO_PTYPE_HV2                                                 9
#define RSZVL_BTO_PTYPE_HV3                                                 10
#define RSZVL_BTO_PTYPE_DV                                                  11
#define RSZVL_BTO_PTYPE_NULL                                                12
#define RSZVL_BTO_PTYPE_POLL                                                13
#define RSZVL_BTO_PTYPE_AUX1                                                14
#define RSZVL_BTO_PTYPE_UNDEF                                               15

// rszvl_rngSpemColor
#define RSZVL_VAL_SPEM_COL_COLOR                                            0
#define RSZVL_VAL_SPEM_COL_RADAR                                            1
#define RSZVL_VAL_SPEM_COL_GRAY                                             2

// rszvl_rngSpemSearchArea
#define RSZVL_VAL_SPEM_SAR_MEMORY                                           1
#define RSZVL_VAL_SPEM_SAR_VISIBLE                                          0

// rszvl_rngNoiseCalcLimLine
#define RSZVL_VAL_NOISE_CALC_LIM_LINE_NFIGURE                               0
#define RSZVL_VAL_NOISE_CALC_LIM_LINE_TEFFECTIVE                            1
#define RSZVL_VAL_NOISE_CALC_LIM_LINE_GAIN                                  2

// rszvl_rngNoiseERNModeSel
#define RSZVL_VAL_NOISE_ERN_MODE_SEL_TABLE                                  0
#define RSZVL_VAL_NOISE_ERN_MODE_SEL_SPOT                                   1

// rszvl_rngNoiseLossMode
#define RSZVL_VAL_NOISE_LOSS_MODE_TABLE                                     0
#define RSZVL_VAL_NOISE_LOSS_MODE_SPOT                                      1

// rszvl_rngNoiseConfDUT
#define RSZVL_VAL_NOISE_CONF_DUT_AMPLIFIER                                  0
#define RSZVL_VAL_NOISE_CONF_DUT_DOWNCONV                                   1
#define RSZVL_VAL_NOISE_CONF_DUT_UPCONV                                     2

// rszvl_rngNoiseDispDataTrac
#define RSZVL_VAL_NOISE_DISP_DATA_TEFF                                      1
#define RSZVL_VAL_NOISE_DISP_DATA_NFIGURE                                   2

// rszvl_rngNoiseXAxisFrequency
#define RSZVL_VAL_NOISE_XAXIS_IF                                            0
#define RSZVL_VAL_NOISE_XAXIS_RF                                            1

// rszvl_rng3GBSMeas
#define RSZVL_VAL_3GPPBS_MEAS_POW                                           0
#define RSZVL_VAL_3GPPBS_MEAS_ACLR                                          1
#define RSZVL_VAL_3GPPBS_MEAS_ESP                                           2
#define RSZVL_VAL_3GPPBS_MEAS_WCDP                                          3

// rszvl_rng3GBSSyncType
#define RSZVL_VAL_3GPP_BS_SETTINGS_SYNC_TYPE_CPIC                           0
#define RSZVL_VAL_3GPP_BS_SETTINGS_SYNC_TYPE_SCH                            1

// rszvl_rng3GBSPowRef
#define RSZVL_VAL_3GPP_BS_SETTINGS_POW_REF_TOT                              0
#define RSZVL_VAL_3GPP_BS_SETTINGS_POW_REF_CPIC                             1

// rszvl_rng3GBSAntenna
#define RSZVL_VAL_3GPP_BS_SETTINGS_ANTENNA_OFF                              0
#define RSZVL_VAL_3GPP_BS_SETTINGS_ANTENNA_1                                1
#define RSZVL_VAL_3GPP_BS_SETTINGS_ANTENNA_2                                2

// rszvl_rng3GBSFeed
#define RSZVL_VAL_3GBS_RES_CDP                                              0
#define RSZVL_VAL_3GBS_RES_CDPABS                                           1
#define RSZVL_VAL_3GBS_RES_SUM                                              2
#define RSZVL_VAL_3GBS_RES_CTAB                                             3

// rszvl_rng3GPPResultSumView
#define RSZVL_VAL_3GPP_RES_SUM_NORMAL                                       0
#define RSZVL_VAL_3GPP_RES_SUM_EXTENDED                                     1

// rszvl_rngWlanPvtRPow
#define RSZVL_VAL_PVT_MEAN                                                  0
#define RSZVL_VAL_PVT_MAX                                                   1

// rszvl_rngWlanPvtSel
#define RSZVL_VAL_PVT_EDGE                                                  0
#define RSZVL_VAL_PVT_FULL                                                  1
#define RSZVL_VAL_PVT_RISE                                                  2
#define RSZVL_VAL_PVT_FALL                                                  3

// rszvl_rngWlanBurstType
#define RSZVL_VAL_WLAN_BTYP_DIRECT                                          0
#define RSZVL_VAL_WLAN_BTYP_LOFDM                                           1
#define RSZVL_VAL_WLAN_BTYP_SOFDM                                           2
#define RSZVL_VAL_WLAN_BTYP_LONG                                            3
#define RSZVL_VAL_WLAN_BTYP_SHORT                                           4

// rszvl_rngSpecMask
#define RSZVL_VAL_SPEC_IEEE                                                 0
#define RSZVL_VAL_SPEC_ETSI                                                 1

// rszvl_rngBurstSel
#define RSZVL_VAL_SELECTION_ALL                                             0
#define RSZVL_VAL_SELECTION_PIL                                             1

// rszvl_rngPreamble
#define RSZVL_VAL_PRE_FREQ                                                  0
#define RSZVL_VAL_PRE_PHAS                                                  1

// rszvl_rngResults
#define RSZVL_VAL_RES_FAIL                                                  0
#define RSZVL_VAL_RES_PASS                                                  1

// rszvl_rngTrigMode
#define RSZVL_VAL_TRIG_MODE_IMM                                             0
#define RSZVL_VAL_TRIG_MODE_EXT                                             1
#define RSZVL_VAL_TRIG_MODE_POW                                             2

// rszvl_rngWlanDemod
#define RSZVL_VAL_WLAN_BPSK                                                 0
#define RSZVL_VAL_WLAN_QPSK                                                 1
#define RSZVL_VAL_WLAN_QAM16                                                2
#define RSZVL_VAL_WLAN_QAM64                                                3
#define RSZVL_VAL_WLAN_BPSK6                                                4
#define RSZVL_VAL_WLAN_BPSK9                                                5
#define RSZVL_VAL_WLAN_QPSK12                                               6
#define RSZVL_VAL_WLAN_QPSK18                                               7
#define RSZVL_VAL_WLAN_QAM1624                                              8
#define RSZVL_VAL_WLAN_QAM1636                                              9
#define RSZVL_VAL_WLAN_QAM6448                                              10
#define RSZVL_VAL_WLAN_QAM6454                                              11
#define RSZVL_VAL_WLAN_DBPSK                                                12
#define RSZVL_VAL_WLAN_DQPSK                                                13
#define RSZVL_VAL_WLAN_CCK55                                                14
#define RSZVL_VAL_WLAN_CCK11                                                15
#define RSZVL_VAL_WLAN_PBCC55                                               16
#define RSZVL_VAL_WLAN_PBCC11                                               17
#define RSZVL_VAL_WLAN_PBCC22                                               18
#define RSZVL_VAL_WLAN_BPSK3                                                19
#define RSZVL_VAL_WLAN_BPSK45                                               20
#define RSZVL_VAL_WLAN_QPSK6                                                21
#define RSZVL_VAL_WLAN_QPSK9                                                22
#define RSZVL_VAL_WLAN_QAM1612                                              23
#define RSZVL_VAL_WLAN_QAM1618                                              24
#define RSZVL_VAL_WLAN_QAM6424                                              25
#define RSZVL_VAL_WLAN_QAM6427                                              26

// rszvl_rngWiMAXFreqBand
#define RSZVL_VAL_WIMAX_FBAN_UNSP                                           0
#define RSZVL_VAL_WIMAX_FBAN_ETSI                                           1
#define RSZVL_VAL_WIMAX_FBAN_MMDS                                           2
#define RSZVL_VAL_WIMAX_FBAN_WCS                                            3
#define RSZVL_VAL_WIMAX_FBAN_CEPT                                           4
#define RSZVL_VAL_WIMAX_FBAN_UNI                                            5
#define RSZVL_VAL_WIMAX_FBAN_ETSI2                                          6
#define RSZVL_VAL_WIMAX_FBAN_MMDS2                                          7
#define RSZVL_VAL_WIMAX_FBAN_WCS2                                           8
#define RSZVL_VAL_WIMAX_FBAN_CEPT2                                          9
#define RSZVL_VAL_WIMAX_FBAN_UNI2                                           10

// rszvl_rngWiMAXModulForm
#define RSZVL_VAL_WIMAX_MODUL_ALL                                           0
#define RSZVL_VAL_WIMAX_MODUL_QPSK                                          1
#define RSZVL_VAL_WIMAX_MODUL_QAM16                                         2
#define RSZVL_VAL_WIMAX_MODUL_QAM64                                         3
#define RSZVL_VAL_WIMAX_MODUL_BPSK                                          4

// rszvl_rngWiMAXFlatSel
#define RSZVL_VAL_WIMAX_FLAT_FLAT                                           0
#define RSZVL_VAL_WIMAX_FLAT_GRD                                            1
#define RSZVL_VAL_WIMAX_FLAT_DIF                                            2

// rszvl_rngWiMAXDemod
#define RSZVL_VAL_WIMAX_BPSK1_2                                             0
#define RSZVL_VAL_WIMAX_QPSK1_2                                             1
#define RSZVL_VAL_WIMAX_QPSK3_4                                             2
#define RSZVL_VAL_WIMAX_16QAM1_2                                            3
#define RSZVL_VAL_WIMAX_16QAM3_4                                            4
#define RSZVL_VAL_WIMAX_64QAM2_3                                            5
#define RSZVL_VAL_WIMAX_64QAM3_4                                            6

// rszvl_rngWiMAXBconAuto
#define RSZVL_VAL_WIMAX_BCON_NONE                                           0
#define RSZVL_VAL_WIMAX_BCON_FIRST                                          1
#define RSZVL_VAL_WIMAX_BCON_USER                                           2
#define RSZVL_VAL_WIMAX_BCON_ALL                                            3
#define RSZVL_VAL_WIMAX_BCON_SIGNAL                                         4
#define RSZVL_VAL_WIMAX_BCON_PREDEF                                         5

// rszvl_rngWiMAXLinkMode
#define RSZVL_VAL_WIMAX_LMOD_DL                                             0
#define RSZVL_VAL_WIMAX_LMOD_UL                                             1

// rszvl_rngWiMAXCestType
#define RSZVL_VAL_WIMAX_CEST_PREAMONLY                                      0
#define RSZVL_VAL_WIMAX_CEST_PREAMPAYL                                      1
#define RSZVL_VAL_WIMAX_CEST_PAYLONLY                                       2

// rszvl_rngWiMAXTracPilots
#define RSZVL_VAL_WIMAX_TRAC_PIL_PRED                                       0
#define RSZVL_VAL_WIMAX_TRAC_PIL_DET                                        1

// rszvl_rngWiMAXDLFramePreMode
#define RSZVL_VAL_WIMAX_PRE_AUTO                                            0
#define RSZVL_VAL_WIMAX_PRE_USER                                            1

// rszvl_rngWiMAXZoneType
#define RSZVL_VAL_WIMAX_ZONE_TYPE_DLFUSC                                    0
#define RSZVL_VAL_WIMAX_ZONE_TYPE_DLPUSC                                    1
#define RSZVL_VAL_WIMAX_ZONE_TYPE_ULPUSC                                    2

// rszvl_rngWiMAXZoneBurstFormat
#define RSZVL_VAL_WIMAX_ZONE_BURST_FORM_QPSK_12                             0
#define RSZVL_VAL_WIMAX_ZONE_BURST_FORM_QPSK_34                             1
#define RSZVL_VAL_WIMAX_ZONE_BURST_FORM_QAM16_12                            2
#define RSZVL_VAL_WIMAX_ZONE_BURST_FORM_QAM16_34                            3
#define RSZVL_VAL_WIMAX_ZONE_BURST_FORM_QAM64_12                            4
#define RSZVL_VAL_WIMAX_ZONE_BURST_FORM_QAM64_23                            5
#define RSZVL_VAL_WIMAX_ZONE_BURST_FORM_QAM64_34                            6

// rszvl_rngWiMAXZoneBurstType
#define RSZVL_VAL_WIMAX_ZONE_BURS_TYPE_FCH                                  0
#define RSZVL_VAL_WIMAX_ZONE_BURS_TYPE_DLMAP                                1
#define RSZVL_VAL_WIMAX_ZONE_BURS_TYPE_ULMAP                                2
#define RSZVL_VAL_WIMAX_ZONE_BURS_TYPE_DATA                                 3

// rszvl_rngTriggerSourceB6
#define RSZVL_VAL_TRG_B6_TV                                                 0

// rszvl_rngTriggerTVVertSig
#define RSZVL_VAL_TRIG_TV_VFIEL_ALL                                         0
#define RSZVL_VAL_TRIG_TV_VFIEL_ODD                                         1
#define RSZVL_VAL_TRIG_TV_VFIEL_EVEN                                        2
#define RSZVL_VAL_TRIG_TV_HORIZONTAL                                        3

// rszvl_rngDISPFormat
#define RSZVL_VAL_DISP_FORMAT_SINGLE                                        0
#define RSZVL_VAL_DISP_FORMAT_SPLIT                                         1

// rszvl_rngSpemAnalysis
#define RSZVL_VAL_SPEM_ANALYSIS_DL                                          0
#define RSZVL_VAL_SPEM_ANALYSIS_UL                                          1

// rszvl_rngWiMAXSignalInput
#define RSZVL_VAL_WIMAX_SIGNAL_INP_RF                                       0

// rszvl_rngWiMAXESPTTA
#define RSZVL_VAL_WIMAX_ESP_TTA_STAN                                        0
#define RSZVL_VAL_WIMAX_ESP_TTA_USER                                        1

// rszvl_rng3GPPESPMode
#define RSZVL_VAL_3GPP_ESP_MODE_AUTO                                        0
#define RSZVL_VAL_3GPP_ESP_MODE_MAN                                         1
#define RSZVL_VAL_3GPP_ESP_MODE_USER                                        2

// rszvl_rngNoiseMarkerTrace
#define RSZVL_VAL_NOISE_MARKER_TRACE_NOIS                                   0
#define RSZVL_VAL_NOISE_MARKER_TRACE_GAIN                                   1

// rszvl_rngUnitSecMin
#define RSZVL_VAL_UNIT_SEC                                                  0
#define RSZVL_VAL_UNIT_MIN                                                  1

// rszvl_rngCatvAtvCarrMeas
#define RSZVL_VAL_ATV_MEAS_CARR_CARR                                        0
#define RSZVL_VAL_ATV_MEAS_CARR_NOIS                                        1

// rszvl_rngCatvAtvMeas
#define RSZVL_VAL_ATV_MEAS_ASP                                              0
#define RSZVL_VAL_ATV_MEAS_CARR                                             1
#define RSZVL_VAL_ATV_MEAS_CN                                               2
#define RSZVL_VAL_ATV_MEAS_CSO                                              3
#define RSZVL_VAL_ATV_MEAS_CTB                                              4
#define RSZVL_VAL_ATV_MEAS_HUM                                              5
#define RSZVL_VAL_ATV_MEAS_VMOD                                             6
#define RSZVL_VAL_ATV_MEAS_VSC                                              7

// rszvl_rngCatvDtvMeas
#define RSZVL_VAL_DTV_MEAS_DSP                                              0
#define RSZVL_VAL_DTV_MEAS_CHP                                              1
#define RSZVL_VAL_DTV_MEAS_CONS                                             2
#define RSZVL_VAL_DTV_MEAS_MERR                                             3
#define RSZVL_VAL_DTV_MEAS_EPATT                                            4
#define RSZVL_VAL_DTV_MEAS_APD                                              5
#define RSZVL_VAL_DTV_MEAS_CCDF                                             6
#define RSZVL_VAL_DTV_MEAS_OVER                                             7

// rszvl_rngCatvTvMeas
#define RSZVL_VAL_TV_MEAS_TILT                                              0

// rszvl_rngCatvMerrZoom
#define RSZVL_VAL_CATV_MERR_ZOOM_NONE                                       0
#define RSZVL_VAL_CATV_MERR_ZOOM_MERR                                       1
#define RSZVL_VAL_CATV_MERR_ZOOM_MERP                                       2
#define RSZVL_VAL_CATV_MERR_ZOOM_EVMR                                       3
#define RSZVL_VAL_CATV_MERR_ZOOM_EVMP                                       4
#define RSZVL_VAL_CATV_MERR_ZOOM_IMB                                        5
#define RSZVL_VAL_CATV_MERR_ZOOM_QERR                                       6
#define RSZVL_VAL_CATV_MERR_ZOOM_SUP                                        7
#define RSZVL_VAL_CATV_MERR_ZOOM_PJIT                                       8

// rszvl_rngCatvOverZoom
#define RSZVL_VAL_CATV_OVER_ZOOM_NONE                                       0
#define RSZVL_VAL_CATV_OVER_ZOOM_MERR                                       1
#define RSZVL_VAL_CATV_OVER_ZOOM_MERP                                       2
#define RSZVL_VAL_CATV_OVER_ZOOM_EVMR                                       3
#define RSZVL_VAL_CATV_OVER_ZOOM_EVMP                                       4
#define RSZVL_VAL_CATV_OVER_ZOOM_CFOF                                       5
#define RSZVL_VAL_CATV_OVER_ZOOM_SROF                                       6

// rszvl_rngCatvConstDiagZoom
#define RSZVL_VAL_CATV_QUAD_ZOOM_NONE                                       0
#define RSZVL_VAL_CATV_QUAD_ZOOM_1                                          1
#define RSZVL_VAL_CATV_QUAD_ZOOM_2                                          2
#define RSZVL_VAL_CATV_QUAD_ZOOM_3                                          3
#define RSZVL_VAL_CATV_QUAD_ZOOM_4                                          4

// rszvl_rngCatvAtvCarrMeasMode
#define RSZVL_VAL_ATV_MEAS_MODE_INSERVICE                                   0
#define RSZVL_VAL_ATV_MEAS_MODE_OFFSERVICE                                  1
#define RSZVL_VAL_ATV_MEAS_MODE_QUIETLINE                                   2

// rszvl_rngCatvAtvRefPowerMode
#define RSZVL_VAL_ATV_RPOW_MODE_RCH                                         0
#define RSZVL_VAL_ATV_RPOW_MODE_MCH                                         1
#define RSZVL_VAL_ATV_RPOW_MODE_MAN                                         2

// rszvl_rngCatvDtvFilterAlpha
#define RSZVL_VAL_CATV_DTV_ALPHA_R012                                       0
#define RSZVL_VAL_CATV_DTV_ALPHA_R013                                       1
#define RSZVL_VAL_CATV_DTV_ALPHA_R015                                       2
#define RSZVL_VAL_CATV_DTV_ALPHA_R018                                       3

// rszvl_rngCatvSBand
#define RSZVL_VAL_CATV_SBAND_NORM                                           0
#define RSZVL_VAL_CATV_SBAND_INV                                            1
#define RSZVL_VAL_CATV_SBAND_AUTO                                           2

// rszvl_rngCatvModStandGroupDelay
#define RSZVL_VAL_CATV_GDELAY_GENERAL                                       0
#define RSZVL_VAL_CATV_GDELAY_GHALF                                         1
#define RSZVL_VAL_CATV_GDELAY_AUSTRALIA                                     2
#define RSZVL_VAL_CATV_GDELAY_DANMARK                                       3
#define RSZVL_VAL_CATV_GDELAY_NEWZELAND                                     4
#define RSZVL_VAL_CATV_GDELAY_NORWAY                                        5
#define RSZVL_VAL_CATV_GDELAY_SWEDENFULL                                    6
#define RSZVL_VAL_CATV_GDELAY_FLAT                                          7
#define RSZVL_VAL_CATV_GDELAY_OIRT                                          8
#define RSZVL_VAL_CATV_GDELAY_CHINA                                         9
#define RSZVL_VAL_CATV_GDELAY_CCIR                                          10
#define RSZVL_VAL_CATV_GDELAY_TDF                                           11
#define RSZVL_VAL_CATV_GDELAY_FFC                                           12

// rszvl_rngCatvModStandSignalType
#define RSZVL_VAL_CATV_SIG_TYPE_ATV                                         0
#define RSZVL_VAL_CATV_SIG_TYPE_DTV                                         1

// rszvl_rngCatvAtvStandard
#define RSZVL_VAL_ATV_STAN_BG                                               0
#define RSZVL_VAL_ATV_STAN_DK                                               1
#define RSZVL_VAL_ATV_STAN_I                                                2
#define RSZVL_VAL_ATV_STAN_K1                                               3
#define RSZVL_VAL_ATV_STAN_L                                                4
#define RSZVL_VAL_ATV_STAN_M                                                5
#define RSZVL_VAL_ATV_STAN_N                                                6

// rszvl_rngCatvDtvStandard
#define RSZVL_VAL_DTV_STAN_J83A                                             0
#define RSZVL_VAL_DTV_STAN_J83B                                             1
#define RSZVL_VAL_DTV_STAN_J83C                                             2

// rszvl_rngCatvAtvSoundSystem
#define RSZVL_VAL_ATV_AUDIO_FM55NICAM585                                    0
#define RSZVL_VAL_ATV_AUDIO_FM55FM5742                                      1
#define RSZVL_VAL_ATV_AUDIO_FM55MONO                                        2
#define RSZVL_VAL_ATV_AUDIO_FM65NICAM585                                    3
#define RSZVL_VAL_ATV_AUDIO_FM65FM6258                                      4
#define RSZVL_VAL_ATV_AUDIO_FM65FM6742                                      5
#define RSZVL_VAL_ATV_AUDIO_FM65MONO                                        6
#define RSZVL_VAL_ATV_AUDIO_FM60NICAM6552                                   7
#define RSZVL_VAL_ATV_AUDIO_FM60MONO                                        8
#define RSZVL_VAL_ATV_AUDIO_AM65NICAM585                                    9
#define RSZVL_VAL_ATV_AUDIO_AM60MONO                                        10
#define RSZVL_VAL_ATV_AUDIO_FM45BTSC                                        11
#define RSZVL_VAL_ATV_AUDIO_FM45EIA_J                                       12
#define RSZVL_VAL_ATV_AUDIO_FM45FM4742                                      13
#define RSZVL_VAL_ATV_AUDIO_FM45MONO                                        14

// rszvl_rngCatvAtvColorSystem
#define RSZVL_VAL_ATV_COLOR_SYSTEM_PAL                                      0
#define RSZVL_VAL_ATV_COLOR_SYSTEM_NTSC                                     1
#define RSZVL_VAL_ATV_COLOR_SYSTEM_SECAM                                    2

// rszvl_rngCatvDtvFormat
#define RSZVL_VAL_DTV_FORMAT_QAM4                                           0
#define RSZVL_VAL_DTV_FORMAT_QAM16                                          1
#define RSZVL_VAL_DTV_FORMAT_QAM32                                          2
#define RSZVL_VAL_DTV_FORMAT_QAM64                                          3
#define RSZVL_VAL_DTV_FORMAT_QAM128                                         4
#define RSZVL_VAL_DTV_FORMAT_QAM256                                         5
#define RSZVL_VAL_DTV_FORMAT_QAM512                                         6
#define RSZVL_VAL_DTV_FORMAT_QAM1024                                        7

// rszvl_rngCatvAtvTriggBarLineType
#define RSZVL_VAL_ATV_BLIN_TYPE_CCIR17                                      0
#define RSZVL_VAL_ATV_BLIN_TYPE_FCC                                         1
#define RSZVL_VAL_ATV_BLIN_TYPE_NTC7                                        2

// rszvl_rngPmetMeasTime
#define RSZVL_VAL_PMET_MEASTIME_NORM                                        0
#define RSZVL_VAL_PMET_MEASTIME_SHORT                                       1
#define RSZVL_VAL_PMET_MEASTIME_LONG                                        2

// rszvl_rngPmetCoupling
#define RSZVL_VAL_PMET_COUP_OFF                                             0
#define RSZVL_VAL_PMET_COUP_CENT                                            1
#define RSZVL_VAL_PMET_COUP_MARK                                            2

// rszvl_rngUnitLength
#define RSZVL_VAL_UNIT_LEN_METER                                            0
#define RSZVL_VAL_UNIT_LEN_FOOT                                             1

// rszvl_rngRipplePhysicalUnits
#define RSZVL_VAL_RIPP_PHYS_UNIT_FLIN                                       0
#define RSZVL_VAL_RIPP_PHYS_UNIT_FLOG                                       1
#define RSZVL_VAL_RIPP_PHYS_UNIT_FSEG                                       2
#define RSZVL_VAL_RIPP_PHYS_UNIT_FSIN                                       3
#define RSZVL_VAL_RIPP_PHYS_UNIT_TLIN                                       4
#define RSZVL_VAL_RIPP_PHYS_UNIT_TLOG                                       5
#define RSZVL_VAL_RIPP_PHYS_UNIT_PLIN                                       6
#define RSZVL_VAL_RIPP_PHYS_UNIT_PLOG                                       7
#define RSZVL_VAL_RIPP_PHYS_UNIT_PSIN                                       8

// rszvl_rngRippleFormUnits
#define RSZVL_VAL_RIPP_FORM_UNIT_COMP                                       0
#define RSZVL_VAL_RIPP_FORM_UNIT_MAGN                                       1
#define RSZVL_VAL_RIPP_FORM_UNIT_PHAS                                       2
#define RSZVL_VAL_RIPP_FORM_UNIT_REAL                                       3
#define RSZVL_VAL_RIPP_FORM_UNIT_IMAG                                       4
#define RSZVL_VAL_RIPP_FORM_UNIT_SWR                                        5
#define RSZVL_VAL_RIPP_FORM_UNIT_GDEL                                       6
#define RSZVL_VAL_RIPP_FORM_UNIT_L                                          7
#define RSZVL_VAL_RIPP_FORM_UNIT_C                                          8

// rszvl_rngEventStart
#define RSZVL_VAL_TRG_START_SWEEP                                           0
#define RSZVL_VAL_TRG_START_POINT                                           1
/* END GENERATE */

/*****************************************************************************
 *- Hidden Attribute Declarations -------------------------------------------*
 *****************************************************************************/

#define RSZVL_ATTR_OPC_TIMEOUT               (RS_ATTR_OPC_TIMEOUT)           /* ViInt32 */
#define RSZVL_ATTR_IO_SESSION                (RS_ATTR_IO_SESSION)            /* ViSession */
#define RSZVL_ATTR_OPC_CALLBACK              (RS_ATTR_OPC_CALLBACK)          /* ViAddr */
#define RSZVL_ATTR_CHECK_STATUS_CALLBACK     (RS_ATTR_CHECK_STATUS_CALLBACK) /* ViAddr */

/****************************************************************************
 *---------------- Constant definition  ------------------------------------*
 ****************************************************************************/

#define RSZVL_VAL_MAX_TIME_IMMEDIATE                             (0x0L)
#define RSZVL_VAL_MAX_TIME_INFINITE                              (0xFFFFFFFFUL)

#define RSZVL_VAL_MARKER_SEARCH_HIGHEST                          (1L)
#define RSZVL_VAL_MARKER_SEARCH_MINIMUM                          (2L)
#define RSZVL_VAL_MARKER_SEARCH_NEXT_PEAK                        (3L)
#define RSZVL_VAL_MARKER_SEARCH_NEXT_PEAK_LEFT                   (4L)
#define RSZVL_VAL_MARKER_SEARCH_NEXT_PEAK_RIGHT                  (5L)
#define RSZVL_VAL_MARKER_SEARCH_NEXT_MINIMUM                     (6L) 
#define RSZVL_VAL_MARKER_SEARCH_NEXT_MINIMUM_LEFT                (7L) 
#define RSZVL_VAL_MARKER_SEARCH_NEXT_MINIMUM_RIGHT               (8L) 

#define RSZVL_VAL_MARKER_SEARCH_CLASS_EXT_BASE                   (500L)
#define RSZVL_VAL_MARKER_SEARCH_SPECIFIC_EXT_BASE                (1000L)

#define RSZVL_VAL_INSTRUMENT_SETTING_FREQUENCY_CENTER            (1L)
#define RSZVL_VAL_INSTRUMENT_SETTING_FREQUENCY_STEP              (2L)
#define RSZVL_VAL_INSTRUMENT_SETTING_REFERENCE_LEVEL             (3L)

#define RSZVL_VAL_INSTRUMENT_SETTING_CLASS_EXT_BASE              (500L)
#define RSZVL_VAL_INSTRUMENT_SETTING_SPECIFIC_EXT_BASE           (1000L)

#define RSZVL_VAL_COUPLING_RBW        0
#define RSZVL_VAL_COUPLING_VBW        1
#define RSZVL_VAL_COUPLING_SWEEP_TIME 2

#define RSZVL_VAL_COUPLING_RATIO_RBW 0
#define RSZVL_VAL_COUPLING_RATIO_VBW 1

#define RSZVL_VAL_LIMIT_ALL         3
#define RSZVL_VAL_LIMIT_CONTROL     3
#define RSZVL_VAL_LIMIT_MOVE_OFFSET 0
#define RSZVL_VAL_LIMIT_MOVE_SHIFT  1

#define RSZVL_VAL_MEASTYPE_NONE  0
#define RSZVL_VAL_MEASTYPE_MIN   1
#define RSZVL_VAL_MEASTYPE_MAX   2
#define RSZVL_VAL_MEASTYPE_PEAK  3
#define RSZVL_VAL_MEASTYPE_AVER  4

#define RSZVL_VAL_MEASTYPE_MAXMIN    0
#define RSZVL_VAL_MEASTYPE_MAXMAX    1
#define RSZVL_VAL_MEASTYPE_MAXAVER   2
#define RSZVL_VAL_MEASTYPE_AVERMIN   3
#define RSZVL_VAL_MEASTYPE_AVERMAX   4

#define RSZVL_VAL_POW_CONT_MODE_REL  0
#define RSZVL_VAL_POW_CONT_MODE_ABS  1

#define RSZVL_VAL_DATE       0

#define RSZVL_VAL_EATT_OFF   0
#define RSZVL_VAL_EATT_AUTO  1
#define RSZVL_VAL_EATT_MAN   2

#define RSZVL_VAL_LINE_THR           0
#define RSZVL_VAL_LINE_DISPLAY       1
#define RSZVL_VAL_LINE_FREQ          2
#define RSZVL_VAL_LINE_TIME          3

#define RSZVL_VAL_ACP_TX         0
#define RSZVL_VAL_ACP_ADJ        1
#define RSZVL_VAL_ACP_ALT        2

#define RSZVL_VAL_SSTAT_OFF      0
#define RSZVL_VAL_SSTAT_APD      1
#define RSZVL_VAL_SSTAT_CCDF     2

#define RSZVL_VAL_SSTAT_PROB_0_1         0
#define RSZVL_VAL_SSTAT_PROB_1           1
#define RSZVL_VAL_SSTAT_PROB_10          2

#define RSZVL_VAL_SSTAT_RESULT_MEAN      0
#define RSZVL_VAL_SSTAT_RESULT_PEAK      1
#define RSZVL_VAL_SSTAT_RESULT_CREST     2

#define RSZVL_VAL_TDOM_PEAK            0
#define RSZVL_VAL_TDOM_RMS             1
#define RSZVL_VAL_TDOM_MEAN            2
#define RSZVL_VAL_TDOM_SDEV            3

#define RSZVL_VAL_TDOM_RESULT              0
#define RSZVL_VAL_TDOM_AVG_RESULT          1
#define RSZVL_VAL_TDOM_MAX_HOLD_RESULT     2

#define RSZVL_VAL_BURS_MEAS_MEAN      0
#define RSZVL_VAL_BURS_MEAS_PEAK      1

#define RSZVL_VAL_FILE_ITEMS_NONE            0
#define RSZVL_VAL_FILE_ITEMS_DEFAULT         1
#define RSZVL_VAL_FILE_ITEMS_ALL             2
#define RSZVL_VAL_FILE_ITEMS_HWSETTINGS      3
#define RSZVL_VAL_FILE_ITEMS_ALLTRACES       4
#define RSZVL_VAL_FILE_ITEMS_ALLLINES        5
#define RSZVL_VAL_FILE_ITEMS_SOUR_CAL_DATA   6
#define RSZVL_VAL_FILE_ITEMS_ALLTRAN         7
#define RSZVL_VAL_FILE_ITEMS_SPECTROGRAM     8

#define RSZVL_VAL_FILE_NEW    0
#define RSZVL_VAL_FILE_COPY   1
#define RSZVL_VAL_FILE_RENAME 2
#define RSZVL_VAL_FILE_DELETE 3
#define RSZVL_VAL_FILE_RDIR   4
#define RSZVL_VAL_FILE_CDIR   5
#define RSZVL_VAL_FILE_CDISC  6

#define RSZVL_VAL_MEM_SAVE    0
#define RSZVL_VAL_MEM_RECALL  1
#define RSZVL_VAL_MEM_DELETE  2
#define RSZVL_VAL_MEM_STARTUP 3

#define RSZVL_VAL_HCOPY_DEST_BMP  0
#define RSZVL_VAL_HCOPY_DEST_WMF  1
#define RSZVL_VAL_HCOPY_DEST_EMF  2
#define RSZVL_VAL_HCOPY_DEST_CLP  3
#define RSZVL_VAL_HCOPY_DEST_PRN  4

#define RSZVL_VAL_HCOP_ALL   0
#define RSZVL_VAL_HCOP_TRACE 1

#define RSZVL_VAL_FORMAT_ASC  0
#define RSZVL_VAL_FORMAT_REAL 1

#define RSZVL_VAL_TABLE_OFF          0
#define RSZVL_VAL_TABLE_SELECT       1
#define RSZVL_VAL_TABLE_NEW          2
#define RSZVL_VAL_TABLE_DELETE       3
#define RSZVL_VAL_TABLE_COPY         4

#define RSZVL_VAL_USER_PORT_INP      0
#define RSZVL_VAL_USER_PORT_OUT      1

#define RSZVL_VAL_FREQ_START         0
#define RSZVL_VAL_FREQ_STOP          1
#define RSZVL_VAL_FREQ_CENTER        2
#define RSZVL_VAL_FREQ_SPAN          3

#define RSZVL_VAL_ATTEN_AREC         0
#define RSZVL_VAL_ATTEN_BREC         1

#define RSZVL_VAL_SWE_SEG_OP_ADD     0
#define RSZVL_VAL_SWE_SEG_OP_INS     1
#define RSZVL_VAL_SWE_SEG_OP_EDIT    2

#define RSZVL_VAL_SEG_TIME           0
#define RSZVL_VAL_SEG_POINT          1

#define RSZVL_VAL_CORR_MAN_COPY      0
#define RSZVL_VAL_CORR_MAN_APPLY     1
#define RSZVL_VAL_CORR_MAN_RESOLVE   2
#define RSZVL_VAL_CORR_MAN_DELETE    3

#define RSZVL_VAL_CONNECTOR_GENDER_MALE   0
#define RSZVL_VAL_CONNECTOR_GENDER_FEMALE 1

#define RSZVL_VAL_CONNECTION_MODE_TEM     0
#define RSZVL_VAL_CONNECTION_MODE_WGUIDE  1

#define RSZVL_VAL_CONNECTION_CONNECTOR_GENDER   0
#define RSZVL_VAL_CONNECTION_CONNECTOR_NGENDER  1

#define RSZVL_VAL_CALTYPE_REFL       0
#define RSZVL_VAL_CALTYPE_RSH        1
#define RSZVL_VAL_CALTYPE_FOP        2
#define RSZVL_VAL_CALTYPE_FTR        3
#define RSZVL_VAL_CALTYPE_RTR        4
#define RSZVL_VAL_CALTYPE_FRTR       5
#define RSZVL_VAL_CALTYPE_OPTP       6
#define RSZVL_VAL_CALTYPE_TOSM       7

#define RSZVL_VAL_CALCOLLSTDGEN_THR    0
#define RSZVL_VAL_CALCOLLSTDGEN_OPEN   1
#define RSZVL_VAL_CALCOLLSTDGEN_SHORT  2
#define RSZVL_VAL_CALCOLLSTDGEN_MATCH  3
#define RSZVL_VAL_CALCOLLSTDGEN_NET    4
#define RSZVL_VAL_CALCOLLSTDGEN_ATT    5
#define RSZVL_VAL_CALCOLLSTDGEN_REFL   6
#define RSZVL_VAL_CALCOLLSTDGEN_SLIDE  7
#define RSZVL_VAL_CALCOLLSTDGEN_LINE   8

#define RSZVL_VAL_CAL_DATA_DIRECTIVITY 0
#define RSZVL_VAL_CAL_DATA_SRCMATCH    1
#define RSZVL_VAL_CAL_DATA_REFLTRACK   2
#define RSZVL_VAL_CAL_DATA_ISOLATION   3
#define RSZVL_VAL_CAL_DATA_LOADMATCH   4
#define RSZVL_VAL_CAL_DATA_TRANSTRACK  5

#define RSZVL_VAL_CONNECTOR_KIT_N50   0
#define RSZVL_VAL_CONNECTOR_KIT_N75   1
#define RSZVL_VAL_CONNECTOR_KIT_PC7   2
#define RSZVL_VAL_CONNECTOR_KIT_PC35  3
#define RSZVL_VAL_CONNECTOR_KIT_PC292 4
#define RSZVL_VAL_CONNECTOR_KIT_USER1 5
#define RSZVL_VAL_CONNECTOR_KIT_USER2 6
#define RSZVL_VAL_CONNECTOR_KIT_SMA   7

#define RSZVL_VAL_CALSTD_MMTH        0
#define RSZVL_VAL_CALSTD_FFTH        1
#define RSZVL_VAL_CALSTD_MFTH        2
#define RSZVL_VAL_CALSTD_MMLI        3
#define RSZVL_VAL_CALSTD_FFLI        4
#define RSZVL_VAL_CALSTD_MFLI        5
#define RSZVL_VAL_CALSTD_MML2        6
#define RSZVL_VAL_CALSTD_FFL2        7
#define RSZVL_VAL_CALSTD_MFL2        8
#define RSZVL_VAL_CALSTD_OSH         9
#define RSZVL_VAL_CALSTD_MOSH        10
#define RSZVL_VAL_CALSTD_FOSH        11
#define RSZVL_VAL_CALSTD_MMAT        12
#define RSZVL_VAL_CALSTD_FFAT        13
#define RSZVL_VAL_CALSTD_MFAT        14
#define RSZVL_VAL_CALSTD_MMSN        15
#define RSZVL_VAL_CALSTD_FFSN        16
#define RSZVL_VAL_CALSTD_MFSN        17
#define RSZVL_VAL_CALSTD_MOP         18
#define RSZVL_VAL_CALSTD_FOP         19
#define RSZVL_VAL_CALSTD_MSH         20
#define RSZVL_VAL_CALSTD_FSH         21
#define RSZVL_VAL_CALSTD_MREF        22
#define RSZVL_VAL_CALSTD_FREF        23
#define RSZVL_VAL_CALSTD_MMTC        24
#define RSZVL_VAL_CALSTD_FMTC        25
#define RSZVL_VAL_CALSTD_MSM         26
#define RSZVL_VAL_CALSTD_FSM         27

#define RSZVL_VAL_OPEN               0
#define RSZVL_VAL_SHORT              1
#define RSZVL_VAL_MATCH              2

#define RSZVB_DISPLAY_RESULTS_EPD    0
#define RSZVB_DISPLAY_RESULTS_MMPT   1
#define RSZVB_DISPLAY_RESULTS_MSTD   2
#define RSZVB_DISPLAY_RESULTS_RMS    3
#define RSZVB_DISPLAY_RESULTS_SFL    4
#define RSZVB_DISPLAY_RESULTS_COMP   5

#define RSZVL_VAL_UNFORMATTED        0
#define RSZVL_VAL_FORMATTED          1
#define RSZVL_VAL_UNFORMATTED_MATH   2

#define RSZVL_VAL_COMPLEX            0
#define RSZVL_VAL_LINP               1
#define RSZVL_VAL_LOGP               2

#define RSZVL_VAL_INSTR_SETTING_FREQ_CENTER 0
#define RSZVL_VAL_INSTR_SETTING_FREQ_START  1
#define RSZVL_VAL_INSTR_SETTING_FREQ_STOP   2

#define RSZVL_VAL_MEAN               0
#define RSZVL_VAL_STDDEV             1
#define RSZVL_VAL_MAX                2
#define RSZVL_VAL_MIN                3
#define RSZVL_VAL_RMS                4
#define RSZVL_VAL_PTPEAK             5
#define RSZVL_VAL_ELENGTH            6
#define RSZVL_VAL_PDELAY             7
#define RSZVL_VAL_ALL                8

#define RSZVL_VAL_SFACTOR_K          0
#define RSZVL_VAL_SFACTOR_U1         1
#define RSZVL_VAL_SFACTOR_U2         2

#define RSZVL_VAL_ELEMENT_BACKGROUND      1
#define RSZVL_VAL_ELEMENT_TEXT            2
#define RSZVL_VAL_ELEMENT_SELTEXT         3
#define RSZVL_VAL_ELEMENT_GRID            4
#define RSZVL_VAL_ELEMENT_REFLINE         5
#define RSZVL_VAL_ELEMENT_ALLMAKERS       6
#define RSZVL_VAL_ELEMENT_HLINE           7
#define RSZVL_VAL_ELEMENT_DTITLE          8
#define RSZVL_VAL_ELEMENT_LIMITFAILTRACE  9
#define RSZVL_VAL_ELEMENT_LIMITLINEOFF    10
#define RSZVL_VAL_ELEMENT_LIMITLINEUPPER  11
#define RSZVL_VAL_ELEMENT_LIMITLINELOWER  12
#define RSZVL_VAL_ELEMENT_TRACE1          13
#define RSZVL_VAL_ELEMENT_TRACE2          14
#define RSZVL_VAL_ELEMENT_TRACE3          15
#define RSZVL_VAL_ELEMENT_TRACE4          16
#define RSZVL_VAL_ELEMENT_TRACE5          17
#define RSZVL_VAL_ELEMENT_TRACE6          18
#define RSZVL_VAL_ELEMENT_TRACE7          19
#define RSZVL_VAL_ELEMENT_TRACE8          20
#define RSZVL_VAL_ELEMENT_TRACE9          21
#define RSZVL_VAL_ELEMENT_TRACE10         22
#define RSZVL_VAL_ELEMENT_TRACE11         23
#define RSZVL_VAL_ELEMENT_TRACE12         24
#define RSZVL_VAL_ELEMENT_TRACE13         25
#define RSZVL_VAL_ELEMENT_TRACE14         26
#define RSZVL_VAL_ELEMENT_TRACE15         27
#define RSZVL_VAL_ELEMENT_TRACE16         28

#define RSZVL_VAL_CMAP_SAVE          0
#define RSZVL_VAL_CMAP_LOAD          1

#define RSZVL_VAL_RESTORE            0
#define RSZVL_VAL_MAXIMIZE           1

#define RSZVL_VAL_FILE_MAN_CDRIVE    0
#define RSZVL_VAL_FILE_MAN_CDIR      1
#define RSZVL_VAL_FILE_MAN_MDIR      2
#define RSZVL_VAL_FILE_MAN_RDIR      3
#define RSZVL_VAL_FILE_MAN_COPY      4
#define RSZVL_VAL_FILE_MAN_MOVE      5
#define RSZVL_VAL_FILE_MAN_DELETE    6

#define RSZVL_VAL_DISPLAY_RESULTS_ALL   0
#define RSZVL_VAL_DISPLAY_RESULTS_EPD   1
#define RSZVL_VAL_DISPLAY_RESULTS_MMPT  2
#define RSZVL_VAL_DISPLAY_RESULTS_MSTD  3
#define RSZVL_VAL_DISPLAY_RESULTS_RMS   4

#define RSZVL_VAL_TRACE_STYLE_SOLID     0
#define RSZVL_VAL_TRACE_STYLE_DASHED    1
#define RSZVL_VAL_TRACE_STYLE_DOTTED    2
#define RSZVL_VAL_TRACE_STYLE_DDOTTED   3
#define RSZVL_VAL_TRACE_STYLE_DDDOTTED  4

#define RSZVL_VAL_LIN                   0
#define RSZVL_VAL_LOG                   1

#define RSZVL_VAL_ACP_LIMIT_CHECK_OFF   0
#define RSZVL_VAL_ACP_LIMIT_CHECK_ABS   1
#define RSZVL_VAL_ACP_LIMIT_CHECK_REL   2

#define RSZVL_VAL_X  0
#define RSZVL_VAL_Y  1

/* K9 Power Meter - constants */
#define RSZVL_VAL_PWMET_MEASTIME_NORM    0
#define RSZVL_VAL_PWMET_MEASTIME_SHORT   1
#define RSZVL_VAL_PWMET_MEASTIME_LONG    2
#define RSZVL_VAL_PWMET_MEASTIME_MANUAL  3

/* K7 - FM Demodulator - constants  */

#define RSZVL_VAL_FMDEM_AM         0
#define RSZVL_VAL_FMDEM_FM         1
#define RSZVL_VAL_FMDEM_PM         2
#define RSZVL_VAL_FMDEM_AFSFM      3
#define RSZVL_VAL_FMDEM_AFSPM      4
#define RSZVL_VAL_FMDEM_RF         5
#define RSZVL_VAL_FMDEM_AMREL      6
#define RSZVL_VAL_FMDEM_AFSAMREL   7

#define RSZVL_VAL_FMDEM_RESULT_OFF  0
#define RSZVL_VAL_FMDEM_RESULT_CURR 1
#define RSZVL_VAL_FMDEM_RESULT_AVG  2
#define RSZVL_VAL_FMDEM_RESULT_MAX  3
#define RSZVL_VAL_FMDEM_RESULT_MIN  4
#define RSZVL_VAL_FMDEM_RESULT_VIEW 5

#define RSZVL_VAL_FMDEM_FILTER_HP     0
#define RSZVL_VAL_FMDEM_FILTER_LP     1
#define RSZVL_VAL_FMDEM_FILTER_DEEMP  2
#define RSZVL_VAL_FMDEM_FILTER_LP_REL 3

#define RSZVL_VAL_FMDEM_COUP_DC 0
#define RSZVL_VAL_FMDEM_COUP_AC 1

#define RSZVL_VAL_FMDEM_RLEN  0
#define RSZVL_VAL_FMDEM_SRATE 1

#define RSZVL_VAL_FMDEM_IMM 0
#define RSZVL_VAL_FMDEM_AVG 1

#define RSZVL_VAL_FMDEM_DET_PPE 0
#define RSZVL_VAL_FMDEM_DET_MPE 1
#define RSZVL_VAL_FMDEM_DET_MID 2
#define RSZVL_VAL_FMDEM_DET_RMS 3

#define RSZVL_VAL_FMDEM_MEAS_AFR 0
#define RSZVL_VAL_FMDEM_MEAS_FER 1
#define RSZVL_VAL_FMDEM_MEAS_SIN 2
#define RSZVL_VAL_FMDEM_MEAS_THD 3
#define RSZVL_VAL_FMDEM_MEAS_CAR 4

#define RSZVL_VAL_FM_FEED_AM_REL         0
#define RSZVL_VAL_FM_FEED_AM             1 
#define RSZVL_VAL_FM_FEED_SPEC           2
#define RSZVL_VAL_FM_FEED_FM             3  
#define RSZVL_VAL_FM_FEED_PM             4
#define RSZVL_VAL_FM_FEED_AMS            5
#define RSZVL_VAL_FM_FEED_AMS_REL        6
#define RSZVL_VAL_FM_FEED_FMS            7
#define RSZVL_VAL_FM_FEED_PMS            8 
#define RSZVL_VAL_FM_FEED_FM_AFSP        9
#define RSZVL_VAL_FM_FEED_PM_AFSP        10
#define RSZVL_VAL_FM_FEED_AM_REL_AFSP    11

/* K30 - Noise Figure - constants  */

#define RSZVL_VAL_NOISE_DUT_MODE_DIRECT        0
#define RSZVL_VAL_NOISE_DUT_MODE_FIXED_LO_PLUS 1
#define RSZVL_VAL_NOISE_DUT_MODE_FIXED_LO_ABS  2

#define RSZVL_VAL_NOISE_SECOND_STAGE_CORR_MEAS 0
#define RSZVL_VAL_NOISE_CONT_FREQ_LIST_MEAS    1
#define RSZVL_VAL_NOISE_SINGLE_FREQ_LIST_MEAS  2
#define RSZVL_VAL_NOISE_SINGLE_FREQ_MEAS       3

#define RSZVL_VAL_DISP_FORMAT_SING             0
#define RSZVL_VAL_DISP_FORMAT_SPL              1

#define RSZVL_VAL_NOISE_DISP_DATA_OFF          0

#define RSZVL_VAL_NOISE_NOISE                1
#define RSZVL_VAL_NOISE_GAIN                 2

#define RSZVL_VAL_NOISE_DISP_FORM_SPLIT      0
#define RSZVL_VAL_NOISE_DISP_FORM_SINGLE     1

#define RSZVL_VAL_NOISE_FETC_FIGURE       0
#define RSZVL_VAL_NOISE_FETC_TEMPERATURE  1
#define RSZVL_VAL_NOISE_FETC_GAIN         2

/* 91 - WLAN constants  */   

#define RSZVL_VAL_WLAN_GATE_TIME      0
#define RSZVL_VAL_WLAN_GATE_SAMPLE    1

#define RSZVL_VAL_WLAN_EVM_MODE_SYMBOL    0
#define RSZVL_VAL_WLAN_EVM_MODE_CARRIER   1
#define RSZVL_VAL_WLAN_EVM_MODE_PHAS      2
#define RSZVL_VAL_WLAN_EVM_MODE_FREQ      3

#define RSZVL_VAL_WLAN_SPEC_MODE_IEEE     0
#define RSZVL_VAL_WLAN_SPEC_MODE_ETSI     1
#define RSZVL_VAL_WLAN_SPEC_MODE_FFT      2
#define RSZVL_VAL_WLAN_SPEC_MODE_ACPR     3
#define RSZVL_VAL_WLAN_SPEC_MODE_FLAT     4

#define RSZVL_VAL_WLAN_CONS_SYM       0
#define RSZVL_VAL_WLAN_CONS_CARR      1

#define RSZVL_VAL_WLAN_STAT_CCDF     0
#define RSZVL_VAL_WLAN_STAT_BSTR     1
#define RSZVL_VAL_WLAN_STAT_SFI      2

#define RSZVL_VAL_WLAN_POS_X          0
#define RSZVL_VAL_WLAN_POS_Y          1

#define RSZVL_VAL_WLAN_MARK_MIN       0
#define RSZVL_VAL_WLAN_MARK_MAX       1

#define RSZVL_VAL_WLAN_DEM_DUR_MS     0
#define RSZVL_VAL_WLAN_DEM_DUR_BYTE   1

#define RSZVL_VAL_WLAN_LIM_FERR      0
#define RSZVL_VAL_WLAN_LIM_SYMB      1
#define RSZVL_VAL_WLAN_LIM_IQOFF     2
#define RSZVL_VAL_WLAN_LIM_EVM_ALL   3
#define RSZVL_VAL_WLAN_LIM_EVM_DATA  4
#define RSZVL_VAL_WLAN_LIM_EVM_PIL   5
#define RSZVL_VAL_WLAN_LIM_RTIME     6
#define RSZVL_VAL_WLAN_LIM_FTIME     7
#define RSZVL_VAL_WLAN_LIM_EVM       8


#define RSZVL_VAL_WLAN_ACP_ADJ       0
#define RSZVL_VAL_WLAN_ACP_ALT       1

#define RSZVL_VAL_POW                1

#define RSZVL_VAL_ACP_CURR           0
#define RSZVL_VAL_ACP_MAX            1

#define RSZVL_VAL_WLAN_MEAS_PREA       0
#define RSZVL_VAL_WLAN_MEAS_PAYL       1
#define RSZVL_VAL_WLAN_MEAS_RMS_MAX    2
#define RSZVL_VAL_WLAN_MEAS_RMS_MIN    3
#define RSZVL_VAL_WLAN_MEAS_RMS_AVER   4
#define RSZVL_VAL_WLAN_MEAS_PEAK       5
#define RSZVL_VAL_WLAN_MEAS_CREST_MAX  6
#define RSZVL_VAL_WLAN_MEAS_CREST_MIN  7
#define RSZVL_VAL_WLAN_MEAS_CREST_AVER 8

#define RSZVL_VAL_WLAN_TIME_FALL     0
#define RSZVL_VAL_WLAN_TIME_RISE     1

#define RSZVL_VAL_WLAN_ERR_FREQ      0
#define RSZVL_VAL_WLAN_ERR_SYMB      1

#define RSZVL_VAL_WLAN_IQ_OFFS       0
#define RSZVL_VAL_WLAN_IQ_IMBA       1
#define RSZVL_VAL_WLAN_IQ_QUAD       2

#define RSZVL_VAL_WLAN_IQ_I          0
#define RSZVL_VAL_WLAN_IQ_Q          1
#define RSZVL_VAL_WLAN_IQ_IQ         2

#define RSZVL_VAL_WLAN_EVM_ALL       0
#define RSZVL_VAL_WLAN_EVM_DATA      1
#define RSZVL_VAL_WLAN_EVM_PILOT     2
#define RSZVL_VAL_WLAN_EVM           3
#define RSZVL_VAL_WLAN_EVM_PEAK      4

#define RSZVL_VAL_WLAN_BURS          0
#define RSZVL_VAL_WLAN_SYMB          1

#define RSZVL_VAL_SELECTION_NUM    2

/* K93 - WiMax constants */

#define RSZVL_VAL_WIMAX_MODE_IEEE       0
#define RSZVL_VAL_WIMAX_MODE_ETSI       1
#define RSZVL_VAL_WIMAX_MODE_FFT        2
#define RSZVL_VAL_WIMAX_MODE_FLAT       3
#define RSZVL_VAL_WIMAX_MODE_FLAT_GRDEL 4
#define RSZVL_VAL_WIMAX_MODE_FLAT_DIFF  5
#define RSZVL_VAL_WIMAX_MODE_ACPR_ABS   6
#define RSZVL_VAL_WIMAX_MODE_ACPR_REL   7
#define RSZVL_VAL_WIMAX_MODE_MASK       8
#define RSZVL_VAL_WIMAX_MODE_SFLAT      9

#define RSZVL_VAL_WIMAX_EVM_MODE_SYMBOL    0
#define RSZVL_VAL_WIMAX_EVM_MODE_CARRIER   1
#define RSZVL_VAL_WIMAX_EVM_MODE_PHAS  2
#define RSZVL_VAL_WIMAX_EVM_MODE_FREQ  3

#define RSZVL_VAL_WIMAX_STAT_CCDF    0
#define RSZVL_VAL_WIMAX_STAT_BSTR    1
#define RSZVL_VAL_WIMAX_STAT_BSUM    2

#define RSZVL_VAL_WIMAX_BCON_FIRS    1

#define RSZVL_VAL_WIMAX_CONS_SYM     0
#define RSZVL_VAL_WIMAX_CONS_CARR    1
#define RSZVL_VAL_WIMAX_CONS_BURST   2

#define RSZVL_VAL_WIMAX_ERR_FREQ      0
#define RSZVL_VAL_WIMAX_ERR_SYMB      1

#define RSZVL_VAL_WIMAX_IQ_OFFS 0
#define RSZVL_VAL_WIMAX_IQ_IMBA 1
#define RSZVL_VAL_WIMAX_IQ_QUAD 2

#define RSZVL_VAL_WIMAX_EVM_ALL       0
#define RSZVL_VAL_WIMAX_EVM_DATA      1
#define RSZVL_VAL_WIMAX_EVM_PILOT     2

#define RSZVL_VAL_WIMAX_LIM_FERR      0
#define RSZVL_VAL_WIMAX_LIM_SYMB      1
#define RSZVL_VAL_WIMAX_LIM_SST       2
#define RSZVL_VAL_WIMAX_LIM_EVM_ALL   3
#define RSZVL_VAL_WIMAX_LIM_EVM_DATA  4
#define RSZVL_VAL_WIMAX_LIM_IQOFF     5
#define RSZVL_VAL_WIMAX_LIM_BERP      6

#define RSZVL_VAL_WIMAX_BURS          0
#define RSZVL_VAL_WIMAX_SYMB          1
#define RSZVL_VAL_WIMAX_ZONE          2

#define RSZVL_VAL_WIMAX_MEAS_RMS       0
#define RSZVL_VAL_WIMAX_MEAS_CREST     1
#define RSZVL_VAL_WIMAX_MEAS_RMS_ALL   2
#define RSZVL_VAL_WIMAX_MEAS_RMS_PILOT 3
#define RSZVL_VAL_WIMAX_MEAS_RMS_DATA  4
#define RSZVL_VAL_WIMAX_MEAS_RMS_DLPR  5

#define RSZVL_VAL_WIMAX_GATE_TIME    0
#define RSZVL_VAL_WIMAX_GATE_SAMPLE  1

#define RSZVL_VAL_WIMAX_IQ_I         0
#define RSZVL_VAL_WIMAX_IQ_Q         1
#define RSZVL_VAL_WIMAX_IQ_IQ        2

#define RSZVL_VAL_WIMAX_PREA          0
#define RSZVL_VAL_WIMAX_SUBF          1 

/* K8  Bluetooth constants */
#define RSZVL_VAL_BTO_CFS_FERR_INITIAL           0
#define RSZVL_VAL_BTO_CFS_FERR_BLOCK             1
#define RSZVL_VAL_BTO_CFS_FERR_TOTAL             2

#define RSZVL_VAL_BTO_CFS_DEVM_RMS               0
#define RSZVL_VAL_BTO_CFS_DEVM_99PCT             1
#define RSZVL_VAL_BTO_CFS_DEVM_PEAK              2

#define RSZVL_VAL_BTO_DPEN_TESTED                0
#define RSZVL_VAL_BTO_DPEN_BER                   1
#define RSZVL_VAL_BTO_DPEN_PASSED                2

#define RSZVL_VAL_BTO_SE_TX                  0
#define RSZVL_VAL_BTO_SE_HADJ_LOW            1
#define RSZVL_VAL_BTO_SE_HADJ_UPP            2

#define RSZVL_VAL_BTO_RTP_GFSK           0
#define RSZVL_VAL_BTO_RTP_DPSK           1
#define RSZVL_VAL_BTO_RTP_RATIO          2

/* K14 constants */
#define RSZVL_VAL_SPECM_MARK_MAX_PEAK     0
#define RSZVL_VAL_SPECM_MARK_MAX_NEXT     1
#define RSZVL_VAL_SPECM_MARK_MAX_LEFT     2
#define RSZVL_VAL_SPECM_MARK_MAX_RIGHT    3
#define RSZVL_VAL_SPECM_MARK_MIN_PEAK     4
#define RSZVL_VAL_SPECM_MARK_MIN_NEXT     5
#define RSZVL_VAL_SPECM_MARK_MIN_LEFT     6
#define RSZVL_VAL_SPECM_MARK_MIN_RIGHT    7
#define RSZVL_VAL_SPECM_MARK_Y_MAX_PEAK   8
#define RSZVL_VAL_SPECM_MARK_Y_MAX_NEXT   9
#define RSZVL_VAL_SPECM_MARK_Y_MAX_ABOVE 10
#define RSZVL_VAL_SPECM_MARK_Y_MAX_BELOW 11
#define RSZVL_VAL_SPECM_MARK_Y_MIN_PEAK  12
#define RSZVL_VAL_SPECM_MARK_Y_MIN_NEXT  13
#define RSZVL_VAL_SPECM_MARK_Y_MIN_ABOVE 14
#define RSZVL_VAL_SPECM_MARK_Y_MIN_BELOW 15
#define RSZVL_VAL_SPECM_MARK_XY_MAX      16
#define RSZVL_VAL_SPECM_MARK_XY_MIN      17

#define RSZVL_VAL_SPECM_TSTAMP_CURR   0
#define RSZVL_VAL_SPECM_TSTAMP_ALL   1


/* K72 constants */

#define RSZVL_VAL_CDP_TABLE_OFF          0
#define RSZVL_VAL_CDP_TABLE_SELECT       1
#define RSZVL_VAL_CDP_TABLE_NEW          2
#define RSZVL_VAL_CDP_TABLE_DELETE       3
#define RSZVL_VAL_CDP_TABLE_COPY         4

#define RSZVL_VAL_CDP_TABLE_CHANNEL_DPCH     0
#define RSZVL_VAL_CDP_TABLE_CHANNEL_PICH     1
#define RSZVL_VAL_CDP_TABLE_CHANNEL_SCCPCH   2
#define RSZVL_VAL_CDP_TABLE_CHANNEL_HSSCCH   3
#define RSZVL_VAL_CDP_TABLE_CHANNEL_HSPDSCH  4
#define RSZVL_VAL_CDP_TABLE_CHANNEL_CHAN     5
#define RSZVL_VAL_CDP_TABLE_CHANNEL_CPRSD    10
#define RSZVL_VAL_CDP_TABLE_CHANNEL_CPRTPC   11
#define RSZVL_VAL_CDP_TABLE_CHANNEL_CPRSF    12
#define RSZVL_VAL_CDP_TABLE_CHANNEL_CPRSFTPC 13
#define RSZVL_VAL_CDP_TABLE_CHANNEL_EHICH    14
#define RSZVL_VAL_CDP_TABLE_CHANNEL_EAGCH    15
#define RSZVL_VAL_CDP_TABLE_CHANNEL_SCPICH   16

#define RSZVL_VAL_3GPP_MARKER_CPICH  0
#define RSZVL_VAL_3GPP_MARKER_PCCPCH 1

#define RSZVL_VAL_3GBS_TRAC_1           0
#define RSZVL_VAL_3GBS_TRAC_CTAB        1
#define RSZVL_VAL_3GBS_TRAC_CWCD        2
#define RSZVL_VAL_3GBS_TRAC_PWCD        3

#define RSZVL_VAL_WCDP_RES_PTOT      0
#define RSZVL_VAL_WCDP_RES_FERR      1
#define RSZVL_VAL_WCDP_RES_TFR       2
#define RSZVL_VAL_WCDP_RES_MACC      3
#define RSZVL_VAL_WCDP_RES_PCD       4  
#define RSZVL_VAL_WCDP_RES_EVMR      5  
#define RSZVL_VAL_WCDP_RES_CERR      6  
#define RSZVL_VAL_WCDP_RES_CSL       7  
#define RSZVL_VAL_WCDP_RES_SRAT      8  
#define RSZVL_VAL_WCDP_RES_CHAN      9 
#define RSZVL_VAL_WCDP_RES_CDPA      10 
#define RSZVL_VAL_WCDP_RES_CDPR      11 
#define RSZVL_VAL_WCDP_RES_IQOF      12 
#define RSZVL_VAL_WCDP_RES_IQIM      13 
#define RSZVL_VAL_WCDP_RES_MTYP      14 
#define RSZVL_VAL_WCDP_RES_RHO       15 
#define RSZVL_VAL_WCDP_RES_TOFF      16 

#define RSZVL_VAL_POW_RES_ACP         0
#define RSZVL_VAL_POW_RES_CPOW        1
#define RSZVL_VAL_POW_RES_MCAC        2
#define RSZVL_VAL_POW_RES_OBAN        3

/* End of K74 constants*/

/* K20 constants */

#define RSZVL_VAL_ATV_VCP             0
#define RSZVL_VAL_ATV_VCF             1
#define RSZVL_VAL_ATV_SC1PR           2
#define RSZVL_VAL_ATV_SC1IF           3
#define RSZVL_VAL_ATV_SC2PR           4
#define RSZVL_VAL_ATV_SC2IF           5

#define RSZVL_VAL_LIM_LOW             0
#define RSZVL_VAL_LIM_UPP             1

#define RSZVL_VAL_ATV_RPC             1
#define RSZVL_VAL_ATV_MDEP            2

#define RSZVL_VAL_STAT_DEFAULT        0
#define RSZVL_VAL_STAT_ADJUST         1
#define RSZVL_VAL_STAT_XRLEV          2
#define RSZVL_VAL_STAT_XRANGE         3
#define RSZVL_VAL_STAT_YMAX           4
#define RSZVL_VAL_STAT_YMIN           5

#define RSZVL_VAL_DTV_MERR            0
#define RSZVL_VAL_DTV_MERP            1
#define RSZVL_VAL_DTV_EVMR            2
#define RSZVL_VAL_DTV_EVMP            3
#define RSZVL_VAL_DTV_CFOF            4
#define RSZVL_VAL_DTV_SROF            5
#define RSZVL_VAL_DTV_IMB             4
#define RSZVL_VAL_DTV_QERR            5
#define RSZVL_VAL_DTV_SUP             6
#define RSZVL_VAL_DTV_PJIT            7

#define RSZVL_VAL_ATV_RES_RPOW        0
#define RSZVL_VAL_ATV_RES_NFC         1
#define RSZVL_VAL_ATV_RES_MFR         2
#define RSZVL_VAL_ATV_RES_CR          3
#define RSZVL_VAL_ATV_RES_NRBW        4

#define RSZVL_VAL_DTV_SPEC_SAL        0
#define RSZVL_VAL_DTV_SPEC_SAUP       1

#define RSZVL_VAL_STAT_MEAN           0
#define RSZVL_VAL_STAT_PEAK           1
#define RSZVL_VAL_STAT_CFAC           2

/* End of K20 constants */

#define RSZVL_VAL_DISP_TRAC_DATA      0
#define RSZVL_VAL_DISP_TRAC_MEM       1
#define RSZVL_VAL_DISP_TRAC_SING      2

#define RSZVL_VAL_POIN                0
#define RSZVL_VAL_COMM                1

#define RSZVL_VAL_SEM                 0
#define RSZVL_VAL_TAB                 2
#define RSZVL_VAL_SPAC                3

#define RSZVL_LIMIT_OFF               0
#define RSZVL_LIMIT_ON                1

#define RSZVL_SAVE                    0
#define RSZVL_RECALL                  1

#define RSZVL_VAL_AUTO_CALTYPE_FOP    0
#define RSZVL_VAL_AUTO_CALTYPE_OPTP   1
#define RSZVL_VAL_AUTO_CALTYPE_FNP    2
#define RSZVL_VAL_AUTO_CALTYPE_SFTP   3

/****************************************************************************
 *---------------- Instrument Driver Function Declarations -----------------*
 ****************************************************************************/

ViStatus _VI_FUNC rszvl_init (ViRsrc resourceName, ViBoolean IDQuery,
                              ViBoolean resetDevice, ViPSession instrumentHandle);
ViStatus _VI_FUNC rszvl_InitWithOptions (ViRsrc resourceName, ViBoolean IDQuery,
                                         ViBoolean resetDevice,
                                         ViString optionString,
                                         ViPSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureInstrumentMode (ViSession instrumentHandle,
                                                 ViInt32 instrumentMode);
ViStatus _VI_FUNC rszvl_QueryInstrumentMode (ViSession instrumentHandle,
                                             ViPInt32 instrumentMode);
ViStatus _VI_FUNC rszvl_ConfigureAcquisition (ViSession instrumentHandle,
                                              ViBoolean sweepModeContinuous,
                                              ViInt32 numberOfSweeps);
ViStatus _VI_FUNC rszvl_ConfigureAverage (ViSession instrumentHandle,
                                          ViInt32 channel, ViBoolean averageState,
                                          ViInt32 averageFactor);
ViStatus _VI_FUNC rszvl_RestartAverage (ViSession instrumentHandle,
                                        ViInt32 channel);
ViStatus _VI_FUNC rszvl_ConfigureBandwidth (ViSession instrumentHandle,
                                            ViInt32 channel,
                                            ViReal64 measBandwidth);
ViStatus _VI_FUNC rszvl_ConfigureSweepPoints (ViSession instrumentHandle,
                                              ViInt32 channel,
                                              ViInt32 numberOfPoints);
ViStatus _VI_FUNC rszvl_ConfigureActiveChannel (ViSession instrumentHandle,
                                                ViInt32 channel);
ViStatus _VI_FUNC rszvl_ChannelAdd (ViSession instrumentHandle, ViInt32 channel,
                                    ViString channelName);
ViStatus _VI_FUNC rszvl_ChannelAddTrace (ViSession instrumentHandle,
                                         ViInt32 diagramArea, ViInt32 window_Trace,
                                         ViInt32 channel, ViString channelName,
                                         ViString traceName);
ViStatus _VI_FUNC rszvl_ChannelAddTraceDiagramArea (ViSession instrumentHandle,
                                                    ViInt32 diagramArea,
                                                    ViInt32 window_Trace,
                                                    ViInt32 channel,
                                                    ViString channelName,
                                                    ViString traceName);
ViStatus _VI_FUNC rszvl_ChannelDelete (ViSession instrumentHandle, ViInt32 channel);
ViStatus _VI_FUNC rszvl_ChannelRename (ViSession instrumentHandle, ViInt32 channel,
                                       ViString channelName);
ViStatus _VI_FUNC rszvl_GetChannelList (ViSession instrumentHandle,
                                        ViInt32 bufferSize,
                                        ViChar _VI_FAR catalog[]);
ViStatus _VI_FUNC rszvl_GetChannelName (ViSession instrumentHandle, ViInt32 channel,
                                        ViChar _VI_FAR channelName[]);
ViStatus _VI_FUNC rszvl_GetChannelNumber (ViSession instrumentHandle,
                                          ViString channelName,
                                          ViPInt32 channelNumber);
ViStatus _VI_FUNC rszvl_ConfigureFrequencyStartStop (ViSession instrumentHandle,
                                                     ViReal64 startFrequency,
                                                     ViReal64 stopFrequency);
ViStatus _VI_FUNC rszvl_ConfigureFrequencyCenter (ViSession instrumentHandle,
                                                  ViReal64 centerFrequency);
ViStatus _VI_FUNC rszvl_ConfigureFrequencyCenterSpan (ViSession instrumentHandle,
                                                      ViReal64 centerFrequency,
                                                      ViReal64 span);
ViStatus _VI_FUNC rszvl_QueryFrequency (ViSession instrumentHandle,
                                        ViInt32 frequency, ViPReal64 value);
ViStatus _VI_FUNC rszvl_SetReceiverStepAttenuators (ViSession instrumentHandle,
                                                    ViInt32 channel,
                                                    ViInt32 analyzerPort,
                                                    ViReal64 attenuationFactor);
ViStatus _VI_FUNC rszvl_GetReceiverStepAttenuators (ViSession instrumentHandle,
                                                    ViInt32 channel,
                                                    ViInt32 analyzerPort,
                                                    ViPReal64 attenuationFactor);
ViStatus _VI_FUNC rszvl_ConfigureSourcePort (ViSession instrumentHandle,
                                             ViInt32 channel, ViInt32 sourcePort);
ViStatus _VI_FUNC rszvl_QuerySourcePort (ViSession instrumentHandle,
                                         ViInt32 channel, ViPInt32 sourcePort);
ViStatus _VI_FUNC rszvl_ConfigureRFOutput (ViSession instrumentHandle,
                                           ViBoolean RFOutput);
ViStatus _VI_FUNC rszvl_ConfigureSweepType (ViSession instrumentHandle,
                                            ViInt32 channel, ViInt32 sweepType);
ViStatus _VI_FUNC rszvl_QuerySweepType (ViSession instrumentHandle, ViInt32 channel,
                                        ViPInt32 sweepType);
ViStatus _VI_FUNC rszvl_QuerySweepPoints (ViSession instrumentHandle,
                                          ViInt32 channel, ViPInt32 numberOfPoints);
ViStatus _VI_FUNC rszvl_ConfigureSweepMeasDelay (ViSession instrumentHandle,
                                                 ViInt32 channel,
                                                 ViReal64 measDelay);
ViStatus _VI_FUNC rszvl_QuerySweepMeasDelay (ViSession instrumentHandle,
                                             ViInt32 channel, ViPReal64 measDelay);
ViStatus _VI_FUNC rszvl_ConfigureFrequencyStepSize (ViSession instrumentHandle,
                                                    ViInt32 channel,
                                                    ViReal64 stepSize);
ViStatus _VI_FUNC rszvl_QueryFrequencyStepSize (ViSession instrumentHandle,
                                                ViInt32 channel,
                                                ViPReal64 stepSize);
ViStatus _VI_FUNC rszvl_ConfigureSweepSegment (ViSession instrumentHandle,
                                               ViInt32 channel, ViInt32 segment,
                                               ViInt32 operation,
                                               ViReal64 startFrequency,
                                               ViReal64 stopFrequency,
                                               ViInt32 numberOfPoints,
                                               ViReal64 power,
                                               ViInt32 sweepTimeSelect,
                                               ViReal64 time, ViReal64 pointDelay,
                                               ViReal64 measBandwidth);
ViStatus _VI_FUNC rszvl_ConfigureSweepSegmentState (ViSession instrumentHandle,
                                                    ViInt32 channel,
                                                    ViInt32 segment,
                                                    ViBoolean state);
ViStatus _VI_FUNC rszvl_ConfigureSweepSegmentStartFrequency
             (ViSession instrumentHandle, ViInt32 channel, ViInt32 segment,
              ViReal64 startFrequency);
ViStatus _VI_FUNC rszvl_ConfigureSweepSegmentStopFrequency
             (ViSession instrumentHandle, ViInt32 channel, ViInt32 segment,
              ViReal64 stopFrequency);
ViStatus _VI_FUNC rszvl_ConfigureSweepSegmentIndividualSettings
             (ViSession instrumentHandle, ViInt32 channel, ViInt32 segment,
              ViBoolean power, ViBoolean measBandwidth, ViBoolean selectivity,
              ViBoolean time, ViBoolean pointDelay);
ViStatus _VI_FUNC rszvl_ConfigureSweepSegmentMeasBandwidth
             (ViSession instrumentHandle, ViInt32 channel, ViInt32 segment,
              ViReal64 measBandwidth);
ViStatus _VI_FUNC rszvl_ConfigureSegmentAttenuatorFactors
             (ViSession instrumentHandle, ViInt32 channel, ViInt32 segment,
              ViInt32 port, ViReal64 attenuationFactor);
ViStatus _VI_FUNC rszvl_ConfigureSweepSegmentSelectivity (ViSession instrumentHandle,
                                                          ViInt32 channel,
                                                          ViInt32 segment,
                                                          ViInt32 selectivity);
ViStatus _VI_FUNC rszvl_ConfigureSweepSegmentSweepTime (ViSession instrumentHandle,
                                                        ViInt32 channel,
                                                        ViInt32 segment,
                                                        ViReal64 time);
ViStatus _VI_FUNC rszvl_ConfigureSweepSegmentPointDelay (ViSession instrumentHandle,
                                                         ViInt32 channel,
                                                         ViInt32 segment,
                                                         ViReal64 pointDelay);
ViStatus _VI_FUNC rszvl_ConfigureSweepSegmentPoints (ViSession instrumentHandle,
                                                     ViInt32 channel,
                                                     ViInt32 segment,
                                                     ViInt32 numberOfPoints);
ViStatus _VI_FUNC rszvl_ConfigureSweepSegmentPower (ViSession instrumentHandle,
                                                    ViInt32 channel,
                                                    ViInt32 segment,
                                                    ViReal64 power);
ViStatus _VI_FUNC rszvl_GetSweepSegmentsCount (ViSession instrumentHandle,
                                               ViInt32 channel, ViPInt32 count);
ViStatus _VI_FUNC rszvl_DeleteSelectedSegment (ViSession instrumentHandle,
                                               ViInt32 channel, ViInt32 segment);
ViStatus _VI_FUNC rszvl_DeleteAllSegments (ViSession instrumentHandle,
                                           ViInt32 channel);
ViStatus _VI_FUNC rszvl_GetSweepSegmentCenterFrequency (ViSession instrumentHandle,
                                                        ViInt32 channel,
                                                        ViInt32 segment,
                                                        ViPReal64 centerFrequency);
ViStatus _VI_FUNC rszvl_GetSweepSegmentFrequencySpan (ViSession instrumentHandle,
                                                      ViInt32 channel,
                                                      ViInt32 segment,
                                                      ViPReal64 frequencySpan);
ViStatus _VI_FUNC rszvl_SaveSegment (ViSession instrumentHandle, ViInt32 channel,
                                     ViString fileName);
ViStatus _VI_FUNC rszvl_LoadSegment (ViSession instrumentHandle, ViInt32 channel,
                                     ViString fileName);
ViStatus _VI_FUNC rszvl_SetCorrectionState (ViSession instrumentHandle,
                                            ViInt32 channel,
                                            ViBoolean correctionState);
ViStatus _VI_FUNC rszvl_CorrectionManager (ViSession instrumentHandle,
                                           ViInt32 channel,
                                           ViInt32 operation_to_bePerformed,
                                           ViString fileName);
ViStatus _VI_FUNC rszvl_ConfigureCalibrationDataCurrentState
             (ViSession instrumentHandle, ViInt32 channel, ViBoolean keepMeasData);
ViStatus _VI_FUNC rszvl_ConfigureCalibrationDataDefaultState
             (ViSession instrumentHandle, ViInt32 channel,
              ViBoolean keepMeasDataDefault);
ViStatus _VI_FUNC rszvl_ExpCalDataTouchstoneFile (ViSession instrumentHandle,
                                                  ViString directoryName);
ViStatus _VI_FUNC rszvl_ConfigurePortConnector (ViSession instrumentHandle,
                                                ViInt32 channel, ViInt32 port,
                                                ViInt32 connector);
ViStatus _VI_FUNC rszvl_ConfigureUserPortConnector (ViSession instrumentHandle,
                                                    ViInt32 channel, ViInt32 port,
                                                    ViString connector,
                                                    ViInt32 connectorGender);
ViStatus _VI_FUNC rszvl_ConfigureIndependentPorts (ViSession instrumentHandle,
                                                   ViInt32 channel,
                                                   ViBoolean sameConnectorAtAllPorts);
ViStatus _VI_FUNC rszvl_ConfigureUserCalibrationConnector
             (ViSession instrumentHandle, ViInt32 channel, ViString connectorName,
              ViInt32 propagationMode, ViInt32 connectorType,
              ViReal64 relativePermittivity, ViReal64 impedance);
ViStatus _VI_FUNC rszvl_GetUserCalibrationConnector (ViSession instrumentHandle,
                                                     ViInt32 channel,
                                                     ViString connectorName,
                                                     ViPInt32 propagationMode,
                                                     ViPInt32 connectorType,
                                                     ViPReal64 relativePermittivity,
                                                     ViPReal64 impedance);
ViStatus _VI_FUNC rszvl_DeleteUserCalibrationConnector (ViSession instrumentHandle,
                                                        ViInt32 channel,
                                                        ViString connectorName);
ViStatus _VI_FUNC rszvl_GetCalibrationState (ViSession instrumentHandle,
                                             ViInt32 channel,
                                             ViInt32 *calibrationState);
ViStatus _VI_FUNC rszvl_GetCalibrationDataParameters (ViSession instrumentHandle,
                                                      ViInt32 channel,
                                                      ViPReal64 frequencyStart,
                                                      ViPReal64 frequencyStop,
                                                      ViPInt32 numberOfPoints,
                                                      ViPReal64 internalSignalSourcePower,
                                                      ViPInt32 sweepType);
ViStatus _VI_FUNC rszvl_GetCalibrationDate (ViSession instrumentHandle,
                                            ViInt32 channel, ViInt32 bufferSize,
                                            ViChar _VI_FAR calibrationDate[]);
ViStatus _VI_FUNC rszvl_ConfigureFactoryCalibrationState (ViSession instrumentHandle,
                                                          ViInt32 channel,
                                                          ViBoolean factoryCalibration);
ViStatus _VI_FUNC rszvl_CalibrationAutoSimplified (ViSession instrumentHandle,
                                                   ViInt32 channel,
                                                   ViString calibrationKitName,
                                                   ViInt32 analyzerPort1,
                                                   ViInt32 analyzerPort2,
                                                   ViInt32 timeout);
ViStatus _VI_FUNC rszvl_CalibrationAuto (ViSession instrumentHandle,
                                         ViInt32 channel,
                                         ViString calibrationKitName,
                                         ViInt32 analyzerPort1,
                                         ViInt32 calUnitPort1,
                                         ViInt32 analyzerPort2,
                                         ViInt32 calUnitPort2, ViInt32 timeout);
ViStatus _VI_FUNC rszvl_GenerateCalibrationKitFile (ViSession instrumentHandle,
                                                    ViString fileName);
ViStatus _VI_FUNC rszvl_GetCalibrationPortAssignment (ViSession instrumentHandle,
                                                      ViInt32 channel,
                                                      ViInt32 *analyzerPort1,
                                                      ViInt32 *analyzerPort2);
ViStatus _VI_FUNC rszvl_SelectActiveCalibrationUnit (ViSession instrumentHandle,
                                                  ViString calibrationUnit);
ViStatus _VI_FUNC rszvl_GetAllCalibrationUnits (ViSession instrumentHandle,
                                                ViInt32 bufferSize,
                                                ViChar calibrationUnit[]);
ViStatus _VI_FUNC rszvl_ConfigureCalibrationUnitStandard (ViSession instrumentHandle,
                                                          ViInt32 standard,
                                                          ViInt32 port1,
                                                          ViInt32 port2);
ViStatus _VI_FUNC rszvl_ExportCalibrationDataToTouchstoneFormat (ViSession instrumentHandle,
                                                                 ViString directoryName);
ViStatus _VI_FUNC rszvl_ConfigureCalibrationType (ViSession instrumentHandle,
                                                  ViInt32 channel,
                                                  ViString calibrationName,
                                                  ViInt32 parameters, ViInt32 port1,
                                                  ViInt32 port2);
ViStatus _VI_FUNC rszvl_StartCalibration (ViSession instrumentHandle,
                                          ViInt32 channel, ViInt32 standard,
                                          ViInt32 port1, ViInt32 port2);
ViStatus _VI_FUNC rszvl_SaveCalibrationData (ViSession instrumentHandle,
                                             ViInt32 channel);
ViStatus _VI_FUNC rszvl_GenerateDefaultCalibrationData (ViSession instrumentHandle,
                                                        ViInt32 channel);
ViStatus _VI_FUNC rszvl_DeleteCalibrationData (ViSession instrumentHandle,
                                               ViInt32 channel,
                                               ViString calibrationName);
ViStatus _VI_FUNC rszvl_ReadCalibrationData (ViSession instrumentHandle,
                                             ViInt32 channel,
                                             ViInt32 errorTermParameters,
                                             ViInt32 port1, ViInt32 port2,
                                             ViReal64 _VI_FAR calibrationData[]);
ViStatus _VI_FUNC rszvl_WriteCalibrationData (ViSession instrumentHandle,
                                              ViInt32 channel,
                                              ViInt32 errorTermParameters,
                                              ViInt32 port1, ViInt32 port2,
                                              ViReal64 _VI_FAR calibrationData[]);
ViStatus _VI_FUNC rszvl_ConfigureCalibrationKit (ViSession instrumentHandle,
                                                 ViInt32 connector,
                                                 ViString calibrationKitName);
ViStatus _VI_FUNC rszvl_GetCalibrationKit (ViSession instrumentHandle,
                                           ViInt32 connector, ViInt32 bufferSize,
                                           ViChar _VI_FAR calibrationKitName[]);
ViStatus _VI_FUNC rszvl_ConfigureCalibrationKitUserConnectorType
             (ViSession instrumentHandle, ViString connector,
              ViString calibrationKitName);
ViStatus _VI_FUNC rszvl_GetCalibrationKitUserConnectorType
             (ViSession instrumentHandle, ViString connector, ViInt32 bufferSize,
              ViChar _VI_FAR calibrationKitName[]);
ViStatus _VI_FUNC rszvl_ConfigureCalibrationStandard (ViSession instrumentHandle,
                                                      ViInt32 connector,
                                                      ViInt32 standard,
                                                      ViString kit,
                                                      ViString serialNumber,
                                                      ViReal64 minFreqHz,
                                                      ViReal64 maxFreqHz,
                                                      ViReal64 lengthmm,
                                                      ViReal64 loss, ViReal64 c0L0,
                                                      ViReal64 c1L1, ViReal64 c2L2,
                                                      ViReal64 c3L3,
                                                      ViInt32 approximation);
ViStatus _VI_FUNC rszvl_LoadCalibrationKit (ViSession instrumentHandle,
                                            ViString fileName);
ViStatus _VI_FUNC rszvl_LoadCalibrationKitFromTouchstoneFile
             (ViSession instrumentHandle, ViString connectorName,
              ViString calibrationKitName, ViInt32 standard,
              ViString calibrationKitLabel, ViString fileName, ViInt32 portNumber1,
              ViInt32 portNumber2);
ViStatus _VI_FUNC rszvl_SaveCalibrationKit (ViSession instrumentHandle,
                                            ViString kitName, ViString fileName);
ViStatus _VI_FUNC rszvl_ConfigureLength (ViSession instrumentHandle,
                                         ViInt32 channel, ViInt32 port,
                                         ViReal64 electricalLength,
                                         ViReal64 mechanicalLength,
                                         ViReal64 permittivity, ViReal64 delay);
ViStatus _VI_FUNC rszvl_AutoLength (ViSession instrumentHandle, ViInt32 channel,
                                    ViInt32 port);
ViStatus _VI_FUNC rszvl_ConfigureLoss (ViSession instrumentHandle, ViInt32 channel,
                                       ViInt32 port, ViReal64 lossAtDC,
                                       ViReal64 lossAtFrequency,
                                       ViReal64 lossReferenceFrequency);
ViStatus _VI_FUNC rszvl_AutoLengthAndLoss (ViSession instrumentHandle,
                                           ViInt32 channel, ViInt32 port);
ViStatus _VI_FUNC rszvl_ResetOffsets (ViSession instrumentHandle, ViInt32 channel);
ViStatus _VI_FUNC rszvl_ConfigureSourcePower (ViSession instrumentHandle,
                                              ViInt32 channel,
                                              ViReal64 sourcePower);
ViStatus _VI_FUNC rszvl_TraceSelect (ViSession instrumentHandle, ViInt32 channel,
                                     ViString traceName);
ViStatus _VI_FUNC rszvl_TraceAdd (ViSession instrumentHandle, ViInt32 channel,
                                  ViString traceName);
ViStatus _VI_FUNC rszvl_TraceAddDiagramArea (ViSession instrumentHandle,
                                             ViInt32 diagramArea,
                                             ViInt32 window_Trace, ViInt32 channel,
                                             ViString traceName);
ViStatus _VI_FUNC rszvl_TraceAddDiagramAreaByName (ViSession instrumentHandle,
                                                   ViInt32 diagramArea,
                                                   ViInt32 channel,
                                                   ViString traceName);
ViStatus _VI_FUNC rszvl_TraceDelete (ViSession instrumentHandle, ViInt32 channel,
                                     ViString traceName);
ViStatus _VI_FUNC rszvl_TraceList (ViSession instrumentHandle, ViInt32 channel,
                                   ViInt32 bufferSize, ViChar _VI_FAR catalog[]);
ViStatus _VI_FUNC rszvl_TraceAssignDiagramArea (ViSession instrumentHandle,
                                                ViInt32 diagramArea,
                                                ViInt32 window_Trace,
                                                ViString traceName);
ViStatus _VI_FUNC rszvl_TraceUnassignDiagramArea (ViSession instrumentHandle,
                                                  ViInt32 diagramArea,
                                                  ViInt32 window_Trace);
ViStatus _VI_FUNC rszvl_TraceShiftStimulusValue (ViSession instrumentHandle,
                                                 ViInt32 diagramArea,
                                                 ViInt32 window_Trace,
                                                 ViReal64 shiftStimulusValue);
ViStatus _VI_FUNC rszvl_TraceShiftResponseValue (ViSession instrumentHandle,
                                                 ViInt32 diagramArea,
                                                 ViInt32 window_Trace,
                                                 ViReal64 magnitude, ViReal64 phase,
                                                 ViReal64 real, ViReal64 imaginary);
ViStatus _VI_FUNC rszvl_ConfigureTraceFormat (ViSession instrumentHandle,
                                              ViInt32 channel_Trace,
                                              ViInt32 format);
ViStatus _VI_FUNC rszvl_ConfigureTraceAperture (ViSession instrumentHandle,
                                                ViInt32 channel_Trace,
                                                ViInt32 aperture);
ViStatus _VI_FUNC rszvl_ConfigureTraceMathFunction (ViSession instrumentHandle,
                                                    ViInt32 channel_Trace,
                                                    ViInt32 mathematicalFunction);
ViStatus _VI_FUNC rszvl_ConfigureTraceSmoothing (ViSession instrumentHandle,
                                                 ViInt32 channel_Trace,
                                                 ViBoolean smoothing,
                                                 ViReal64 aperture);
ViStatus _VI_FUNC rszvl_ConfigureTraceMaxHold (ViSession instrumentHandle,
                                               ViInt32 channel_Trace,
                                               ViInt32 maxHoldState);
ViStatus _VI_FUNC rszvl_TraceDataToMemory (ViSession instrumentHandle,
                                           ViInt32 channel_Trace);
ViStatus _VI_FUNC rszvl_TraceDataToMemoryTrace (ViSession instrumentHandle,
                                                ViString memoryTrace,
                                                ViString dataTrace);
ViStatus _VI_FUNC rszvl_TraceMathToMemoryTrace (ViSession instrumentHandle,
                                                ViString memoryTrace,
                                                ViString dataTrace);
ViStatus _VI_FUNC rszvl_DeleteMemoryTrace (ViSession instrumentHandle,
                                           ViInt32 memoryTrace);
ViStatus _VI_FUNC rszvl_TraceListCatalog (ViSession instrumentHandle,
                                          ViInt32 bufferSize,
                                          ViChar _VI_FAR catalog[]);
ViStatus _VI_FUNC rszvl_TraceDisplayState (ViSession instrumentHandle,
                                           ViInt32 traceType,
                                           ViString singleTraceName,
                                           ViBoolean showTrace);
ViStatus _VI_FUNC rszvl_QueryTraceDisplayState (ViSession instrumentHandle,
                                                ViInt32 traceType,
                                                ViString singleTraceName,
                                                ViBoolean *showTrace);
ViStatus _VI_FUNC rszvl_TraceAutoscale (ViSession instrumentHandle,
                                        ViInt32 diagramArea, ViInt32 window_Trace);
ViStatus _VI_FUNC rszvl_TraceAutoscaleByName (ViSession instrumentHandle,
                                              ViString traceName);
ViStatus _VI_FUNC rszvl_ConfigureTraceScale (ViSession instrumentHandle,
                                             ViInt32 diagramArea,
                                             ViInt32 window_Trace, ViReal64 bottom,
                                             ViReal64 top);
ViStatus _VI_FUNC rszvl_ConfigureTraceScaleDivision (ViSession instrumentHandle,
                                                     ViInt32 diagramArea,
                                                     ViInt32 window_Trace,
                                                     ViReal64 scaleDivisions);
ViStatus _VI_FUNC rszvl_ConfigureTraceReferenceValue (ViSession instrumentHandle,
                                                      ViInt32 diagramArea,
                                                      ViInt32 window_Trace,
                                                      ViReal64 referenceLevel);
ViStatus _VI_FUNC rszvl_ConfigureTraceReferencePosition (ViSession instrumentHandle,
                                                         ViInt32 diagramArea,
                                                         ViInt32 window_Trace,
                                                         ViReal64 referencePosition);
ViStatus _VI_FUNC rszvl_ConfigureTraceEvaluationRange (ViSession instrumentHandle,
                                                       ViInt32 channel_Trace,
                                                       ViInt32 evaluationRange,
                                                       ViReal64 start,
                                                       ViReal64 stop);
ViStatus _VI_FUNC rszvl_ConfigureDisplayResultsState (ViSession instrumentHandle,
                                                      ViInt32 channel_Trace,
                                                      ViInt32 resultType,
                                                      ViBoolean displayResults);
ViStatus _VI_FUNC rszvl_ConfigureTraceCustomHeader (ViSession instrumentHandle,
                                           ViBoolean showCustomHeader,
                                           ViBoolean showInstrumentInfo,
                                           ViBoolean showDateAndTimeInfo,
                                           ViBoolean showText,
                                           ViString customHeaderText,
                                           ViBoolean showCalibrationInfo);
ViStatus _VI_FUNC rszvl_TraceImportData (ViSession instrumentHandle,
                                         ViString traceName, ViString fileName);
ViStatus _VI_FUNC rszvl_TraceExportData (ViSession instrumentHandle,
                                         ViString traceName, ViString fileName);
ViStatus _VI_FUNC rszvl_TraceExportDataWithOptions (ViSession instrumentHandle,
                                                    ViString traceName,
                                                    ViString fileName,
                                                    ViInt32 exportFormat,
                                                    ViInt32 exportData);
ViStatus _VI_FUNC rszvl_TraceExportDataWithOptionsExtended (ViSession instrumentHandle,
                                                            ViString traceName,
                                                            ViString fileName,
                                                            ViInt32 exportFormat,
                                                            ViInt32 exportData,
                                                            ViInt32 decimalSeparator,
                                                            ViInt32 fieldSeparator);
ViStatus _VI_FUNC rszvl_ConfigureMarkerState (ViSession instrumentHandle,
                                              ViInt32 channel_Trace, ViInt32 marker,
                                              ViBoolean markerState);
ViStatus _VI_FUNC rszvl_ConfigureMarkerStimulus (ViSession instrumentHandle,
                                                 ViInt32 channel_Trace,
                                                 ViInt32 marker,
                                                 ViReal64 markerStimulus);
ViStatus _VI_FUNC rszvl_ConfigureReferenceMarkerState (ViSession instrumentHandle,
                                                       ViInt32 channel_Trace,
                                                       ViInt32 marker,
                                                       ViBoolean referenceMarkerState);
ViStatus _VI_FUNC rszvl_ConfigureReferenceMarkerStimulus (ViSession instrumentHandle,
                                                          ViInt32 channel_Trace,
                                                          ViInt32 marker,
                                                          ViReal64 referenceMarkerStimulus);

ViStatus _VI_FUNC rszvl_ConfigureReferenceDiscreteMarker (ViSession instrumentHandle,
                                                              ViInt32 channel_Trace,
                                                              ViInt32 marker,
                                                              ViInt32 discreteMode);

ViStatus _VI_FUNC rszvl_ConfigureReferenceFixedMarker (ViSession instrumentHandle,
                                                       ViInt32 channel_Trace,
                                                       ViInt32 marker,
                                                       ViInt32 fixedMarker);

ViStatus _VI_FUNC rszvl_ConfigureDeltaMarkerState (ViSession instrumentHandle,
                                                   ViInt32 channel_Trace,
                                                   ViInt32 marker,
                                                   ViBoolean deltaMarkerState);
ViStatus _VI_FUNC rszvl_ConfigureCoupledMarkers (ViSession instrumentHandle,
                                                 ViInt32 channel_Trace,
                                                 ViBoolean markerCoupled);
ViStatus _VI_FUNC rszvl_ConfigureDiscreteMarker (ViSession instrumentHandle,
                                                 ViInt32 channel_Trace,
                                                 ViInt32 marker,
                                                 ViInt32 discreteMode);
ViStatus _VI_FUNC rszvl_ConfigureFixedMarker (ViSession instrumentHandle,
                                              ViInt32 channel_Trace, ViInt32 marker,
                                              ViInt32 fixedMarker);
ViStatus _VI_FUNC rszvl_ConfigureSignalTrack (ViSession instrumentHandle,
                                              ViInt32 channel_Trace, ViInt32 marker,
                                              ViBoolean signalTrackEnabled);
ViStatus _VI_FUNC rszvl_DisableAllMarkers (ViSession instrumentHandle,
                                           ViInt32 channel_Trace);
ViStatus _VI_FUNC rszvl_SetInstrumentFromMarker (ViSession instrumentHandle,
                                                 ViInt32 channel_Trace,
                                                 ViInt32 marker,
                                                 ViInt32 instrumentSetting);
ViStatus _VI_FUNC rszvl_ConfigureMarkerFormat (ViSession instrumentHandle,
                                               ViInt32 channel_Trace,
                                               ViInt32 marker,
                                               ViInt32 markerFormat);
ViStatus _VI_FUNC rszvl_MarkerSearch (ViSession instrumentHandle,
                                      ViInt32 channel_Trace, ViInt32 marker,
                                      ViInt32 search);
ViStatus _VI_FUNC rszvl_MarkerTargetSearch (ViSession instrumentHandle,
                                            ViInt32 channel_Trace, ViInt32 marker,
                                            ViInt32 search);
ViStatus _VI_FUNC rszvl_MarkerBandfilterSearch (ViSession instrumentHandle,
                                                ViInt32 channel_Trace,
                                                ViInt32 bandfilter);
ViStatus _VI_FUNC rszvl_ConfigureMarkerSearchLimits (ViSession instrumentHandle,
                                                     ViInt32 channel_Trace,
                                                     ViInt32 marker,
                                                     ViInt32 searchRange,
                                                     ViReal64 searchRangeStart,
                                                     ViReal64 searchRangeStop);
ViStatus _VI_FUNC rszvl_ConfigureMarkerTargetValue (ViSession instrumentHandle,
                                                    ViInt32 channel_Trace,
                                                    ViInt32 marker,
                                                    ViReal64 targetValue);
ViStatus _VI_FUNC rszvl_ConfigureMarkerxdBBandwidth (ViSession instrumentHandle,
                                                     ViInt32 channel_Trace,
                                                     ViReal64 x_dBBandwidth);
ViStatus _VI_FUNC rszvl_MarkerBandfilterTracking (ViSession instrumentHandle,
                                                  ViInt32 channel_Trace,
                                                  ViBoolean bandfilterTracking);
ViStatus _VI_FUNC rszvl_SetMarkerSearchResultState (ViSession instrumentHandle,
                                                    ViInt32 channel_Trace,
                                                    ViBoolean searchResults);
ViStatus _VI_FUNC rszvl_GetMarkerSearchResultState (ViSession instrumentHandle,
                                                    ViInt32 channel_Trace,
                                                    ViPBoolean searchResults);
ViStatus _VI_FUNC rszvl_SaveAllMarkers (ViSession instrumentHandle,
                                        ViString fileName);
ViStatus _VI_FUNC rszvl_ConfigureLimitLineDisplay (ViSession instrumentHandle,
                                                   ViInt32 channel_Trace,
                                                   ViBoolean displayLimitLine);
ViStatus _VI_FUNC rszvl_ConfigureLimitCheck (ViSession instrumentHandle,
                                             ViInt32 channel_Trace,
                                             ViInt32 limitLine,
                                             ViBoolean limitCheck);
ViStatus _VI_FUNC rszvl_ConfigureLimitResponseDomainComplexUnits
             (ViSession instrumentHandle, ViInt32 channel_Trace,
              ViInt32 responseUnits);
ViStatus _VI_FUNC rszvl_ConfigureLimitResponseDomainFormatUnits
             (ViSession instrumentHandle, ViInt32 channel_Trace,
              ViInt32 responseUnits);
ViStatus _VI_FUNC rszvl_ConfigureLimitResponseDomainSpacingUnits
             (ViSession instrumentHandle, ViInt32 channel_Trace,
              ViInt32 responseUnits);
ViStatus _VI_FUNC rszvl_ConfigureDisplayLineState (ViSession instrumentHandle,
                                                   ViInt32 channel_Trace,
                                                   ViBoolean state);
ViStatus _VI_FUNC rszvl_ConfigureDisplayLinePosition (ViSession instrumentHandle,
                                                      ViInt32 channel_Trace,
                                                      ViReal64 position);
ViStatus _VI_FUNC rszvl_AddLimitLineSegment (ViSession instrumentHandle,
                                             ViInt32 channel_Trace, ViInt32 segment,
                                             ViInt32 type, ViReal64 startStimulus,
                                             ViReal64 stopStimulus,
                                             ViReal64 startResponse,
                                             ViReal64 stopResponse);
ViStatus _VI_FUNC rszvl_EditLimitLineSegment (ViSession instrumentHandle,
                                              ViInt32 channel_Trace,
                                              ViInt32 segment, ViInt32 type,
                                              ViReal64 startStimulus,
                                              ViReal64 stopStimulus,
                                              ViReal64 startResponse,
                                              ViReal64 stopResponse);
ViStatus _VI_FUNC rszvl_ReadLimitLineSegmentList (ViSession instrumentHandle,
                                                  ViInt32 channel_Trace,
                                                  ViInt32 listSize,
                                                  ViPInt32 segmentsCount,
                                                  ViInt32 _VI_FAR type[],
                                                  ViReal64 _VI_FAR startStimulus[],
                                                  ViReal64 _VI_FAR stopStimulus[],
                                                  ViReal64 _VI_FAR startResponse[],
                                                  ViReal64 _VI_FAR stopResponse[]);
ViStatus _VI_FUNC rszvl_WriteLimitLineSegmentList (ViSession instrumentHandle,
                                                   ViInt32 channel_Trace,
                                                   ViInt32 listSize, ViInt32 type,
                                                   ViReal64 _VI_FAR startStimulus[],
                                                   ViReal64 _VI_FAR stopStimulus[],
                                                   ViReal64 _VI_FAR startResponse[],
                                                   ViReal64 _VI_FAR stopResponse[]);
ViStatus _VI_FUNC rszvl_ShiftLimitLineSegmentList (ViSession instrumentHandle,
                                                   ViInt32 channel_Trace,
                                                   ViInt32 limitLineType,
                                                   ViReal64 stimulusOffset,
                                                   ViReal64 responseOffset);
ViStatus _VI_FUNC rszvl_QueryLimitLineSegmentCount (ViSession instrumentHandle,
                                                    ViInt32 channel_Trace,
                                                    ViInt32 *segmentCount);
ViStatus _VI_FUNC rszvl_DeleteLimitLineSegments (ViSession instrumentHandle,
                                                 ViInt32 channel_Trace);
ViStatus _VI_FUNC rszvl_SaveLimitLine (ViSession instrumentHandle,
                                       ViString traceName, ViString fileName);
ViStatus _VI_FUNC rszvl_RecallLimitLine (ViSession instrumentHandle,
                                         ViString traceName, ViString fileName);
ViStatus _VI_FUNC rszvl_LoadLimitLineWithOptions (ViSession instrumentHandle,
                                                  ViString traceName,
                                                  ViString fileName,
                                                  ViString sParameter,
                                                  ViReal64 xOffset,
                                                  ViReal64 yOffset, ViInt32 type);
ViStatus _VI_FUNC rszvl_ImportTraceasLimitLine (ViSession instrumentHandle,
                                                ViInt32 channel_Trace,
                                                ViInt32 limitLineType,
                                                ViReal64 stimulusOffset,
                                                ViReal64 responseOffset,
                                                ViString traceName);
ViStatus _VI_FUNC rszvl_ShowRippleLimits (ViSession instrumentHandle,
                                          ViInt32 channel_Trace,
                                          ViBoolean showLimits);
ViStatus _VI_FUNC rszvl_ConfigureRippleLimitCheck (ViSession instrumentHandle,
                                                   ViInt32 channel_Trace,
                                                   ViBoolean limitCheck);
ViStatus _VI_FUNC rszvl_GetRippleGlobalLimitCheckResult (ViSession instrumentHandle,
                                                         ViInt32 channel_Trace,
                                                         ViInt32 *limitCheckResult);
ViStatus _VI_FUNC rszvl_ConfigureRippleSegmentLimitCheck (ViSession instrumentHandle,
                                                          ViInt32 channel_Trace,
                                                          ViInt32 segment,
                                                          ViBoolean limitCheck);
ViStatus _VI_FUNC rszvl_GetRippleLimitCheckSegmentResult (ViSession instrumentHandle,
                                                          ViInt32 channel_Trace,
                                                          ViInt32 segment,
                                                          ViInt32 *fail,
                                                          ViReal64 *limitCheckResult);
ViStatus _VI_FUNC rszvl_GetRippleNumberOfLimitRanges (ViSession instrumentHandle,
                                                      ViInt32 channel_Trace,
                                                      ViInt32 segment,
                                                      ViInt32 *numberOfRanges);
ViStatus _VI_FUNC rszvl_AddRippleLimitLineRangesSegment (ViSession instrumentHandle,
                                                         ViInt32 channel_Trace,
                                                         ViInt32 numberOfValues,
                                                         ViInt32 type[],
                                                         ViReal64 startStimulus[],
                                                         ViReal64 stopStimulus[],
                                                         ViReal64 limit[]);
ViStatus _VI_FUNC rszvl_EditRippleLimitLineSegment (ViSession instrumentHandle,
                                                    ViInt32 channel_Trace,
                                                    ViInt32 segment,
                                                    ViReal64 startStimulus,
                                                    ViReal64 stopStimulus,
                                                    ViReal64 limit);
ViStatus _VI_FUNC rszvl_DeleteAllRippleLimitRanges (ViSession instrumentHandle,
                                                    ViInt32 channel_Trace);
ViStatus _VI_FUNC rszvl_ConfigureRippleLimitPhysicalUnits (ViSession instrumentHandle,
                                                           ViInt32 channel_Trace,
                                                           ViInt32 physicalUnits);
ViStatus _VI_FUNC rszvl_ConfigureRippleLimitResponseDomainFormatUnits (ViSession instrumentHandle,
                                                                       ViInt32 channel_Trace,
                                                                       ViInt32 responseUnits);
ViStatus _VI_FUNC rszvl_SaveRecallRippleLimit (ViSession instrumentHandle,
                                               ViInt32 operationToBePerformed,
                                               ViString traceName,
                                               ViString fileName);
ViStatus _VI_FUNC rszvl_GetTraceResponseData (ViSession instrumentHandle,
                                              ViInt32 channel_Trace,
                                              ViInt32 dataFormat,
                                              ViPInt32 no_ofValues,
                                              ViReal64 _VI_FAR traceData[]);
ViStatus _VI_FUNC rszvl_GetTraceResponseDataError (ViSession instrumentHandle,
                                                   ViInt32 channel_Trace,
                                                   ViInt32 errorTerm,
                                                   ViPInt32 no_ofValues,
                                                   ViReal64 _VI_FAR traceData[]);
ViStatus _VI_FUNC rszvl_GetTraceResponseSingleSweepData (ViSession instrumentHandle,
                                                         ViInt32 channel_Trace,
                                                         ViInt32 sweepNumber,
                                                         ViPInt32 no_ofValues,
                                                         ViReal64 _VI_FAR traceData[]);
ViStatus _VI_FUNC rszvl_GetTraceStimulusData (ViSession instrumentHandle,
                                              ViInt32 channel_Trace,
                                              ViPInt32 no_ofValues,
                                              ViReal64 _VI_FAR traceData[]);
ViStatus _VI_FUNC rszvl_GetTraceStatisticalEvaluation (ViSession instrumentHandle,
                                                       ViInt32 channel_Trace,
                                                       ViInt32 statisticalParameter,
                                                       ViReal64 _VI_FAR responseValue_s[]);
ViStatus _VI_FUNC rszvl_GetLimitCheckResult (ViSession instrumentHandle,
                                             ViInt32 channel_Trace,
                                             ViPInt32 limitCheckResult);
ViStatus _VI_FUNC rszvl_GetCombinedLimitCheckResult (ViSession instrumentHandle,
                                                     ViPInt32 state);
ViStatus _VI_FUNC rszvl_GetMarkerResponse (ViSession instrumentHandle,
                                           ViInt32 channel_Trace, ViInt32 marker,
                                           ViReal64 _VI_FAR markerResponse[]);
ViStatus _VI_FUNC rszvl_GetReferenceMarkerResponse (ViSession instrumentHandle,
                                                    ViInt32 channel_Trace,
                                                    ViPReal64 referenceMarkerResponse);
ViStatus _VI_FUNC rszvl_QueryMarker (ViSession instrumentHandle,
                                     ViInt32 channel_Trace, ViInt32 marker,
                                     ViPReal64 markerPosition,
                                     ViPReal64 markerAmplitude);
ViStatus _VI_FUNC rszvl_GetMarkerSearchResults (ViSession instrumentHandle,
                                                ViInt32 channel_Trace,
                                                ViPReal64 markerStimulus,
                                                ViReal64 _VI_FAR markerResponse[]);
ViStatus _VI_FUNC rszvl_GetMarkerBandfilterResults (ViSession instrumentHandle,
                                                    ViInt32 channel_Trace,
                                                    ViPReal64 bandwidth,
                                                    ViPReal64 centerStimulus,
                                                    ViPReal64 q, ViPReal64 loss,
                                                    ViPReal64 LBE, ViPReal64 UBE);
ViStatus _VI_FUNC rszvl_Initiate (ViSession instrumentHandle, ViInt32 channel_Trace,
                                  ViUInt32 timeout);
ViStatus _VI_FUNC rszvl_SendSoftwareTrigger (ViSession instrumentHandle,
                                             ViUInt32 timeout);
ViStatus _VI_FUNC rszvl_ConfigureSParameters (ViSession instrumentHandle,
                                              ViInt32 channel, ViString traceName,
                                              ViInt32 outPort, ViInt32 inPort);
ViStatus _VI_FUNC rszvl_ConfigureImpedances (ViSession instrumentHandle,
                                             ViInt32 channel, ViString traceName,
                                             ViInt32 outPort, ViInt32 inPort);
ViStatus _VI_FUNC rszvl_ConfigureAdmitances (ViSession instrumentHandle,
                                             ViInt32 channel, ViString traceName,
                                             ViInt32 outPort, ViInt32 inPort);
ViStatus _VI_FUNC rszvl_ConfigureStabilityFactors (ViSession instrumentHandle,
                                                   ViInt32 channel,
                                                   ViString traceName,
                                                   ViInt32 DUTOut, ViInt32 DUTIn,
                                                   ViInt32 stabilityFactor);
ViStatus _VI_FUNC rszvl_ParametersConversion (ViSession instrumentHandle,
                                              ViInt32 channel_Trace,
                                              ViInt32 conversion);
ViStatus _VI_FUNC rszvl_ConfigureDistanceToFaultState (ViSession instrumentHandle,
                                                       ViInt32 channel,
                                                       ViBoolean state);
ViStatus _VI_FUNC rszvl_ConfigureDistanceToFaultStartStop
             (ViSession instrumentHandle, ViInt32 channel, ViReal64 startDistance,
              ViReal64 stopDistance);
ViStatus _VI_FUNC rszvl_ConfigureDistanceToFaultCenterSpan
             (ViSession instrumentHandle, ViInt32 channel, ViReal64 centerDistance,
              ViReal64 span);
ViStatus _VI_FUNC rszvl_ConfigureDistanceToFaultAutoNumberOfPoints
             (ViSession instrumentHandle, ViInt32 channel);
ViStatus _VI_FUNC rszvl_ConfigureDistanceToFaultAddCable (ViSession instrumentHandle,
                                                          ViInt32 channel,
                                                          ViString cableName,
                                                          ViReal64 permittivity,
                                                          ViReal64 attenuation,
                                                          ViReal64 frequency);
ViStatus _VI_FUNC rszvl_ConfigureDistanceToFaultSelectCable
             (ViSession instrumentHandle, ViInt32 channel, ViString cableName);
ViStatus _VI_FUNC rszvl_ConfigureDistanceToFaultDeleteCable
             (ViSession instrumentHandle, ViInt32 channel, ViString cableName);
ViStatus _VI_FUNC rszvl_SaveDistanceToFaultCable (ViSession instrumentHandle,
                                                  ViString targetDirectory);
ViStatus _VI_FUNC rszvl_LoadDistanceToFaultCable (ViSession instrumentHandle,
                                                  ViString sourceDirectory);
ViStatus _VI_FUNC rszvl_ConfigureFaultLimitCheck (ViSession instrumentHandle,
                                                  ViInt32 channel,
                                                  ViBoolean limitCheck);
ViStatus _VI_FUNC rszvl_ConfigureFaultThreshold (ViSession instrumentHandle,
                                                 ViInt32 channel,
                                                 ViReal64 threshold);
ViStatus _VI_FUNC rszvl_QueryFaultCount (ViSession instrumentHandle,
                                         ViInt32 channel, ViInt32 *peakCount);
ViStatus _VI_FUNC rszvl_QueryFaultData (ViSession instrumentHandle,
                                        ViInt32 channel, ViInt32 faultNumber,
										ViReal64 *distance, ViReal64 *peakResponseValue);
ViStatus _VI_FUNC rszvl_ConfigureDisplay (ViSession instrumentHandle,
                                          ViBoolean frequencyInfo,
                                          ViBoolean markerColor,
                                          ViBoolean traceColor);
ViStatus _VI_FUNC rszvl_ConfigureRGBColor (ViSession instrumentHandle,
                                           ViInt32 element, ViReal64 red,
                                           ViReal64 green, ViReal64 blue,
                                           ViInt32 traceStyle, ViInt32 traceWidth);
ViStatus _VI_FUNC rszvl_ConfigureColorScheme (ViSession instrumentHandle,
                                              ViInt32 colorScheme);
ViStatus _VI_FUNC rszvl_ColorSchemeFileManager (ViSession instrumentHandle,
                                                ViInt32 operation,
                                                ViString fileName);
ViStatus _VI_FUNC rszvl_ConfigureDisplayUpdate (ViSession instrumentHandle,
                                                ViBoolean displayUpdate,
                                                ViBoolean errorDisplay);
ViStatus _VI_FUNC rszvl_ConfigureRemoteDisplay (ViSession instrumentHandle,
                                                ViString title);
ViStatus _VI_FUNC rszvl_DiagramAreaAdd (ViSession instrumentHandle,
                                        ViInt32 diagramArea);
ViStatus _VI_FUNC rszvl_DiagramAreaDelete (ViSession instrumentHandle,
                                           ViInt32 diagramArea);
ViStatus _VI_FUNC rszvl_DiagramAreaMaximize (ViSession instrumentHandle,
                                             ViInt32 diagramArea,
                                             ViInt32 diagramAreaOperation);
ViStatus _VI_FUNC rszvl_DiagramAreaTitle (ViSession instrumentHandle,
                                          ViInt32 diagramArea, ViBoolean title,
                                          ViString titleString);
ViStatus _VI_FUNC rszvl_ConfigureNewSetup (ViSession instrumentHandle,
                                           ViString setupName);
ViStatus _VI_FUNC rszvl_SelectSetup (ViSession instrumentHandle,
                                     ViString setupName);
ViStatus _VI_FUNC rszvl_CloseSetup (ViSession instrumentHandle, ViString setupName);
ViStatus _VI_FUNC rszvl_GetSetupList (ViSession instrumentHandle,
                                      ViInt32 bufferSize, ViChar _VI_FAR catalog[]);
ViStatus _VI_FUNC rszvl_SaveSetup (ViSession instrumentHandle, ViString fileName);
ViStatus _VI_FUNC rszvl_LoadSetup (ViSession instrumentHandle, ViString fileName);
ViStatus _VI_FUNC rszvl_ConfigurePresets (ViSession instrumentHandle,
                                          ViInt32 presetScope);
ViStatus _VI_FUNC rszvl_ConfigureUserDefinedPreset (ViSession instrumentHandle,
                                                    ViBoolean userDefinedPreset,
                                                    ViString userDefinedPresetFile);
ViStatus _VI_FUNC rszvl_ImmediateSettingsUpdate (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureSound (ViSession instrumentHandle,
                                        ViBoolean alarmSounds,
                                        ViBoolean statusSounds);
ViStatus _VI_FUNC rszvl_ConfigureLengthUnit (ViSession instrumentHandle,
                                             ViInt32 lengthUnit);
ViStatus _VI_FUNC rszvl_ConfigureSANFrequencyStartStop (ViSession instrumentHandle,
                                                        ViReal64 startFrequency,
                                                        ViReal64 stopFrequency);
ViStatus _VI_FUNC rszvl_ConfigureSANFrequencyCenter (ViSession instrumentHandle,
                                                     ViReal64 centerFrequency);
ViStatus _VI_FUNC rszvl_ConfigureSANFrequencyCenterSpan (ViSession instrumentHandle,
                                                         ViReal64 centerFrequency,
                                                         ViReal64 span);
ViStatus _VI_FUNC rszvl_ConfigureSANFrequencyStepSize (ViSession instrumentHandle,
                                                       ViReal64 stepSize);
ViStatus _VI_FUNC rszvl_ConfigureSANFrequencyCouplingFactor
             (ViSession instrumentHandle, ViInt32 coupling, ViInt32 couplingFactor);
ViStatus _VI_FUNC rszvl_ConfigureSANFrequencySpanFull (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureSANFrequencyOffset (ViSession instrumentHandle,
                                                     ViReal64 frequencyOffset);
ViStatus _VI_FUNC rszvl_ConfigureSANSweepCouplingAuto (ViSession instrumentHandle,
                                                       ViInt32 sweepCoupling);
ViStatus _VI_FUNC rszvl_ConfigureSANSweepCoupling (ViSession instrumentHandle,
                                                   ViInt32 sweepCoupling,
                                                   ViReal64 couplingValue);
ViStatus _VI_FUNC rszvl_ConfigureSANSweepCouplingAdvanced
             (ViSession instrumentHandle, ViInt32 ratioSelection, ViReal64 value);
ViStatus _VI_FUNC rszvl_ConfigureSANFilterType (ViSession instrumentHandle,
                                                ViInt32 filterType);
ViStatus _VI_FUNC rszvl_ConfigureSANVBWMode (ViSession instrumentHandle,
                                             ViInt32 videoFilterType);
ViStatus _VI_FUNC rszvl_ConfigureSANAveraging (ViSession instrumentHandle,
                                               ViInt32 trace, ViBoolean state);
ViStatus _VI_FUNC rszvl_ConfigureSANAveragingType (ViSession instrumentHandle,
                                                   ViInt32 averagingType);
ViStatus _VI_FUNC rszvl_ConfigureSANAcquisition (ViSession instrumentHandle,
                                                 ViBoolean sweepModeContinuous,
                                                 ViInt32 numberOfSweeps);
ViStatus _VI_FUNC rszvl_ConfigureSANVerticalScale (ViSession instrumentHandle,
                                                   ViInt32 verticalScale);
ViStatus _VI_FUNC rszvl_ConfigureSANSweepPoints (ViSession instrumentHandle,
                                                 ViInt32 sweepPoints);
ViStatus _VI_FUNC rszvl_ConfigureSANAttenuation (ViSession instrumentHandle,
                                                 ViBoolean attenuationAuto,
                                                 ViReal64 attenuation);
ViStatus _VI_FUNC rszvl_ConfigureSANReferenceLevel (ViSession instrumentHandle,
                                                    ViReal64 referenceLevel);
ViStatus _VI_FUNC rszvl_ConfigureSANLevel (ViSession instrumentHandle,
                                           ViInt32 amplitudeUnits,
                                           ViReal64 inputImpedance,
                                           ViReal64 referenceLevel,
                                           ViReal64 referenceLevelOffset,
                                           ViBoolean attenuationAuto,
                                           ViReal64 attenuation);
ViStatus _VI_FUNC rszvl_ConfigureSANTrace (ViSession instrumentHandle,
                                           ViInt32 trace, ViInt32 traceType);
ViStatus _VI_FUNC rszvl_ConfigureSANTraceDetector (ViSession instrumentHandle,
                                                   ViInt32 trace,
                                                   ViBoolean detectorTypeAuto,
                                                   ViInt32 detectorType);
ViStatus _VI_FUNC rszvl_ConfigureSANTraceResetBehavior (ViSession instrumentHandle,
                                                        ViInt32 trace,
                                                        ViBoolean reset_atChange);
ViStatus _VI_FUNC rszvl_SANStoreTraceToFile (ViSession instrumentHandle,
                                             ViInt32 trace, ViString fileName);
ViStatus _VI_FUNC rszvl_SANTraceIQSet (ViSession instrumentHandle,
                                       ViReal64 samplingRate, ViInt32 triggerMode,
                                       ViInt32 triggerSlope,
                                       ViInt32 pretriggerSamples,
                                       ViInt32 no_ofSamples);
ViStatus _VI_FUNC rszvl_ConfigureSANTraceIQSampleRate (ViSession instrumentHandle,
                                                       ViReal64 samplingRate);
ViStatus _VI_FUNC rszvl_ConfigureSANTraceIQDataAcquisition
             (ViSession instrumentHandle, ViBoolean traceState);
ViStatus _VI_FUNC rszvl_ConfigureSANTraceIQDataAveraging (ViSession instrumentHandle,
                                                          ViBoolean averagingState,
                                                          ViInt32 averagingCount);
ViStatus _VI_FUNC rszvl_SANCopyTrace (ViSession instrumentHandle,
                                      ViInt32 destinationTrace,
                                      ViInt32 sourceTrace);
ViStatus _VI_FUNC rszvl_ConfigureSANSubtractTraces (ViSession instrumentHandle,
                                                    ViReal64 position,
                                                    ViBoolean traceMathState);
ViStatus _VI_FUNC rszvl_SANSubtractTraces (ViSession instrumentHandle,
                                           ViInt32 trace2);
ViStatus _VI_FUNC rszvl_ConfigureSANMarker (ViSession instrumentHandle,
                                            ViInt32 marker, ViBoolean markerEnabled,
                                            ViInt32 trace);
ViStatus _VI_FUNC rszvl_ConfigureSANSignalTrack (ViSession instrumentHandle,
                                                 ViBoolean signalTrackEnabled,
                                                 ViReal64 signalTrackBandwidth,
                                                 ViReal64 signalTrackThreshold,
                                                 ViInt32 trace);
ViStatus _VI_FUNC rszvl_SANDisableAllMarkers (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_SANMarkerSearch (ViSession instrumentHandle, ViInt32 marker,
                                         ViInt32 markerSearch);
ViStatus _VI_FUNC rszvl_SANMarkerPeakAuto (ViSession instrumentHandle,
                                           ViInt32 markerSearch,
                                           ViBoolean autoPeak);
ViStatus _VI_FUNC rszvl_SANMoveMarker (ViSession instrumentHandle, ViInt32 marker,
                                       ViReal64 markerPosition);
ViStatus _VI_FUNC rszvl_SetSANFromMarker (ViSession instrumentHandle,
                                          ViInt32 marker,
                                          ViInt32 instrumentSetting);
ViStatus _VI_FUNC rszvl_ConfigureSANMarkerZoom (ViSession instrumentHandle,
                                                ViReal64 markerZoom,
                                                ViUInt32 timeout);
ViStatus _VI_FUNC rszvl_QuerySANMarker (ViSession instrumentHandle, ViInt32 marker,
                                        ViPReal64 markerPosition,
                                        ViPReal64 markerAmplitude);
ViStatus _VI_FUNC rszvl_ConfigureSANMarkerSearchThreshold
             (ViSession instrumentHandle, ViBoolean thresholdState,
              ViReal64 markerThreshold);
ViStatus _VI_FUNC rszvl_ConfigureSANMarkerSearchPeakExcursion
             (ViSession instrumentHandle, ViReal64 peakExcursion);
ViStatus _VI_FUNC rszvl_ConfigureSANMarkerSearchLocalOscillator
             (ViSession instrumentHandle, ViBoolean localOscillatorSupression);
ViStatus _VI_FUNC rszvl_ConfigureSANMarkerSearchLimits (ViSession instrumentHandle,
                                                        ViBoolean searchLimits,
                                                        ViReal64 searchLimitLeft,
                                                        ViReal64 searchLimitRight);
ViStatus _VI_FUNC rszvl_ConfigureSANMarkerDemodulation (ViSession instrumentHandle,
                                                        ViInt32 marker,
                                                        ViBoolean state,
                                                        ViInt32 demodulationType,
                                                        ViReal64 markerStopTime,
                                                        ViBoolean continuousDemodulation);
ViStatus _VI_FUNC rszvl_ConfigureSANMarkerPeakList (ViSession instrumentHandle,
                                                    ViInt32 marker,
                                                    ViInt32 peakListCount,
                                                    ViInt32 peakListSort);
ViStatus _VI_FUNC rszvl_QuerySANMarkerPeakListFound (ViSession instrumentHandle,
                                                     ViInt32 marker,
                                                     ViPInt32 peakListFound);
ViStatus _VI_FUNC rszvl_QuerySANMarkerPeakList (ViSession instrumentHandle,
                                                ViInt32 marker, ViInt32 arraySize,
                                                ViInt32 peakListSelection,
                                                ViReal64 _VI_FAR peakList[]);
ViStatus _VI_FUNC rszvl_ConfigureSANDeltaMarker (ViSession instrumentHandle,
                                                 ViInt32 deltaMarker,
                                                 ViBoolean state, ViInt32 trace);
ViStatus _VI_FUNC rszvl_ConfigureSANDeltaMarkerPosition (ViSession instrumentHandle,
                                                         ViInt32 deltaMarker,
                                                         ViInt32 mode,
                                                         ViReal64 position);
ViStatus _VI_FUNC rszvl_SANDisableAllDeltaMarkers (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_SANDeltaMarkerSearch (ViSession instrumentHandle,
                                              ViInt32 deltaMarker,
                                              ViInt32 markerSearch);
ViStatus _VI_FUNC rszvl_QuerySANDeltaMarker (ViSession instrumentHandle,
                                             ViInt32 deltaMarker, ViInt32 mode,
                                             ViPReal64 position,
                                             ViPReal64 amplitude);
ViStatus _VI_FUNC rszvl_ConfigureSANReferenceFixed (ViSession instrumentHandle,
                                                    ViBoolean fixedReference);
ViStatus _VI_FUNC rszvl_ConfigureSANReferenceFixedPoint (ViSession instrumentHandle,
                                                         ViReal64 refPointLevel,
                                                         ViReal64 refPointLevelOffset,
                                                         ViReal64 refPointFrequencyTime);
ViStatus _VI_FUNC rszvl_SANReferenceFixedPeakSearch (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_SetSANTimeAndDate (ViSession instrumentHandle,
                                           ViInt32 timeDate, ViInt32 hour,
                                           ViInt32 minute, ViInt32 second,
                                           ViInt32 day, ViInt32 month,
                                           ViInt32 year);
ViStatus _VI_FUNC rszvl_GetSANSystemTimeAndDate (ViSession instrumentHandle,
                                                 ViInt32 timeDate, ViPInt32 hour,
                                                 ViPInt32 minute, ViPInt32 second,
                                                 ViPInt32 day, ViPInt32 month,
                                                 ViPInt32 year);
ViStatus _VI_FUNC rszvl_SANSpeakerVolume (ViSession instrumentHandle,
                                          ViReal64 speakerVolume);
ViStatus _VI_FUNC rszvl_ConfigureSANTriggerSource (ViSession instrumentHandle,
                                                   ViInt32 channel,
                                                   ViInt32 triggerSource);
ViStatus _VI_FUNC rszvl_ConfigureSANTriggerEventStart (ViSession instrumentHandle,
                                                       ViInt32 channel,
                                                       ViInt32 triggerEventStart);
ViStatus _VI_FUNC rszvl_ConfigureSANTVTriggerSource (ViSession instrumentHandle,
                                                     ViInt32 triggerSource);
ViStatus _VI_FUNC rszvl_ConfigureSANTrigger (ViSession instrumentHandle,
                                             ViInt32 channel,
                                             ViReal64 triggerDelay,
                                             ViInt32 triggerPolarity);
ViStatus _VI_FUNC rszvl_ConfigureSANVideoTrigger (ViSession instrumentHandle,
                                                  ViReal64 videoTriggerLevel);
ViStatus _VI_FUNC rszvl_ConfigureSANIFPowerTrigger (ViSession instrumentHandle,
                                                    ViReal64 triggerLevel);
ViStatus _VI_FUNC rszvl_ConfigureSANIFPowerTriggerParameters
             (ViSession instrumentHandle, ViReal64 triggerLevel, ViReal64 holdoff,
              ViReal64 hysteresis);
ViStatus _VI_FUNC rszvl_ConfigureSANTimeTrigger (ViSession instrumentHandle,
                                                 ViReal64 timeInterval);
ViStatus _VI_FUNC rszvl_ConfigureSANTVTrigger (ViSession instrumentHandle,
                                            ViInt32 lineSystem,
                                            ViInt32 synchronization,
                                            ViInt32 horizontalSyncLine,
                                            ViInt32 polarity);
ViStatus _VI_FUNC rszvl_ConfigureSANTVTriggerFreeRunState (ViSession instrumentHandle,
                                                  ViBoolean TVFreeRunState);
ViStatus _VI_FUNC rszvl_ConfigureSANExternalGate (ViSession instrumentHandle,
                                                  ViBoolean gating,
                                                  ViInt32 gateSource, ViInt32 mode,
                                                  ViInt32 polarity, ViReal64 delay,
                                                  ViReal64 length);
ViStatus _VI_FUNC rszvl_ConfigureSANPreamplifierState (ViSession instrumentHandle,
                                                       ViBoolean state);
ViStatus _VI_FUNC rszvl_CreateSANLimitLine (ViSession instrumentHandle,
                                            ViInt32 limit, ViString name,
                                            ViInt32 domain, ViString comment,
                                            ViInt32 assign_toTrace,
                                            ViBoolean deleteExistingLimitLine);
ViStatus _VI_FUNC rszvl_ConfigureSANLimitLine (ViSession instrumentHandle,
                                               ViInt32 limit, ViInt32 type,
                                               ViInt32 units,
                                               ViInt32 xAxisInterpolation,
                                               ViInt32 yAxisInterpolation,
                                               ViInt32 xAxisScaling,
                                               ViReal64 margin, ViReal64 threshold);
ViStatus _VI_FUNC rszvl_DefineSANLimitLine (ViSession instrumentHandle,
                                            ViInt32 limit, ViInt32 type,
                                            ViInt32 count, ViReal64 _VI_FAR xAxis[],
                                            ViReal64 _VI_FAR amplitude[]);
ViStatus _VI_FUNC rszvl_SelectSANLimitLine (ViSession instrumentHandle,
                                            ViInt32 limit, ViString name,
                                            ViBoolean limitEnabled);
ViStatus _VI_FUNC rszvl_EnableSANLimitCheck (ViSession instrumentHandle,
                                             ViInt32 limit, ViInt32 type,
                                             ViBoolean limitEnabled,
                                             ViBoolean checkEnabled);
ViStatus _VI_FUNC rszvl_MoveSANLimitLine (ViSession instrumentHandle, ViInt32 limit,
                                          ViInt32 type, ViInt32 method,
                                          ViReal64 value);
ViStatus _VI_FUNC rszvl_CopySANLimitLine (ViSession instrumentHandle, ViInt32 limit,
                                          ViInt32 copyTo);
ViStatus _VI_FUNC rszvl_DeleteSANLimitLine (ViSession instrumentHandle,
                                            ViInt32 limit);
ViStatus _VI_FUNC rszvl_ClearSANLimitLineResults (ViSession instrumentHandle,
                                                  ViInt32 limit);
ViStatus _VI_FUNC rszvl_ConfigureSANDisplayLineState (ViSession instrumentHandle,
                                                      ViInt32 line, ViInt32 type,
                                                      ViBoolean state);
ViStatus _VI_FUNC rszvl_ConfigureSANDisplayLinePosition (ViSession instrumentHandle,
                                                         ViInt32 line, ViInt32 type,
                                                         ViReal64 position);
ViStatus _VI_FUNC rszvl_ConfigureSANDisplayControl (ViSession instrumentHandle,
                                                    ViBoolean frequency,
                                                    ViBoolean time);
ViStatus _VI_FUNC rszvl_ConfigureSANDisplayPowerSave (ViSession instrumentHandle,
                                                      ViBoolean state,
                                                      ViInt32 holdoff);
ViStatus _VI_FUNC rszvl_ConfigureSANDisplayWindowSize (ViSession instrumentHandle,
                                                       ViInt32 windowSize);
ViStatus _VI_FUNC rszvl_ConfigureSANDisplayLogRange (ViSession instrumentHandle,
                                                     ViReal64 range);
ViStatus _VI_FUNC rszvl_ConfigureSANDisplayAmplitudeGridMode
             (ViSession instrumentHandle, ViInt32 yAxisGridMode);
ViStatus _VI_FUNC rszvl_ConfigureSANDisplayReferenceValue
             (ViSession instrumentHandle, ViReal64 referenceValue);
ViStatus _VI_FUNC rszvl_ConfigureSANDisplayReferencePosition
             (ViSession instrumentHandle, ViReal64 referencePosition);
ViStatus _VI_FUNC rszvl_ConfigureSANDisplaySingleSweep (ViSession instrumentHandle,
                                                        ViBoolean state);
ViStatus _VI_FUNC rszvl_ConfigureSANDisplayUpdate (ViSession instrumentHandle,
                                                   ViBoolean displayInRemote);
ViStatus _VI_FUNC rszvl_SANDisplayColorDefault (ViSession instrumentHandle,
                                                ViInt32 defaultSetting);
ViStatus _VI_FUNC rszvl_ConfigureSANDisplayColor (ViSession instrumentHandle,
                                                  ViInt32 colorMap, ViReal64 TINT,
                                                  ViReal64 SATURATION,
                                                  ViReal64 BRIGHTNESS);
ViStatus _VI_FUNC rszvl_ConfigureSANDisplayColorByPredefined
             (ViSession instrumentHandle, ViInt32 colorMap,
              ViInt32 predefinedColors);
ViStatus _VI_FUNC rszvl_ConfigureSANHardcopyDevice (ViSession instrumentHandle,
                                                    ViInt32 device,
                                                    ViInt32 destination,
                                                    ViInt32 pageOrientation);
ViStatus _VI_FUNC rszvl_SANHardcopyAbort (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_SANHardcopyPrint (ViSession instrumentHandle,
                                          ViInt32 device, ViInt32 items);
ViStatus _VI_FUNC rszvl_SANHardcopyPrintNext (ViSession instrumentHandle,
                                              ViInt32 device);
ViStatus _VI_FUNC rszvl_SANHardcopySetFileName (ViSession instrumentHandle,
                                                ViString name);
ViStatus _VI_FUNC rszvl_SANHardcopyGetPrinterList (ViSession instrumentHandle,
                                                   ViInt32 bufferSize,
                                                   ViChar _VI_FAR printerList[]);
ViStatus _VI_FUNC rszvl_SANHardcopySetPrinter (ViSession instrumentHandle,
                                               ViInt32 device,
                                               ViString printerName);
ViStatus _VI_FUNC rszvl_SANHardcopyComment (ViSession instrumentHandle,
                                            ViString comment);
ViStatus _VI_FUNC rszvl_SANHardcopyColor (ViSession instrumentHandle,
                                          ViBoolean color, ViInt32 defaultSetting);
ViStatus _VI_FUNC rszvl_ConfigureSANHardcopyColor (ViSession instrumentHandle,
                                                   ViInt32 colorMap, ViReal64 TINT,
                                                   ViReal64 SATURATION,
                                                   ViReal64 BRIGHTNESS);
ViStatus _VI_FUNC rszvl_ConfigureSANHardcopyColorByPredefined
             (ViSession instrumentHandle, ViInt32 colorMap,
              ViInt32 predefinedColors);
ViStatus _VI_FUNC rszvl_SANFileDecimalSeparator (ViSession instrumentHandle,
                                                 ViInt32 decimalSeparator);
ViStatus _VI_FUNC rszvl_SANDataSetFileSelectItems (ViSession instrumentHandle,
                                                   ViInt32 itemSelector,
                                                   ViBoolean itemState);
ViStatus _VI_FUNC rszvl_SANCalibration (ViSession instrumentHandle, ViBoolean sync,
                                        ViUInt32 timeoutms, ViPInt32 result);
ViStatus _VI_FUNC rszvl_SANCalibrationAbort (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_SANCalibrationResult (ViSession instrumentHandle,
                                              ViInt32 arraySize,
                                              ViChar _VI_FAR result[]);
ViStatus _VI_FUNC rszvl_SANCalibrationState (ViSession instrumentHandle,
                                             ViBoolean state);
ViStatus _VI_FUNC rszvl_ConfigureSANServiceInput (ViSession instrumentHandle,
                                                  ViInt32 input);
ViStatus _VI_FUNC rszvl_ConfigureSANServiceCombFrequency (ViSession instrumentHandle,
                                                          ViInt32 frequency);
ViStatus _VI_FUNC rszvl_ConfigureSANServiceNoiseSource (ViSession instrumentHandle,
                                                        ViBoolean noiseSource);
ViStatus _VI_FUNC rszvl_ConfigureSANServiceHWInfo (ViSession instrumentHandle,
                                                   ViInt32 bufferSize,
                                                   ViChar _VI_FAR HWInfo[]);
ViStatus _VI_FUNC rszvl_ReadSANYTrace (ViSession instrumentHandle, ViInt32 trace,
                                       ViUInt32 maximumTime_ms, ViInt32 arrayLength,
                                       ViPInt32 actualPoints,
                                       ViReal64 _VI_FAR amplitude[]);
ViStatus _VI_FUNC rszvl_FetchSANYTrace (ViSession instrumentHandle, ViInt32 trace,
                                        ViInt32 arrayLength, ViPInt32 actualPoints,
                                        ViReal64 _VI_FAR amplitude[]);
ViStatus _VI_FUNC rszvl_ReadSANTraceIQData (ViSession instrumentHandle,
                                            ViUInt32 timeout, ViInt32 bufferSize,
                                            ViPInt32 no_ofPoints,
                                            ViReal64 _VI_FAR realParts_I[],
                                            ViReal64 _VI_FAR imaginaryParts_Q[]);
ViStatus _VI_FUNC rszvl_FetchSANTraceIQData (ViSession instrumentHandle,
                                             ViInt32 offsetSamples,
                                             ViInt32 no_ofSamples,
                                             ViInt32 bufferSize,
                                             ViPInt32 no_ofPoints,
                                             ViReal64 _VI_FAR realParts_I[],
                                             ViReal64 _VI_FAR imaginaryParts_Q[]);
ViStatus _VI_FUNC rszvl_ReadSANMemoryIQLargeData (ViSession instrumentHandle,
                                                  ViUInt32 timeout, ViInt32 format,
                                                  ViPInt32 no_ofPoints);
ViStatus _VI_FUNC rszvl_ReadSANMemoryIQLargeDataBlock (ViSession instrumentHandle,
                                                       ViInt32 bufferSize,
                                                       ViChar _VI_FAR returnedData[],
                                                       ViPInt32 returnedCounts);
ViStatus _VI_FUNC rszvl_GetSANLimitCheckResult (ViSession instrumentHandle,
                                                ViInt32 limit, ViPInt32 state);
ViStatus _VI_FUNC rszvl_SANInitiate (ViSession instrumentHandle, ViUInt32 timeout);
ViStatus _VI_FUNC rszvl_SANAbort (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_SANContinue (ViSession instrumentHandle, ViUInt32 timeout);
ViStatus _VI_FUNC rszvl_SANSendSoftwareTrigger (ViSession instrumentHandle,
                                                ViUInt32 timeout);
ViStatus _VI_FUNC rszvl_GetSANSweepNumber (ViSession instrumentHandle,
                                           ViPInt32 numberOfSweeps);
ViStatus _VI_FUNC rszvl_ConfigureSANMarkerNoiseMeasurement
             (ViSession instrumentHandle, ViInt32 marker, ViBoolean state);
ViStatus _VI_FUNC rszvl_QuerySANMarkerNoiseMeasurementResult
             (ViSession instrumentHandle, ViInt32 marker, ViPReal64 result);
ViStatus _VI_FUNC rszvl_ConfigureSANMarkerPhaseNoiseMeasurement
             (ViSession instrumentHandle, ViBoolean state);
ViStatus _VI_FUNC rszvl_SANPhaseNoisePeakSearch (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_QuerySANPhaseNoiseResult (ViSession instrumentHandle,
                                                  ViPReal64 phaseNoiseResult);
ViStatus _VI_FUNC rszvl_ConfigureSANMarkerFrequencyCounter
             (ViSession instrumentHandle, ViInt32 marker,
              ViBoolean markerFrequencyCounter,
              ViReal64 frequencyCounterResolution);
ViStatus _VI_FUNC rszvl_QuerySANMarkerFrequencyCounterResult
             (ViSession instrumentHandle, ViInt32 marker,
              ViPReal64 counterFrequency);
ViStatus _VI_FUNC rszvl_ConfigureSANNdBPoints (ViSession instrumentHandle,
                                               ViBoolean n_dBState,
                                               ViReal64 n_dBLevel);
ViStatus _VI_FUNC rszvl_QuerySANNdBResult (ViSession instrumentHandle,
                                           ViPReal64 n_dBResult);
ViStatus _VI_FUNC rszvl_QuerySANNdBFrequencies (ViSession instrumentHandle,
                                                ViPReal64 n_dBFrequencyLower,
                                                ViPReal64 n_dBFrequencyHigher);
ViStatus _VI_FUNC rszvl_QuerySANNdBTimes (ViSession instrumentHandle,
                                          ViPReal64 n_dBTimeLower,
                                          ViPReal64 n_dBTimeHigher);
ViStatus _VI_FUNC rszvl_ConfigureSANListPowerSequence (ViSession instrumentHandle,
                                                       ViInt32 no_ofListItems,
                                                       ViReal64 _VI_FAR analyzerFrequency[],
                                                       ViReal64 _VI_FAR referenceLevel[],
                                                       ViReal64 _VI_FAR RFInputAttenuation[],
                                                       ViReal64 _VI_FAR RFInputElectronicAttn[],
                                                       ViInt32 _VI_FAR filterType[],
                                                       ViReal64 _VI_FAR resolutionBandwidth[],
                                                       ViReal64 _VI_FAR videoBandwidth[],
                                                       ViReal64 _VI_FAR measTime[],
                                                       ViReal64 _VI_FAR triggerLevel[]);
ViStatus _VI_FUNC rszvl_ConfigureSANListPowerSet (ViSession instrumentHandle,
                                                  ViBoolean peakMeas,
                                                  ViBoolean RMSMeas,
                                                  ViBoolean AVGMeas,
                                                  ViInt32 triggerMode,
                                                  ViInt32 triggerSlope,
                                                  ViReal64 triggerOffset,
                                                  ViReal64 gateLength);
ViStatus _VI_FUNC rszvl_SANListPowerMeasurementOff (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_QuerySANListPowerSequence (ViSession instrumentHandle,
                                                   ViInt32 no_ofListItems,
                                                   ViReal64 _VI_FAR analyzerFrequency[],
                                                   ViReal64 _VI_FAR referenceLevel[],
                                                   ViReal64 _VI_FAR RFInputAttenuation[],
                                                   ViReal64 _VI_FAR RFInputElectronicAttn[],
                                                   ViInt32 _VI_FAR filterType[],
                                                   ViReal64 _VI_FAR resolutionBandwidth[],
                                                   ViReal64 _VI_FAR videoBandwidth[],
                                                   ViReal64 _VI_FAR measTime[],
                                                   ViReal64 _VI_FAR triggerLevel[],
                                                   ViReal64 _VI_FAR listPowerResults[]);
ViStatus _VI_FUNC rszvl_QuerySANListPowerResult (ViSession instrumentHandle,
                                                 ViInt32 no_ofResults,
                                                 ViReal64 _VI_FAR listPowerResults[],
                                                 ViPInt32 returnedValues);
ViStatus _VI_FUNC rszvl_ConfigureSANTimeDomainPowerMeasurementState
             (ViSession instrumentHandle, ViBoolean state);
ViStatus _VI_FUNC rszvl_ConfigureSANTimeDomainPowerMeasurement
             (ViSession instrumentHandle, ViBoolean peak, ViBoolean RMS,
              ViBoolean mean, ViBoolean standardDeviation, ViBoolean average,
              ViBoolean maxHold, ViInt32 power);
ViStatus _VI_FUNC rszvl_SANTimeDomainPowerMeasurementAllOff
             (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_SANTimeDomainPowerMeasurementSetReference
             (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_QuerySANTDomPowerResult (ViSession instrumentHandle,
                                                 ViInt32 measurement,
                                                 ViInt32 resultType,
                                                 ViPReal64 result);
ViStatus _VI_FUNC rszvl_QuerySANTDomPowerPulseSequenceResult
             (ViSession instrumentHandle, ViReal64 timeOffset_ofFirstPulse,
              ViReal64 measurementTime, ViReal64 period, ViInt32 number_ofPulses,
              ViPReal64 result);
ViStatus _VI_FUNC rszvl_SelectSANPowerMeasurement (ViSession instrumentHandle,
                                                   ViBoolean state,
                                                   ViInt32 powerMeasurement);
ViStatus _VI_FUNC rszvl_ConfigureSANPowerMeasurement (ViSession instrumentHandle,
                                                      ViInt32 noOfAdjChannels,
                                                      ViInt32 channelMode,
                                                      ViInt32 powerMode,
                                                      ViBoolean resultMode);
ViStatus _VI_FUNC rszvl_ConfigureSANPowerChannelSpacing (ViSession instrumentHandle,
                                                         ViInt32 channel,
                                                         ViInt32 channelNumber,
                                                         ViReal64 spacing);
ViStatus _VI_FUNC rszvl_ConfigureSANPowerChannelBandwidth
             (ViSession instrumentHandle, ViInt32 channel, ViInt32 channelNumber,
              ViReal64 bandwidth);
ViStatus _VI_FUNC rszvl_ConfigureSANChannelPowerTrace (ViSession instrumentHandle,
                                                       ViInt32 traceNumber);
ViStatus _VI_FUNC rszvl_ConfigureSANChannelPowerStandard (ViSession instrumentHandle,
                                                          ViInt32 ACPStandard);
ViStatus _VI_FUNC rszvl_SANPowerAdjustReferenceLevel (ViSession instrumentHandle,
                                                      ViUInt32 timeout);
ViStatus _VI_FUNC rszvl_SANPowerPresetMeasurement (ViSession instrumentHandle,
                                                   ViInt32 channelPowerType,
                                                   ViUInt32 timeout);
ViStatus _VI_FUNC rszvl_QuerySANPowerResults (ViSession instrumentHandle,
                                              ViInt32 powerMeasurement,
                                              ViInt32 arraySize,
                                              ViReal64 _VI_FAR results[],
                                              ViPInt32 returnedValues);
ViStatus _VI_FUNC rszvl_ConfigureSANACPFastACPState (ViSession instrumentHandle,
                                                     ViBoolean fastACPState);
ViStatus _VI_FUNC rszvl_SANACPChannelPowerAutoAdjust (ViSession instrumentHandle,
                                                      ViUInt32 timeout);
ViStatus _VI_FUNC rszvl_ConfigureSANACPLimitState (ViSession instrumentHandle,
                                                   ViBoolean state);
ViStatus _VI_FUNC rszvl_ConfigureSANACPAdjacentChannelLimit
             (ViSession instrumentHandle, ViInt32 limitCheck, ViReal64 limitValue);
ViStatus _VI_FUNC rszvl_ConfigureSANACPAlternateChannelLimit
             (ViSession instrumentHandle, ViInt32 alternateChannelNumber,
              ViInt32 limitCheck, ViReal64 limitValue);
ViStatus _VI_FUNC rszvl_QuerySANACPAdjacentChannelLimitCheckResults
             (ViSession instrumentHandle, ViPInt32 lowerResult,
              ViPInt32 upperResult);
ViStatus _VI_FUNC rszvl_QuerySANACPAlternateChannelLimitCheckResults
             (ViSession instrumentHandle, ViInt32 alternateChannelNumber,
              ViPInt32 lowerResult, ViPInt32 upperResult);
ViStatus _VI_FUNC rszvl_ConfigureSANMulticarrierACPPower (ViSession instrumentHandle,
                                                          ViInt32 noOfAdjChannels,
                                                          ViInt32 noOfTXChannels,
                                                          ViInt32 channelMode,
                                                          ViInt32 powerMode,
                                                          ViBoolean resultMode);
ViStatus _VI_FUNC rszvl_ConfigureSANACPReferenceChannelAuto
             (ViSession instrumentHandle, ViInt32 reference);
ViStatus _VI_FUNC rszvl_ConfigureSANACPReferenceChannelMan
             (ViSession instrumentHandle, ViInt32 channel);
ViStatus _VI_FUNC rszvl_ConfigureSANOccupiedBandwidth (ViSession instrumentHandle,
                                                       ViReal64 powerBandwidth,
                                                       ViReal64 channelBandwidth);
ViStatus _VI_FUNC rszvl_ConfigureSANSignalStatisticMeasurement
             (ViSession instrumentHandle, ViInt32 measurement,
              ViInt32 number_ofSamples, ViReal64 resolutionBandwidth);
ViStatus _VI_FUNC rszvl_ConfigureSANSignalStatisticMeasurementXAxis
             (ViSession instrumentHandle, ViReal64 xAxisRefLevel,
              ViReal64 xAxisRange);
ViStatus _VI_FUNC rszvl_ConfigureSANSignalStatisticMeasurementYAxis
             (ViSession instrumentHandle, ViReal64 yAxisMaxValue,
              ViReal64 yAxisMinValue);
ViStatus _VI_FUNC rszvl_ConfigureSANSignalStatisticSweep (ViSession instrumentHandle,
                                                          ViBoolean sweepModeContinuous);
ViStatus _VI_FUNC rszvl_SANSignalStatisticMeasurementAdjustSettings
             (ViSession instrumentHandle, ViUInt32 timeout);
ViStatus _VI_FUNC rszvl_SANSignalStatisticMeasurementPresetScaling
             (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_QuerySANSignalStatisticResult (ViSession instrumentHandle,
                                                       ViInt32 trace,
                                                       ViInt32 resultType,
                                                       ViPReal64 result);
ViStatus _VI_FUNC rszvl_ConfigureSANModDepthMeasurement (ViSession instrumentHandle,
                                                         ViBoolean state);
ViStatus _VI_FUNC rszvl_QuerySANModulationDepth (ViSession instrumentHandle,
                                                 ViPReal64 modulationDepth);
ViStatus _VI_FUNC rszvl_SANModulationDepthSignalSearch (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureSANTOIMeasurement (ViSession instrumentHandle,
                                                    ViBoolean state);
ViStatus _VI_FUNC rszvl_QuerySANTOI (ViSession instrumentHandle,
                                     ViPReal64 TOIMeasurement);
ViStatus _VI_FUNC rszvl_ConfigureSANHDistMeasurement (ViSession instrumentHandle,
                                                      ViBoolean state,
                                                      ViInt32 no_ofHarmonics,
                                                      ViBoolean harmonicRBWAuto);
ViStatus _VI_FUNC rszvl_SANAdjustHDistSettings (ViSession instrumentHandle,
                                                ViUInt32 timeout);
ViStatus _VI_FUNC rszvl_QuerySANHDist (ViSession instrumentHandle,
                                       ViPReal64 harmonicDistortion_in_percent,
                                       ViPReal64 harmonicDistortion_in_dB);
ViStatus _VI_FUNC rszvl_QuerySANHDistHarmonicsList (ViSession instrumentHandle,
                                                    ViReal64 _VI_FAR harmonicsList[]);
ViStatus _VI_FUNC rszvl_ConfigureSANBurstPowerSequence (ViSession instrumentHandle,
                                                        ViReal64 analyzerFrequencyHz,
                                                        ViReal64 resolutionBandwidthHz,
                                                        ViReal64 measTimes,
                                                        ViInt32 triggerSource,
                                                        ViReal64 triggerLevel,
                                                        ViReal64 triggerOffsets,
                                                        ViInt32 type_ofMeas,
                                                        ViInt32 no_ofMeasurements);
ViStatus _VI_FUNC rszvl_ConfigureSANBurstPowerFilterType (ViSession instrumentHandle,
                                                          ViInt32 filterType);
ViStatus _VI_FUNC rszvl_GetSANBurstPowerSequence (ViSession instrumentHandle,
                                                  ViReal64 analyzerFrequencyHz,
                                                  ViReal64 resolutionBandwidthHz,
                                                  ViReal64 measTimes,
                                                  ViInt32 triggerSource,
                                                  ViReal64 triggerLevel,
                                                  ViReal64 triggerOffsets,
                                                  ViInt32 type_ofMeas,
                                                  ViInt32 no_ofMeasurements,
                                                  ViUInt32 timeout,
                                                  ViReal64 _VI_FAR burstPowerResults[]);
ViStatus _VI_FUNC rszvl_GetSANBurstPowerResults (ViSession instrumentHandle,
                                                 ViInt32 no_ofResults,
                                                 ViReal64 _VI_FAR burstPowerResults[]);
ViStatus _VI_FUNC rszvl_GetSANBurstPwrResultMin (ViSession instrumentHandle,
                                                 ViPReal64 powerResultMin);
ViStatus _VI_FUNC rszvl_ADemodMode (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_SetADemodState (ViSession instrumentHandle,
                                        ViBoolean state);
ViStatus _VI_FUNC rszvl_SelectADemodResult (ViSession instrumentHandle,
                                            ViInt32 operatingMode, ViInt32 trace);
ViStatus _VI_FUNC rszvl_ConfigureADemodTrigger (ViSession instrumentHandle,
                                                ViInt32 triggerSource,
                                                ViReal64 triggerOffset,
                                                ViInt32 triggerPolarity,
                                                ViReal64 triggerValue);
ViStatus _VI_FUNC rszvl_ConfigureADemodSet (ViSession instrumentHandle,
                                            ViReal64 sampleRate,
                                            ViInt32 recordLength,
                                            ViInt32 triggerSource,
                                            ViInt32 triggerSlope,
                                            ViInt32 pretriggerSamples,
                                            ViInt32 numberOfMeas);
ViStatus _VI_FUNC rszvl_ConfigureADemodFilter (ViSession instrumentHandle,
                                               ViInt32 filterType,
                                               ViBoolean filterState,
                                               ViReal64 filterFrequency,
                                               ViReal64 AMFMDeemphasisValue);
ViStatus _VI_FUNC rszvl_ConfigureADemodZoom (ViSession instrumentHandle,
                                             ViBoolean zoomState,
                                             ViReal64 startTime);
ViStatus _VI_FUNC rszvl_ConfigureADemodBandwidth (ViSession instrumentHandle,
                                                  ViReal64 demodulationBandwidth);
ViStatus _VI_FUNC rszvl_ConfigureADemodAFCoupling (ViSession instrumentHandle,
                                                   ViInt32 AFCoupling);
ViStatus _VI_FUNC rszvl_ConfigureADemodMeasuringTime (ViSession instrumentHandle,
                                                      ViReal64 measuringTime);
ViStatus _VI_FUNC rszvl_ConfigureADemodAFFrequencyStartStop
             (ViSession instrumentHandle, ViReal64 startFrequency,
              ViReal64 stopFrequency);
ViStatus _VI_FUNC rszvl_ConfigureADemodAFFrequencyCenterSpan
             (ViSession instrumentHandle, ViReal64 centerFrequency, ViReal64 span);
ViStatus _VI_FUNC rszvl_ConfigureADemodAFFullSpan (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureADemodAFParam (ViSession instrumentHandle,
                                                ViReal64 resolutionBandwidth);
ViStatus _VI_FUNC rszvl_ConfigureADemodRFParam (ViSession instrumentHandle,
                                                ViReal64 resolutionBandwidth,
                                                ViReal64 span);
ViStatus _VI_FUNC rszvl_ConfigureADemodRFZoom (ViSession instrumentHandle,
                                               ViReal64 zoom);
ViStatus _VI_FUNC rszvl_ConfigureAnalogDemodPMPhaseWrap (ViSession instrumentHandle,
                                                         ViInt32 phaseWrap);
ViStatus _VI_FUNC rszvl_ConfigureAnalogDemodPMPhaseUnits (ViSession instrumentHandle,
                                                          ViInt32 PMUnits);
ViStatus _VI_FUNC rszvl_ConfigureAnalogDemodPMZeroPhaseReferencePoint
             (ViSession instrumentHandle, ViReal64 referencePoint);
ViStatus _VI_FUNC rszvl_ConfigureADemodResultType (ViSession instrumentHandle,
                                                   ViInt32 demodulationType,
                                                   ViInt32 _VI_FAR resultTypes[]);
ViStatus _VI_FUNC rszvl_GetADemodCurrentlySetValue (ViSession instrumentHandle,
                                                    ViInt32 value,
                                                    ViPReal64 result);
ViStatus _VI_FUNC rszvl_GetADemodResultValues (ViSession instrumentHandle,
                                               ViInt32 demodulationType,
                                               ViInt32 resultType, ViUInt32 timeout,
                                               ViInt32 arraySize,
                                               ViReal64 _VI_FAR resultValues[],
                                               ViPInt32 returnedValues);
ViStatus _VI_FUNC rszvl_GetADemodOffset (ViSession instrumentHandle,
                                         ViInt32 resultType, ViPReal64 value);
ViStatus _VI_FUNC rszvl_GetADemodMarkerModulationValue (ViSession instrumentHandle,
                                                        ViInt32 demodulationType,
                                                        ViInt32 detectorType,
                                                        ViInt32 trace,
                                                        ViPReal64 value);
ViStatus _VI_FUNC rszvl_GetADemodMarkerMeasValue (ViSession instrumentHandle,
                                                  ViInt32 measurement,
                                                  ViInt32 trace, ViPReal64 value);
ViStatus _VI_FUNC rszvl_BTOMode (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureBTOMeasurement (ViSession instrumentHandle,
                                                 ViInt32 channel,
                                                 ViInt32 outputPowerClass,
                                                 ViInt32 samplesPerSymbols,
                                                 ViInt32 packetType,
                                                 ViReal64 antennaGain);
ViStatus _VI_FUNC rszvl_ConfigureBTOAveragingMethod (ViSession instrumentHandle,
                                                     ViReal64 startPosition,
                                                     ViReal64 stopPosition);
ViStatus _VI_FUNC rszvl_ConfigureBTOSynchronization (ViSession instrumentHandle,
                                                     ViBoolean findSync,
                                                     ViInt32 syncOffset,
                                                     ViBoolean findBurst,
                                                     ViReal64 burstOffset,
                                                     ViBoolean searchLengthAuto,
                                                     ViReal64 searchLength);
ViStatus _VI_FUNC rszvl_ConfigureBTOLAP (ViSession instrumentHandle,
                                         ViInt32 lowest24Bits_ofDevAddress);
ViStatus _VI_FUNC rszvl_ConfigureBTOMeasurementMode (ViSession instrumentHandle,
                                                     ViInt32 measurementMode);
ViStatus _VI_FUNC rszvl_ConfigureBTONumberOfAdjacentChannels
             (ViSession instrumentHandle, ViInt32 number_ofAdjacentChannels);
ViStatus _VI_FUNC rszvl_ConfigureBTOReceiveFilter (ViSession instrumentHandle,
                                                   ViInt32 filter);
ViStatus _VI_FUNC rszvl_ConfigureBTOSweepCouplingAuto (ViSession instrumentHandle,
                                                       ViInt32 sweepCoupling);
ViStatus _VI_FUNC rszvl_ConfigureBTOSweepCoupling (ViSession instrumentHandle,
                                                   ViInt32 sweepCoupling,
                                                   ViReal64 couplingValue);
ViStatus _VI_FUNC rszvl_ConfigureBTOAcquisition (ViSession instrumentHandle,
                                                 ViBoolean sweepModeContinuous,
                                                 ViInt32 numberOfSweeps);
ViStatus _VI_FUNC rszvl_SelectBTOTrace (ViSession instrumentHandle,
                                        ViInt32 trace);
ViStatus _VI_FUNC rszvl_ConfigureBTOTraceMode (ViSession instrumentHandle,
                                           ViInt32 trace, ViInt32 traceMode);
ViStatus _VI_FUNC rszvl_ConfigureBTOTraceDetector (ViSession instrumentHandle,
                                           ViInt32 trace, ViInt32 traceDetector);
ViStatus _VI_FUNC rszvl_ConfigureBTOPacketBytesSCO (ViSession instrumentHandle,
                                                    ViInt32 packetBytesSCO);
ViStatus _VI_FUNC rszvl_ConfigureBTODisplayYAxisScaling (ViSession instrumentHandle,
                                                         ViReal64 scaling);
ViStatus _VI_FUNC rszvl_ConfigureBTOAnalogDemodulationZoom
             (ViSession instrumentHandle, ViBoolean zoomState, ViReal64 startTime);
ViStatus _VI_FUNC rszvl_GetBTOMesurementStatus (ViSession instrumentHandle,
                                                ViPInt32 status);
ViStatus _VI_FUNC rszvl_GetBTOPacketInfo (ViSession instrumentHandle, ViPInt32 type,
                                          ViPInt32 length);
ViStatus _VI_FUNC rszvl_GetBTOOutputPower (ViSession instrumentHandle, ViInt32 type,
                                           ViPReal64 power);
ViStatus _VI_FUNC rszvl_GetBTONumberOfExceptions (ViSession instrumentHandle,
                                                  ViPInt32 numberOfExceptions);
ViStatus _VI_FUNC rszvl_GetBTOPowerOfChannels (ViSession instrumentHandle,
                                               ViInt32 arraySize,
                                               ViReal64 _VI_FAR power[],
                                               ViPInt32 returnedValues);
ViStatus _VI_FUNC rszvl_GetBTOFrequencyDeviation (ViSession instrumentHandle,
                                                  ViInt32 frequencyDeviation,
                                                  ViInt32 type, ViPReal64 value);
ViStatus _VI_FUNC rszvl_GetBTORatioAverageFrequencyDeviation
             (ViSession instrumentHandle, ViInt32 type, ViPReal64 value);
ViStatus _VI_FUNC rszvl_GetBTOPercentageDeviation (ViSession instrumentHandle,
                                                   ViPReal64 value);
ViStatus _VI_FUNC rszvl_GetBTOInitCarrFreqTol (ViSession instrumentHandle,
                                               ViInt32 type, ViPReal64 value);
ViStatus _VI_FUNC rszvl_GetBTOCarrFreqDrift (ViSession instrumentHandle,
                                             ViInt32 type, ViPReal64 value);
ViStatus _VI_FUNC rszvl_ConfigureBTOSENumAdjacentChannels
             (ViSession instrumentHandle, ViInt32 number);
ViStatus _VI_FUNC rszvl_AdjustBTOSEGateSetting (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_GetBTOChannelPowerOfSEMeasurement
             (ViSession instrumentHandle, ViInt32 parameter, ViPReal64 value);
ViStatus _VI_FUNC rszvl_GetBTOSEPowerOfAdjacentChannels (ViSession instrumentHandle,
                                                         ViInt32 arraySize,
                                                         ViReal64 _VI_FAR power[],
                                                         ViPInt32 returnedValues);
ViStatus _VI_FUNC rszvl_GetBTOSENumberOfExceptions (ViSession instrumentHandle,
                                                    ViPInt32 exceptions);
ViStatus _VI_FUNC rszvl_ConfigureBTOCFSBlockCount (ViSession instrumentHandle,
                                                   ViInt32 blockCount);
ViStatus _VI_FUNC rszvl_GetBTOCFSDeviation (ViSession instrumentHandle,
                                            ViInt32 parameter, ViInt32 type,
                                            ViPReal64 value);
ViStatus _VI_FUNC rszvl_GetBTOCFSModulAccuracy (ViSession instrumentHandle,
                                                ViInt32 parameter, ViInt32 type,
                                                ViPReal64 value);
ViStatus _VI_FUNC rszvl_GetBTODifferentialPhaseEncodingValues
             (ViSession instrumentHandle, ViInt32 parameter, ViPInt32 value);
ViStatus _VI_FUNC rszvl_ConfigureBTORTPMeasurement (ViSession instrumentHandle,
                                                    ViInt32 packetSection,
                                                    ViReal64 start, ViReal64 stop);
ViStatus _VI_FUNC rszvl_GetBTORelativeTransmissionPower (ViSession instrumentHandle,
                                                         ViInt32 parameter,
                                                         ViInt32 modificator,
                                                         ViPReal64 value);
ViStatus _VI_FUNC rszvl_PWMMode (ViSession instrumentHandle, ViBoolean state);
ViStatus _VI_FUNC rszvl_ConfigurePWMMeasurement (ViSession instrumentHandle,
                                                 ViInt32 measurementTime,
                                                 ViInt32 coupling,
                                                 ViReal64 frequency);
ViStatus _VI_FUNC rszvl_ConfigurePWMRelative (ViSession instrumentHandle,
                                              ViReal64 referenceValue,
                                              ViInt32 unit);
ViStatus _VI_FUNC rszvl_ConfigurePWMAbsolute (ViSession instrumentHandle,
                                              ViInt32 unit);
ViStatus _VI_FUNC rszvl_ConfigurePWMReferenceLevelOffsetState (ViSession instrumentHandle,
                                                               ViBoolean state);
ViStatus _VI_FUNC rszvl_ConfigurePWMAveraging (ViSession instrumentHandle,
                                               ViBoolean manualAveraging,
                                               ViInt32 averagingCount);
ViStatus _VI_FUNC rszvl_ConfigurePWMMeasToRef (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigurePWMZero (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ReadPWMResult (ViSession instrumentHandle, ViUInt32 timeout,
                                       ViReal64* result);
ViStatus _VI_FUNC rszvl_FetchPWMResult (ViSession instrumentHandle,
                                        ViReal64* result);
ViStatus _VI_FUNC rszvl_SpectrogramMode (ViSession instrumentHandle,
                                         ViBoolean state);
ViStatus _VI_FUNC rszvl_ConfigureSpectrogramSweep (ViSession instrumentHandle,
                                                   ViBoolean continuous);
ViStatus _VI_FUNC rszvl_ConfigureSpectrogramDisplayParameters
             (ViSession instrumentHandle, ViInt32 color, ViInt32 size);
ViStatus _VI_FUNC rszvl_ConfigureSpectrogramMeasurementParameters
             (ViSession instrumentHandle, ViInt32 historyDepth, ViBoolean timeStamp,
              ViInt32 frameCount, ViBoolean continueFrame);
ViStatus _VI_FUNC rszvl_SpectrogramClear (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureSpectrogramMarkerFrameNumber
             (ViSession instrumentHandle, ViInt32 markerNumber,
              ViInt32 frameNumber);
ViStatus _VI_FUNC rszvl_GetSpectrogramMarkerFrameNumber (ViSession instrumentHandle,
                                                         ViInt32 markerNumber,
                                                         ViPInt32 frameNumber);
ViStatus _VI_FUNC rszvl_ConfigureSpectrogramMarkerSearchArea
             (ViSession instrumentHandle, ViInt32 markerNumber, ViInt32 searchArea);
ViStatus _VI_FUNC rszvl_SpectrogramMarkerSearch (ViSession instrumentHandle,
                                                 ViInt32 markerNumber,
                                                 ViInt32 searchMode);
ViStatus _VI_FUNC rszvl_ConfigureSpectrogramDeltaMarkerFrameNumber
             (ViSession instrumentHandle, ViInt32 deltaMarkerNumber,
              ViInt32 frameNumber);
ViStatus _VI_FUNC rszvl_GetSpectrogramDeltaMarkerFrameNumber
             (ViSession instrumentHandle, ViInt32 deltaMarker,
              ViPInt32 frameNumber);
ViStatus _VI_FUNC rszvl_ConfigureSpectrogramDeltaMarkerSearchArea
             (ViSession instrumentHandle, ViInt32 deltaMarker, ViInt32 searchArea);
ViStatus _VI_FUNC rszvl_SpectrogramDeltaMarkerSearch (ViSession instrumentHandle,
                                                      ViInt32 deltaMarker,
                                                      ViInt32 searchMode);
ViStatus _VI_FUNC rszvl_GetSpectrogramTimeStamp (ViSession instrumentHandle,
                                                 ViInt32 frame, ViInt32 arraySize,
                                                 ViInt32 _VI_FAR part1[],
                                                 ViInt32 _VI_FAR part2[],
                                                 ViInt32 _VI_FAR part3[],
                                                 ViInt32 _VI_FAR part4[],
                                                 ViPInt32 returnedValues);
ViStatus _VI_FUNC rszvl_ReadSpectrogramTraceData (ViSession instrumentHandle,
                                                  ViInt32 arraySize,
                                                  ViReal64 _VI_FAR results[],
                                                  ViPInt32 returnedValues);
ViStatus _VI_FUNC rszvl_CATVMode (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureCATVChannelSetup (ViSession instrumentHandle,
                                                      ViString channelTableName,
                                                      ViInt32 measurementChannel);
ViStatus _VI_FUNC rszvl_ConfigureCATVChannelFrequency (ViSession instrumentHandle,
                                                          ViReal64 RFFrequency,
                                                          ViInt32 axisLabeling);
ViStatus _VI_FUNC rszvl_ConfigureCATVChannelFrequencyStepSize
             (ViSession instrumentHandle, ViReal64 stepSize);
ViStatus _VI_FUNC rszvl_ConfigureCATVModulationStandard
             (ViSession instrumentHandle, ViString modulationStandard,
              ViInt32 signalType, ViInt32 sidebandPosition);
ViStatus _VI_FUNC rszvl_CATVChannelAdjustAtt (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_CATVAutoscale (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureCATVMarkerState (ViSession instrumentHandle,
                                                     ViInt32 markerNumber,
                                                     ViBoolean markerState);
ViStatus _VI_FUNC rszvl_ConfigureCATVDeltaMarkerState (ViSession instrumentHandle,
                                                          ViInt32 deltaMarkerNumber,
                                                          ViBoolean deltaMarkerState);
ViStatus _VI_FUNC rszvl_ConfigureCATVAMeasurement (ViSession instrumentHandle,
                                                      ViInt32 measurement);
ViStatus _VI_FUNC rszvl_ConfigureCATVAModulStandard (ViSession instrumentHandle,
                                                        ViInt32 TVStandard,
                                                        ViInt32 soundSystem,
                                                        ViInt32 groupDelay,
                                                        ViInt32 colorSystem,
                                                        ViInt32 barField,
                                                        ViInt32 barLine,
                                                        ViInt32 barLineType,
                                                        ViInt32 quietField,
                                                        ViInt32 quietLine);
ViStatus _VI_FUNC rszvl_ConfigureCATVAVideoScopeLine (ViSession instrumentHandle,
                                                         ViInt32 line);
ViStatus _VI_FUNC rszvl_ConfigureCATVAVideoScopeField (ViSession instrumentHandle,
                                                          ViInt32 activeField);
ViStatus _VI_FUNC rszvl_ConfigureCATVACNMeasurement (ViSession instrumentHandle,
                                                        ViInt32 measCarrier,
                                                        ViInt32 measurementMethod,
                                                        ViReal64 noiseReferenceBandwidth);
ViStatus _VI_FUNC rszvl_ConfigureCATVACNReferencePower
             (ViSession instrumentHandle, ViInt32 referencePower,
              ViInt32 referenceChannel, ViReal64 powerManual);
ViStatus _VI_FUNC rszvl_ConfigureCATVACNMeasurementFrequencies
             (ViSession instrumentHandle, ViInt32 tableRow, ViBoolean state,
              ViReal64 centerFrequency, ViReal64 span);
ViStatus _VI_FUNC rszvl_ConfigureCATVACNNoiseFloorCorrection
             (ViSession instrumentHandle, ViBoolean noiseFloorCorrection,
              ViUInt32 timeout);
ViStatus _VI_FUNC rszvl_CATVACNNextMeasFrequency (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureCATVACSOMeasurement (ViSession instrumentHandle,
                                                         ViInt32 measCarrier,
                                                         ViInt32 measurementMethod);
ViStatus _VI_FUNC rszvl_ConfigureCATVACSOReferencePower
             (ViSession instrumentHandle, ViInt32 referencePower,
              ViInt32 referenceChannel, ViReal64 powerManual);
ViStatus _VI_FUNC rszvl_ConfigureCATVACSOMeasurementFrequencies
             (ViSession instrumentHandle, ViInt32 tableRow, ViBoolean state,
              ViReal64 centerFrequency, ViReal64 span);
ViStatus _VI_FUNC rszvl_ConfigureCATVACSONoiseFloorCorrection
             (ViSession instrumentHandle, ViBoolean noiseFloorCorrection,
              ViUInt32 timeout);
ViStatus _VI_FUNC rszvl_CATVACSONextMeasFrequency (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureCATVACTBMeasurement (ViSession instrumentHandle,
                                                         ViInt32 measCarrier);
ViStatus _VI_FUNC rszvl_ConfigureCATVACTBReferencePower
             (ViSession instrumentHandle, ViInt32 referencePower,
              ViInt32 referenceChannel, ViReal64 powerManual);
ViStatus _VI_FUNC rszvl_ConfigureCATVACTBMeasurementFrequencies
             (ViSession instrumentHandle, ViInt32 tableRow, ViBoolean state,
              ViReal64 centerFrequency, ViReal64 span);
ViStatus _VI_FUNC rszvl_ConfigureCATVACTBNoiseFloorCorrection
             (ViSession instrumentHandle, ViBoolean noiseFloorCorrection,
              ViUInt32 timeout);
ViStatus _VI_FUNC rszvl_CATVACTBNextMeasFrequency (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureCATVACarrierLimit (ViSession instrumentHandle,
                                                       ViInt32 measurement,
                                                       ViInt32 limitType,
                                                       ViReal64 limitValue);
ViStatus _VI_FUNC rszvl_ConfigureCATVACarrierRatioLimit
             (ViSession instrumentHandle, ViInt32 measurement, ViInt32 limitType,
              ViReal64 limitValue);
ViStatus _VI_FUNC rszvl_ConfigureCATVAHumLimit (ViSession instrumentHandle,
                                                   ViInt32 limitType,
                                                   ViReal64 limitValue);
ViStatus _VI_FUNC rszvl_ConfigureCATVAVisionModLimit (ViSession instrumentHandle,
                                                         ViInt32 limit,
                                                         ViInt32 limitType,
                                                         ViReal64 limitValue);
ViStatus _VI_FUNC rszvl_ConfigureCATVAVisionModulationCarrierPowerUnit
             (ViSession instrumentHandle, ViInt32 unit);
ViStatus _VI_FUNC rszvl_ConfigureCATVAHumUnit (ViSession instrumentHandle,
                                                  ViInt32 unit);
ViStatus _VI_FUNC rszvl_ConfigureCATVDMeasurement (ViSession instrumentHandle,
                                                      ViInt32 measurement);
ViStatus _VI_FUNC rszvl_ConfigureCATVDModulStandard (ViSession instrumentHandle,
                                                        ViInt32 TVStandard,
                                                        ViInt32 rollOff,
                                                        ViInt32 constellationParameter,
                                                        ViReal64 symbolRate);
ViStatus _VI_FUNC rszvl_ConfigureCATVDShoulderAttenuation
             (ViSession instrumentHandle, ViBoolean shoulderAttenuation);
ViStatus _VI_FUNC rszvl_ConfigureCATVDEchoPattern (ViSession instrumentHandle,
                                                      ViBoolean zoomState,
                                                      ViInt32 zoomFactor,
                                                      ViReal64 velocityFactor);
ViStatus _VI_FUNC rszvl_ConfigureCATVDModulErrorZoom (ViSession instrumentHandle,
                                                         ViInt32 zoom);
ViStatus _VI_FUNC rszvl_ConfigureCATVDOverviewZoom (ViSession instrumentHandle,
                                                       ViInt32 zoom);
ViStatus _VI_FUNC rszvl_ConfigureCATVDConstellationDiagramZoom
             (ViSession instrumentHandle, ViInt32 zoomDiagram);
ViStatus _VI_FUNC rszvl_ConfigureCATVDEqualizer (ViSession instrumentHandle,
                                                    ViBoolean state,
                                                    ViBoolean freeze);
ViStatus _VI_FUNC rszvl_ConfigureCATVDSignalStatisticsNoSamples
             (ViSession instrumentHandle, ViInt32 no_ofSamples);
ViStatus _VI_FUNC rszvl_ConfigureCATVDSignalStatisticsScaling
             (ViSession instrumentHandle, ViInt32 scalingMode,
              ViReal64 scalingValue);
ViStatus _VI_FUNC rszvl_ConfigureCATVDCCDFPercentMark (ViSession instrumentHandle,
                                                          ViInt32 markerNumber,
                                                          ViReal64 positionValue);
ViStatus _VI_FUNC rszvl_CATVDResetEqualizer (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureCATVDOverLimit (ViSession instrumentHandle,
                                                    ViInt32 measurement,
                                                    ViInt32 limitType,
                                                    ViReal64 limitValue);
ViStatus _VI_FUNC rszvl_ConfigureCATVDChanPowLimit (ViSession instrumentHandle,
                                                       ViInt32 limitType,
                                                       ViReal64 limitValue);
ViStatus _VI_FUNC rszvl_ConfigureCATVDModErrLimit (ViSession instrumentHandle,
                                                      ViInt32 measurement,
                                                      ViInt32 limitType,
                                                      ViReal64 limitValue);
ViStatus _VI_FUNC rszvl_ConfigureCATVDEchoPatternUnit (ViSession instrumentHandle,
                                                          ViInt32 unit);
ViStatus _VI_FUNC rszvl_ConfigureCATVDUnit (ViSession instrumentHandle,
                                               ViInt32 measurement, ViInt32 unit);
ViStatus _VI_FUNC rszvl_ConfigureCATVTVMeasurement (ViSession instrumentHandle,
                                                       ViInt32 measurement);
ViStatus _VI_FUNC rszvl_ConfigureCATVTVStandardForTitlMeasurement
             (ViSession instrumentHandle, ViString modulationStandard,
              ViBoolean state);
ViStatus _VI_FUNC rszvl_QueryCATVACarrierResult (ViSession instrumentHandle,
                                                    ViInt32 measurement,
                                                    ViReal64* result);
ViStatus _VI_FUNC rszvl_QueryCATVACarrierAllResults (ViSession instrumentHandle,
                                                        ViReal64 _VI_FAR results[]);
ViStatus _VI_FUNC rszvl_QueryCATVACarrierRatioResult (ViSession instrumentHandle,
                                                         ViInt32 measurement,
                                                         ViInt32 resultType,
                                                         ViReal64* result);
ViStatus _VI_FUNC rszvl_QueryCATVACarrierAllRatioResults
             (ViSession instrumentHandle, ViInt32 measurement,
              ViReal64 _VI_FAR results[]);
ViStatus _VI_FUNC rszvl_QueryCATVAHumResult (ViSession instrumentHandle,
                                                ViReal64* result);
ViStatus _VI_FUNC rszvl_QueryCATVAVisionModulationResult
             (ViSession instrumentHandle, ViInt32 measurement, ViReal64* result);
ViStatus _VI_FUNC rszvl_QueryCATVAVisionAllModulationResults
             (ViSession instrumentHandle, ViReal64 _VI_FAR results[]);
ViStatus _VI_FUNC rszvl_QueryCATVACarrierLimitResult (ViSession instrumentHandle,
                                                         ViInt32 measurement,
                                                         ViInt32 limitType,
                                                         ViInt32* result);
ViStatus _VI_FUNC rszvl_QueryCATVACarrierRatioLimitResult
             (ViSession instrumentHandle, ViInt32 measurement, ViInt32 limitType,
              ViInt32* result);
ViStatus _VI_FUNC rszvl_QueryCATVAHumLimitResult (ViSession instrumentHandle,
                                                     ViInt32 limitType,
                                                     ViInt32* result);
ViStatus _VI_FUNC rszvl_QueryCATVAVisionModulationLimitResult
             (ViSession instrumentHandle, ViInt32 measurement, ViInt32 limitType,
              ViInt32* result);
ViStatus _VI_FUNC rszvl_QueryCATVDSpectrumResult (ViSession instrumentHandle,
                                                     ViInt32 measurement,
                                                     ViReal64* result);
ViStatus _VI_FUNC rszvl_QueryCATVDChanelPowerResult (ViSession instrumentHandle,
                                                        ViReal64* result);
ViStatus _VI_FUNC rszvl_QueryCATVDOverviewResult (ViSession instrumentHandle,
                                                     ViInt32 measurement,
                                                     ViReal64* result);
ViStatus _VI_FUNC rszvl_QueryCATVDOverviewAllResults (ViSession instrumentHandle,
                                                         ViReal64 _VI_FAR results[]);
ViStatus _VI_FUNC rszvl_QueryCATVDModulationErrorsResult
             (ViSession instrumentHandle, ViInt32 measurement, ViReal64* result);
ViStatus _VI_FUNC rszvl_QueryCATVDModulationErrorsAllResults
             (ViSession instrumentHandle, ViReal64 _VI_FAR results[]);
ViStatus _VI_FUNC rszvl_QueryCATVDSignalStatisticsResult
             (ViSession instrumentHandle, ViInt32 traceNumber, ViInt32 resultType,
              ViReal64* result);
ViStatus _VI_FUNC rszvl_QueryCATVDSignalStatisticsAllResults
             (ViSession instrumentHandle, ViInt32 traceNumber,
              ViReal64 _VI_FAR results[]);
ViStatus _VI_FUNC rszvl_QueryCATVDOverviewLimitResult (ViSession instrumentHandle,
                                                          ViInt32 measurement,
                                                          ViInt32 limitType,
                                                          ViInt32* result);
ViStatus _VI_FUNC rszvl_QueryCATVDChannelPowerLimitResult
             (ViSession instrumentHandle, ViInt32* result);
ViStatus _VI_FUNC rszvl_QueryCATVDModulationErrorsLimitResult
             (ViSession instrumentHandle, ViInt32 measurement, ViInt32 limitType,
              ViInt32* result);
ViStatus _VI_FUNC rszvl_NoiseMode (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureNoiseFrequencySettings (ViSession instrumentHandle,
                                                         ViReal64 startFrequency,
                                                         ViReal64 stopFrequency,
                                                         ViReal64 stepFrequency,
                                                         ViInt32 mode,
                                                         ViReal64 fixedLO,
                                                         ViReal64 imageRejection,
                                                         ViReal64 fixedFrequency);
ViStatus _VI_FUNC rszvl_ConfigureNoiseFrequencyTable (ViSession instrumentHandle,
                                                      ViInt32 numberOfValues,
                                                      ViReal64 _VI_FAR receiverFrequencyValues[],
                                                      ViReal64 _VI_FAR localOscilatorFrequencyVal[],
                                                      ViReal64 _VI_FAR intermediateFrequencyValues[]);
ViStatus _VI_FUNC rszvl_ConfigureNoiseMeasurementMode (ViSession instrumentHandle,
                                                       ViInt32 measurementMode);
ViStatus _VI_FUNC rszvl_ConfigureNoiseCalibration (ViSession instrumentHandle,
                                                   ViBoolean _2ndStageCorrection);
ViStatus _VI_FUNC rszvl_ConfigureNoiseAnalyzerSettings (ViSession instrumentHandle,
                                                        ViReal64 resolutionBandwidth,
                                                        ViReal64 sweepTime,
                                                        ViReal64 settlingTime,
                                                        ViInt32 average,
                                                        ViReal64 RFAttenuation,
                                                        ViReal64 range,
                                                        ViBoolean automaticRefLevel,
                                                        ViReal64 refLevel,
                                                        ViBoolean preamplifier_B22);
ViStatus _VI_FUNC rszvl_ConfigureNoiseENRSettings (ViSession instrumentHandle,
                                                   ViInt32 selectionMode,
                                                   ViReal64 ENRConstant,
                                                   ViReal64 roomTemperature);
ViStatus _VI_FUNC rszvl_ConfigureNoiseENRTable (ViSession instrumentHandle,
                                                ViInt32 numberOfValues,
                                                ViReal64 _VI_FAR receiverFrequencyValues[],
                                                ViReal64 _VI_FAR ENRValues[]);
ViStatus _VI_FUNC rszvl_ConfigureNoiseLossInputSettings (ViSession instrumentHandle,
                                                         ViInt32 selectionMode,
                                                         ViReal64 lossInputConstant);
ViStatus _VI_FUNC rszvl_ConfigureNoiseLossInputTable (ViSession instrumentHandle,
                                                      ViInt32 numberOfValues,
                                                      ViReal64 _VI_FAR receiverFrequencyValues[],
                                                      ViReal64 _VI_FAR lossInputValues[]);
ViStatus _VI_FUNC rszvl_ConfigureNoiseLossOutputSettings (ViSession instrumentHandle,
                                                          ViInt32 selectionMode,
                                                          ViReal64 lossOutputConstant);
ViStatus _VI_FUNC rszvl_ConfigureNoiseLossOutputTable (ViSession instrumentHandle,
                                                       ViInt32 numberOfValues,
                                                       ViReal64 _VI_FAR receiverFrequencyValues[],
                                                       ViReal64 _VI_FAR lossOutputValues[]);
ViStatus _VI_FUNC rszvl_ConfigureNoiseResultSettings (ViSession instrumentHandle,
                                                      ViInt32 combinedTraceDisplay,
                                                      ViBoolean tableFormResults);
ViStatus _VI_FUNC rszvl_ConfigureNoiseTraceSettings (ViSession instrumentHandle,
                                                     ViInt32 yAxis,
                                                     ViBoolean symbols,
                                                     ViBoolean automaticScaling,
                                                     ViReal64 minYAxis,
                                                     ViReal64 maxYAxis);
ViStatus _VI_FUNC rszvl_ConfigureNoiseGainTraceSettings (ViSession instrumentHandle,
                                                         ViBoolean yAxis,
                                                         ViBoolean symbols,
                                                         ViBoolean automaticScaling,
                                                         ViReal64 minYAxis,
                                                         ViReal64 maxYAxis);
ViStatus _VI_FUNC rszvl_ConfigureNoiseTraceMemoryDisplay (ViSession instrumentHandle,
                                                          ViInt32 memory,
                                                          ViBoolean state);
ViStatus _VI_FUNC rszvl_ConfigureNoiseCurrentResultsDisplay
             (ViSession instrumentHandle, ViBoolean state);
ViStatus _VI_FUNC rszvl_ConfigureXAxisFrequencyDisplay (ViSession instrumentHandle,
                                                        ViInt32 measurementResults,
                                                        ViInt32 frequency);
ViStatus _VI_FUNC rszvl_CreateNoiseLimitLine (ViSession instrumentHandle,
                                              ViInt32 limit, ViString name,
                                              ViString comment);
ViStatus _VI_FUNC rszvl_DefineNoiseLimitLine (ViSession instrumentHandle,
                                              ViInt32 limit, ViInt32 resultType,
                                              ViInt32 type, ViInt32 count,
                                              ViReal64 _VI_FAR xAxis[],
                                              ViReal64 _VI_FAR amplitude[]);
ViStatus _VI_FUNC rszvl_EnableNoiseLimitCheck (ViSession instrumentHandle,
                                               ViInt32 limit, ViInt32 type,
                                               ViBoolean limitEnabled,
                                               ViBoolean checkEnabled);
ViStatus _VI_FUNC rszvl_MoveNoiseLimitLine (ViSession instrumentHandle,
                                            ViInt32 limit, ViInt32 type,
                                            ViReal64 value);
ViStatus _VI_FUNC rszvl_CopyNoiseLimitLine (ViSession instrumentHandle,
                                            ViInt32 limit, ViInt32 copyTo);
ViStatus _VI_FUNC rszvl_ClearNoiseLimitLineResults (ViSession instrumentHandle,
                                                    ViInt32 limit);
ViStatus _VI_FUNC rszvl_DeleteNoiseLimitLine (ViSession instrumentHandle,
                                              ViInt32 limit);
ViStatus _VI_FUNC rszvl_SaveNoiseTraceDataToMemory (ViSession instrumentHandle,
                                                    ViInt32 memory);
ViStatus _VI_FUNC rszvl_FetchNoiseArray (ViSession instrumentHandle,
                                         ViInt32 modifier, ViInt32 arraySize,
                                         ViReal64 _VI_FAR values[]);
ViStatus _VI_FUNC rszvl_FetchNoiseScalar (ViSession instrumentHandle,
                                          ViInt32 modifier, ViPReal64 value);
ViStatus _VI_FUNC rszvl_FetchNoiseMemoryArray (ViSession instrumentHandle,
                                               ViInt32 modifier, ViInt32 memory,
                                               ViInt32 arraySize,
                                               ViReal64 _VI_FAR measResult[]);
ViStatus _VI_FUNC rszvl_GetNoiseLimitCheckResult (ViSession instrumentHandle,
                                                  ViInt32 limit, ViPInt32 state);
ViStatus _VI_FUNC rszvl_Set3GPPBSMode (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_Configure3GPPBSMeasurement (ViSession instrumentHandle,
                                                    ViInt32 scramblingCode,
                                                    ViInt32 powerReference,
                                                    ViInt32 syncType,
                                                    ViBoolean normalize,
                                                    ViBoolean HSDPAUPA,
                                                    ViReal64 inactiveChannelTreshold,
                                                    ViBoolean invertQ,
                                                    ViInt32 antenna);
ViStatus _VI_FUNC rszvl_Configure3GPPBSResults (ViSession instrumentHandle,
                                                ViInt32 channel, ViInt32 CPICHSlot);
ViStatus _VI_FUNC rszvl_Configure3GPPBSResultType (ViSession instrumentHandle,
                                                   ViInt32 resultType);
ViStatus _VI_FUNC rszvl_Configure3GPPBSResultSummaryView (ViSession instrumentHandle,
                                                          ViInt32 view);
ViStatus _VI_FUNC rszvl_Configure3GPPBSYScaleDivision (ViSession instrumentHandle,
                                                       ViReal64 yScaleDivision);
ViStatus _VI_FUNC rszvl_Set3GPPBSASequence (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_Search3GPPBSScramblingCode (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_Set3GPPBSMarkerToChannel (ViSession instrumentHandle,
                                                  ViInt32 channel);
ViStatus _VI_FUNC rszvl_Set3GPPBSDeltaMarkerToChannel (ViSession instrumentHandle,
                                                       ViInt32 channel);
ViStatus _VI_FUNC rszvl_Set3GPPBSMeasurement (ViSession instrumentHandle,
                                              ViInt32 measurement);
ViStatus _VI_FUNC rszvl_Get3GPPBSTrace (ViSession instrumentHandle, 
                                        ViInt32 trace, ViInt32 arrayLength,
                                        ViPInt32 actualPoints,
                                        ViReal64 _VI_FAR values[]);
ViStatus _VI_FUNC rszvl_Get3GPPBSResult (ViSession instrumentHandle, ViInt32 type,
                                         ViPReal64 result);
ViStatus _VI_FUNC rszvl_Get3GPPBSCDPResult (ViSession instrumentHandle,
                                            ViInt32 type, ViInt32 arraySize,
                                            ViReal64 _VI_FAR results[],
                                            ViPInt32 returnedValues);
ViStatus _VI_FUNC rszvl_Get3GPPBSCDPScramblingCodeResult (ViSession instrumentHandle,
                                                          ViInt32 arraySize,
                                                          ViInt32 _VI_FAR scrambingCode[],
                                                          ViReal64 _VI_FAR power[],
                                                          ViPInt32 returnedValues);
ViStatus _VI_FUNC rszvl_Get3GPPBSCDPListEvaluationResult (ViSession instrumentHandle,
                                                          ViInt32 noOfValues,
                                                          ViReal64 _VI_FAR rangeNumber[],
                                                          ViReal64 _VI_FAR startFrequency[],
                                                          ViReal64 _VI_FAR stopFrequency[],
                                                          ViReal64 _VI_FAR resolutionBandwidth[],
                                                          ViReal64 _VI_FAR frequency[],
                                                          ViReal64 _VI_FAR powerAbs[],
                                                          ViReal64 _VI_FAR powerRel[],
                                                          ViReal64 _VI_FAR delta[],
                                                          ViInt32 _VI_FAR limitCheck[],
                                                          ViReal64 _VI_FAR reserved1[],
                                                          ViReal64 _VI_FAR reserved2[],
                                                          ViPInt32 returnedValues);
ViStatus _VI_FUNC rszvl_WlanSetWlanMode (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureWlanStandard (ViSession instrumentHandle,
                                               ViInt32 standard);
ViStatus _VI_FUNC rszvl_ConfigureWlanTrigger (ViSession instrumentHandle,
                                              ViInt32 mode, ViReal64 offset,
                                              ViBoolean autoTriggerLevel,
                                              ViReal64 powerTriggerLevel);
ViStatus _VI_FUNC rszvl_ConfigureWlanGate (ViSession instrumentHandle,
                                           ViBoolean gating, ViInt32 mode,
                                           ViReal64 delay, ViReal64 length,
                                           ViBoolean linkGateAndMarker);
ViStatus _VI_FUNC rszvl_ConfigureWlanACPMode (ViSession instrumentHandle,
                                              ViInt32 ACPMode);
ViStatus _VI_FUNC rszvl_ConfigureWlanGainImbalanceUnit (ViSession instrumentHandle,
                                                        ViInt32 unit);
ViStatus _VI_FUNC rszvl_ConfigureWLANSweep (ViSession instrumentHandle,
                                            ViBoolean sweepModeContinuous);
ViStatus _VI_FUNC rszvl_ConfigureWlanAutoLevelOnce (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureWlanBasebandSignalLevel (ViSession instrumentHandle,
                                                          ViBoolean autoLevel,
                                                          ViReal64 autoLevelTime,
                                                          ViReal64 extAtt);
ViStatus _VI_FUNC rszvl_ConfigureWlanRFChannelFrequency (ViSession instrumentHandle,
                                                         ViInt32 channelNo,
                                                         ViReal64 frequency);
ViStatus _VI_FUNC rszvl_ConfigureWlanRFSignalLevel (ViSession instrumentHandle,
                                                    ViBoolean autoLevel,
                                                    ViReal64 autoLevelTime,
                                                    ViReal64 signalLevel,
                                                    ViReal64 RFAtt,
                                                    ViReal64 extAtt);
ViStatus _VI_FUNC rszvl_ConfigureWLANMarkerEnabled (ViSession instrumentHandle,
                                                    ViInt32 window, ViInt32 marker,
                                                    ViBoolean markerEnabled,
                                                    ViInt32 trace);
ViStatus _VI_FUNC rszvl_ConfigureWlanMarkPosition (ViSession instrumentHandle,
                                                   ViInt32 window, ViInt32 marker,
                                                   ViInt32 position,
                                                   ViReal64 value);
ViStatus _VI_FUNC rszvl_ConfigureWlanMarkerPositionConstellation
             (ViSession instrumentHandle, ViInt32 window, ViInt32 marker,
              ViInt32 type, ViInt32 position);
ViStatus _VI_FUNC rszvl_ConfigureWlanMarkerZoom (ViSession instrumentHandle,
                                                 ViInt32 window, ViInt32 marker,
                                                 ViReal64 zoomRatio);
ViStatus _VI_FUNC rszvl_ConfigureWlanMarkerBurstSymbol (ViSession instrumentHandle,
                                                        ViInt32 window,
                                                        ViInt32 marker,
                                                        ViInt32 burstNumber,
                                                        ViInt32 symbolNumber);
ViStatus _VI_FUNC rszvl_SetWlanMarkerPeak (ViSession instrumentHandle,
                                           ViInt32 window, ViInt32 marker,
                                           ViInt32 position);
ViStatus _VI_FUNC rszvl_WlanDisableAllMarkers (ViSession instrumentHandle,
                                               ViInt32 window);
ViStatus _VI_FUNC rszvl_ConfigureWlanDemodulation (ViSession instrumentHandle,
                                                   ViBoolean useSignalFieldContent,
                                                   ViInt32 burstType,
                                                   ViBoolean autoDemodulation,
                                                   ViInt32 demodulator,
                                                   ViBoolean channelEstimation,
                                                   ViBoolean equalBurstLength,
                                                   ViInt32 minNumber_ofDataSymbols,
                                                   ViInt32 maxNumber_ofDataSymbols);
ViStatus _VI_FUNC rszvl_ConfigureWlanPLCPDemodulation (ViSession instrumentHandle,
                                                       ViBoolean usePLCPHeaderContent,
                                                       ViInt32 burstType,
                                                       ViBoolean autoDemodulation,
                                                       ViInt32 demodulator,
                                                       ViInt32 lengthType,
                                                       ViBoolean equalBurstLength,
                                                       ViInt32 minPayloadLength,
                                                       ViInt32 maxPayloadLength);
ViStatus _VI_FUNC rszvl_ConfigureWlanTracking (ViSession instrumentHandle,
                                               ViBoolean phase, ViBoolean timing,
                                               ViBoolean level);
ViStatus _VI_FUNC rszvl_ConfigureWlanSelectFilter (ViSession instrumentHandle,
                                                   ViString TXFilterName,
                                                   ViString RXFilterName);
ViStatus _VI_FUNC rszvl_GetWlanFilterCatalog (ViSession instrumentHandle,
                                              ViInt32 bufferSize,
                                              ViChar _VI_FAR filterList[],
                                              ViPInt32 returnedLength);
ViStatus _VI_FUNC rszvl_ConfigureWlanDisplay (ViSession instrumentHandle,
                                              ViInt32 combinedTraceDisplay,
                                              ViBoolean displayTable);
ViStatus _VI_FUNC rszvl_ConfigureWlanLimit (ViSession instrumentHandle,
                                            ViInt32 window, ViInt32 limit,
                                            ViInt32 limitType, ViInt32 mode,
                                            ViReal64 value);
ViStatus _VI_FUNC rszvl_ConfigureWLANAllLimits (ViSession instrumentHandle,
                                                ViInt32 window, ViInt32 limit,
                                                ViInt32 arraySize,
                                                ViReal64 _VI_FAR values[]);
ViStatus _VI_FUNC rszvl_ConfigureWlanPLCPLimit (ViSession instrumentHandle,
                                                ViInt32 window, ViInt32 limit,
                                                ViInt32 limitType, ViInt32 mode,
                                                ViReal64 value);
ViStatus _VI_FUNC rszvl_GetWlanACPChannelLimit (ViSession instrumentHandle,
                                                ViInt32 window, ViInt32 limit,
                                                ViInt32 channelType,
                                                ViPReal64 value);
ViStatus _VI_FUNC rszvl_GetWlanSpectrumMaxOverstepping (ViSession instrumentHandle,
                                                        ViInt32 window, ViInt32 limit,
                                                        ViInt32 channelType,
                                                        ViPReal64 value);
ViStatus _VI_FUNC rszvl_GetWlanAllLimits (ViSession instrumentHandle,
                                          ViInt32 window, ViInt32 limit,
                                          ViReal64 _VI_FAR values[]);
ViStatus _VI_FUNC rszvl_ConfigureWlanMeas (ViSession instrumentHandle,
                                           ViBoolean overallBurstCount,
                                           ViInt32 no_ofBurst_toAnalyze,
                                           ViReal64 captureTime, ViBoolean swapIQ,
                                           ViInt32 sweepCount);
ViStatus _VI_FUNC rszvl_ConfigureWlanPVTMeas (ViSession instrumentHandle,
                                              ViInt32 risingFalling);
ViStatus _VI_FUNC rszvl_ConfigureWlanPLCPPVT (ViSession instrumentHandle,
                                              ViInt32 referencePower, ViInt32 ramp,
                                              ViInt32 averageLength);
ViStatus _VI_FUNC rszvl_ConfigureWlanEVMMeasurement (ViSession instrumentHandle,
                                                     ViInt32 mode,
                                                     ViBoolean autoScaling,
                                                     ViReal64 perDivision,
                                                     ViInt32 unit);
ViStatus _VI_FUNC rszvl_ConfigureWlanSpectrumMeasurement (ViSession instrumentHandle,
                                                          ViInt32 mode);
ViStatus _VI_FUNC rszvl_ConfigureWlanConstellationMeasurement
             (ViSession instrumentHandle, ViInt32 constellationMode,
              ViInt32 carrierSelection, ViInt32 carrierNumber);
ViStatus _VI_FUNC rszvl_ConfigureWlanPLCPConstellation (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureWlanStatisticsMeasurement
             (ViSession instrumentHandle, ViInt32 mode);
ViStatus _VI_FUNC rszvl_ConfigureWlanIQMeasurementParameters
             (ViSession instrumentHandle, ViReal64 sampleRate);
ViStatus _VI_FUNC rszvl_GetWlanACPMeasurement (ViSession instrumentHandle,
                                               ViInt32 window, ViInt32 marker,
                                               ViInt32 mode,
                                               ViReal64 _VI_FAR values[]);
ViStatus _VI_FUNC rszvl_LoadWlanIQData (ViSession instrumentHandle,
                                        ViString fileName);
ViStatus _VI_FUNC rszvl_StoreWlanIQData (ViSession instrumentHandle,
                                         ViString fileName);
ViStatus _VI_FUNC rszvl_WlanBurstImm (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ReadWlanTraceData (ViSession instrumentHandle,
                                           ViInt32 trace, ViInt32 arraySize,
                                           ViReal64 _VI_FAR traceData[],
                                           ViPInt32 no_ofValues);
ViStatus _VI_FUNC rszvl_ReadWlanMemoryIQData (ViSession instrumentHandle,
                                              ViInt32 offsetSamples,
                                              ViInt32 no_ofSamples,
                                              ViInt32 bufferSize,
                                              ViPInt32 no_ofPoints,
                                              ViReal64 _VI_FAR realParts_I[],
                                              ViReal64 _VI_FAR imaginaryParts_Q[]);
ViStatus _VI_FUNC rszvl_ReadWlanSEMResults (ViSession instrumentHandle,
                                            ViInt32 noOfValues,
                                            ViReal64 _VI_FAR index[],
                                            ViReal64 _VI_FAR startFrequencyBand[],
                                            ViReal64 _VI_FAR stopFrequencyBand[],
                                            ViReal64 _VI_FAR resolutionBandwidth[],
                                            ViReal64 _VI_FAR limitFailFrequency[],
                                            ViReal64 _VI_FAR powerAbs[],
                                            ViReal64 _VI_FAR powerRel[],
                                            ViReal64 _VI_FAR limitDistance[],
                                            ViReal64 _VI_FAR failureFlag[],
                                            ViPInt32 returnedValues);
ViStatus _VI_FUNC rszvl_FetchWlanBurstAll (ViSession instrumentHandle,
                                           ViReal64 _VI_FAR results[],
                                           ViPInt32 returnedValues);
ViStatus _VI_FUNC rszvl_FetchWlanBurstPower (ViSession instrumentHandle,
                                             ViInt32 powerMeasurement,
                                             ViPReal64 value);
ViStatus _VI_FUNC rszvl_FetchWlanBurstTime (ViSession instrumentHandle,
                                            ViInt32 time, ViInt32 mode,
                                            ViPReal64 value);
ViStatus _VI_FUNC rszvl_FetchWlanBurstError (ViSession instrumentHandle,
                                             ViInt32 errorType, ViInt32 mode,
                                             ViPReal64 value);
ViStatus _VI_FUNC rszvl_FetchWlanIQImp (ViSession instrumentHandle,
                                        ViInt32 IQImpairment, ViInt32 mode,
                                        ViPReal64 value);
ViStatus _VI_FUNC rszvl_FetchWlanEVM (ViSession instrumentHandle, ViInt32 EVM,
                                      ViInt32 mode, ViPReal64 value);
ViStatus _VI_FUNC rszvl_FetchWlanCount (ViSession instrumentHandle, ViInt32 type,
                                        ViPInt32 count);
ViStatus _VI_FUNC rszvl_GetWlanLimitCheckResult (ViSession instrumentHandle,
                                                 ViInt32 window, ViInt32 limit,
                                                 ViInt32 limitType, ViInt32 mode,
                                                 ViPInt32 result);
ViStatus _VI_FUNC rszvl_GetWlanPLCPLimitCheckResult (ViSession instrumentHandle,
                                                     ViInt32 window, ViInt32 limit,
                                                     ViInt32 limitType,
                                                     ViInt32 mode, ViPInt32 result);
ViStatus _VI_FUNC rszvl_GetWlanACPChannLimitCheckResult (ViSession instrumentHandle,
                                                         ViInt32 window, ViInt32 limit,
                                                         ViInt32 channelType,
                                                         ViPInt32 result);
ViStatus _VI_FUNC rszvl_GetWlanAllLimitCheckResults (ViSession instrumentHandle,
                                                     ViInt32 window, ViInt32 limit,
                                                     ViInt32 _VI_FAR results[]);
ViStatus _VI_FUNC rszvl_GetWlanLimitCheck (ViSession instrumentHandle,
                                           ViInt32 window, ViInt32 limit,
                                           ViPInt32 result);
ViStatus _VI_FUNC rszvl_SetWiMAXMode (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXStandard (ViSession instrumentHandle,
                                                ViInt32 standard);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXTrigger (ViSession instrumentHandle,
                                               ViInt32 mode, ViReal64 offset,
                                               ViBoolean autoTriggerLevel,
                                               ViReal64 powerTriggerLevel,
											   ViReal64    externalTriggerLevel);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXGate (ViSession instrumentHandle,
                                            ViBoolean gating, ViInt32 mode,
                                            ViReal64 delay, ViReal64 length,
                                            ViBoolean linkGateAndMarker);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXACP (ViSession instrumentHandle,
                                           ViInt32 ACPMode,
                                           ViInt32 numberOfChannels,
                                           ViBoolean noiseCorrection);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXACPChannelBandwidth
             (ViSession instrumentHandle, ViInt32 channel, ViInt32 channelNumber,
              ViReal64 bandwidth);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXACPChannelSpacing (ViSession instrumentHandle,
                                                         ViInt32 channel,
                                                         ViInt32 channelNumber,
                                                         ViReal64 spacing);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXLevel (ViSession instrumentHandle,
                                             ViBoolean autoLevel,
                                             ViReal64 referenceLevel,
                                             ViReal64 externalAttenuation);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXSEMMode (ViSession instrumentHandle,
                                               ViInt32 linkMode);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXAutoLevelOnce (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXBasebandSignalLevel
             (ViSession instrumentHandle, ViBoolean autoLevel,
              ViReal64 autoLevelTime, ViReal64 extAtt);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXRFChannelFrequency (ViSession instrumentHandle,
                                                          ViInt32 channelNo,
                                                          ViReal64 frequency);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXRFSignalLevel (ViSession instrumentHandle,
                                                     ViBoolean autoLevel,
                                                     ViReal64 autoLevelTime,
                                                     ViReal64 signalLevel,
                                                     ViReal64 RFAtt,
                                                     ViReal64 extAtt,
													 ViInt32  signalInput);
ViStatus _VI_FUNC rszvl_GetWiMAXReferenceLevel (ViSession instrumentHandle,
                                                ViPReal64 referenceLevel);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXMarkerEnabled (ViSession instrumentHandle,
                                                     ViInt32 window, ViInt32 marker,
                                                     ViBoolean markerEnabled,
                                                     ViInt32 trace);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXMarkPosition (ViSession instrumentHandle,
                                                    ViInt32 window, ViInt32 marker,
                                                    ViReal64 value);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXMarkerPositionConstellation
             (ViSession instrumentHandle, ViInt32 window, ViInt32 marker,
              ViInt32 type, ViInt32 position);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXMarkerZoom (ViSession instrumentHandle,
                                                  ViInt32 window,
                                                  ViReal64 zoomRatio);
ViStatus _VI_FUNC rszvl_SetWiMAXMarkerPeak (ViSession instrumentHandle,
                                            ViInt32 window, ViInt32 marker,
                                            ViInt32 position);
ViStatus _VI_FUNC rszvl_WiMAXDisableAllMarkers (ViSession instrumentHandle,
                                                ViInt32 window);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXDemodulation (ViSession instrumentHandle,
                                                    ViInt32 modulationDetectionMode,
                                                    ViInt32 demodulator,
                                                    ViInt32 linkMode,
                                                    ViBoolean channelEstimation,
                                                    ViBoolean equalBurstLength,
                                                    ViInt32 minNumber_ofDataSymbols,
                                                    ViInt32 maxNumber_ofDataSymbols);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXDemodulationOFDMA (ViSession instrumentHandle,
                                                         ViInt32 channelEstimation,
                                                         ViReal64 FFTOffset);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXSubchannelization (ViSession instrumentHandle,
                                                         ViBoolean state,
                                                         ViInt32 subchannel,
                                                         ViInt32 upLinkPhysicalModifier);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXTracking (ViSession instrumentHandle,
                                                ViBoolean phase, ViBoolean timing,
                                                ViBoolean level);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXOFDMATracking (ViSession instrumentHandle,
                                                     ViBoolean phase,
                                                     ViBoolean timing,
                                                     ViInt32 trackingPilot);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXSelectFilter (ViSession instrumentHandle,
                                                    ViString TXFilterName,
                                                    ViString RXFilterName);
ViStatus _VI_FUNC rszvl_GetWiMAXFilterCatalog (ViSession instrumentHandle,
                                               ViInt32 bufferSize,
                                               ViChar _VI_FAR filterList[],
                                               ViPInt32 returnedLength);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXDisplay (ViSession instrumentHandle,
                                               ViInt32 combinedTraceDisplay,
                                                ViBoolean displayTable);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXLimit (ViSession instrumentHandle,
                                             ViInt32 window, ViInt32 limit,
                                             ViInt32 limitType, ViInt32 mode,
                                             ViReal64 value);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXAllLimits (ViSession instrumentHandle,
                                                 ViInt32 window, ViInt32 limit,
                                                 ViInt32 arraySize,
                                                 ViReal64 _VI_FAR values[]);
ViStatus _VI_FUNC rszvl_GetWiMAXSpectrumMaxOverstepping (ViSession instrumentHandle,
                                                         ViInt32 window,
                                                         ViInt32 limit,
                                                         ViInt32 channelType,
                                                         ViPReal64 value);
ViStatus _VI_FUNC rszvl_GetWiMAXAllLimits (ViSession instrumentHandle,
                                           ViInt32 window, ViInt32 limit,
                                           ViReal64 _VI_FAR values[]);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXFrame (ViSession instrumentHandle,
                                             ViInt32 IDCell,
                                             ViInt32 ULSubframeFrameNumber,
                                             ViReal64 TDDFrame);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXFrameOFDMAMAP (ViSession instrumentHandle,
                                                     ViInt32 use_forAnalysis);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXDLFramePreambleIndex
             (ViSession instrumentHandle, ViInt32 mode, ViInt32 preamblePattern);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXFrameDLUsedSubchannel
             (ViSession instrumentHandle, ViInt32 segment, ViInt32 subchannelUsage);
ViStatus _VI_FUNC rszvl_LoadWiMAXFrameData (ViSession instrumentHandle,
                                            ViString fileName);
ViStatus _VI_FUNC rszvl_StoreWiMAXFrameData (ViSession instrumentHandle,
                                             ViString fileName);
ViStatus _VI_FUNC rszvl_CreateWiMAXNewZone (ViSession instrumentHandle,
                                            ViInt32 zone, ViBoolean analyze,
                                            ViInt32 type, ViInt32 segment,
                                            ViInt32 length, ViInt32 offset,
                                            ViInt32 permbase, ViInt32 PRBS_ID);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXZone (ViSession instrumentHandle,
                                            ViInt32 zone, ViBoolean analyze,
                                            ViInt32 type, ViInt32 segment,
                                            ViInt32 length, ViInt32 offset,
                                            ViInt32 permbase, ViInt32 PRBS_ID);
ViStatus _VI_FUNC rszvl_CreateWiMAXNewZoneBurst (ViSession instrumentHandle,
                                                 ViInt32 zone, ViInt32 burst,
                                                 ViInt32 modulation,
                                                 ViInt32 subchannels,
                                                 ViInt32 symbols,
                                                 ViInt32 slotDuration,
                                                 ViInt32 subchannelOffset,
                                                 ViInt32 symbolOffset,
                                                 ViReal64 burstPower,
                                                 ViInt32 burstType);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXZoneBurst (ViSession instrumentHandle,
                                                 ViInt32 zone, ViInt32 burst,
                                                 ViInt32 modulation,
                                                 ViInt32 subchannels,
                                                 ViInt32 symbols,
                                                 ViInt32 slotDuration,
                                                 ViReal64 burstPower,
                                                 ViInt32 burstType,
                                                 ViBoolean offsetAuto,
                                                 ViInt32 subchannelOffset,
                                                 ViInt32 symbolOffset);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXExtFrameZoneGeneratorAddress
             (ViSession instrumentHandle, ViString TCPIPAddress);
ViStatus _VI_FUNC rszvl_GetWiMAXZoneCount (ViSession instrumentHandle,
                                           ViPInt32 count);
ViStatus _VI_FUNC rszvl_GetWiMAXZoneBurstCount (ViSession instrumentHandle,
                                                ViInt32 zone, ViPInt32 count);
ViStatus _VI_FUNC rszvl_WiMAXZoneDelete (ViSession instrumentHandle, ViInt32 zone);
ViStatus _VI_FUNC rszvl_WiMAXZoneBurstDelete (ViSession instrumentHandle,
                                              ViInt32 zone, ViInt32 burst);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXMeas (ViSession instrumentHandle,
                                            ViBoolean overallBurstCount,
                                            ViInt32 no_ofBurst_toAnalyze,
                                            ViReal64 captureTime, ViBoolean swapIQ,
                                            ViInt32 sweepCount,
                                            ViBoolean sweepTimeAuto_MaskACP,
                                            ViReal64 sweepTime_MaskACP,
                                            ViInt32 frequencyBand,
                                            ViReal64 channelBandwidth,
                                            ViInt32 guardRatio);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXMeasOFDMA (ViSession instrumentHandle,
                                                 ViReal64 captureTime,
                                                 ViBoolean swapIQ,
                                                 ViInt32 sweepCount,
                                                 ViBoolean sweepTimeAuto_MaskACP,
                                                 ViReal64 sweepTime_MaskACP,
                                                 ViInt32 frequencyBand,
                                                 ViReal64 channelBandwidth,
                                                 ViInt32 guardRatio,
                                                 ViInt32 FFTSize,
                                                 ViInt32 number_ofSubframes_toAnalyze);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXMeasWiBro (ViSession instrumentHandle,
                                                 ViReal64 captureTime,
                                                 ViBoolean swapIQ,
                                                 ViInt32 sweepCount,
                                                 ViBoolean sweepTimeAuto_MaskACP,
                                                 ViReal64 sweepTime_MaskACP,
                                                 ViReal64 channelBandwidth,
                                                 ViInt32 number_ofSubframes_toAnalyze);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXPVTMeas (ViSession instrumentHandle,
                                               ViInt32 risingFalling,
                                               ViInt32 burstSelection);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXPVTMeasOFDMA (ViSession instrumentHandle,
                                                    ViInt32 risingFalling);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXEVMMeasurement (ViSession instrumentHandle,
                                                      ViInt32 mode,
                                                      ViBoolean autoScaling,
                                                      ViReal64 perDivision,
                                                      ViInt32 unit);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXSpectrumMeasurement
             (ViSession instrumentHandle, ViInt32 mode);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXConstellationMeasurement
             (ViSession instrumentHandle, ViInt32 constellationMode,
              ViInt32 carrierSelection, ViInt32 carrierNumber);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXConstellationSelection
             (ViSession instrumentHandle, ViInt32 modulation,
              ViInt32 burstSelection, ViInt32 burstNumber, ViInt32 carrierSelection,
              ViInt32 carrierNumber, ViInt32 symbolSelection, ViInt32 symbolNumber);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXBitstreamSelection (ViSession instrumentHandle,
                                                          ViInt32 modulation,
                                                          ViInt32 burstSelection,
                                                          ViInt32 burstNumber,
                                                          ViInt32 symbolSelection,
                                                          ViInt32 symbolNumber);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXStatisticsMeasurement
             (ViSession instrumentHandle, ViInt32 mode);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXIQMeasurementParameters
             (ViSession instrumentHandle, ViReal64 sampleRate);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXSEMMeasurement (ViSession instrumentHandle,
                                                      ViInt32 powerClass);
ViStatus _VI_FUNC rszvl_GetWiMAXACPMeasurement (ViSession instrumentHandle,
                                                ViInt32 window, ViInt32 marker,
                                                ViInt32 mode,
                                                ViReal64 _VI_FAR values[]);
ViStatus _VI_FUNC rszvl_LoadWiMAXIQData (ViSession instrumentHandle,
                                         ViString fileName);
ViStatus _VI_FUNC rszvl_StoreWiMAXIQData (ViSession instrumentHandle,
                                          ViString fileName);
ViStatus _VI_FUNC rszvl_WiMAXBurstImm (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_WiMAXRefreshIQMeasurement (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ReadWiMAXTraceData (ViSession instrumentHandle,
                                            ViInt32 trace, ViInt32 arraySize,
                                            ViReal64 _VI_FAR traceData[],
                                            ViPInt32 no_ofValues);
ViStatus _VI_FUNC rszvl_ReadWiMAXMemoryIQData (ViSession instrumentHandle,
                                               ViInt32 offsetSamples,
                                               ViInt32 no_ofSamples,
                                               ViInt32 bufferSize,
                                               ViPInt32 no_ofPoints,
                                               ViReal64 _VI_FAR realParts_I[],
                                               ViReal64 _VI_FAR imaginaryParts_Q[]);
ViStatus _VI_FUNC rszvl_ReadWiMAXSEMResults (ViSession instrumentHandle,
                                             ViInt32 noOfValues,
                                             ViReal64 _VI_FAR index[],
                                             ViReal64 _VI_FAR startFrequencyBand[],
                                             ViReal64 _VI_FAR stopFrequencyBand[],
                                             ViReal64 _VI_FAR resolutionBandwidth[],
                                             ViReal64 _VI_FAR limitFailFrequency[],
                                             ViReal64 _VI_FAR powerAbs[],
                                             ViReal64 _VI_FAR powerRel[],
                                             ViReal64 _VI_FAR limitDistance[],
                                             ViReal64 _VI_FAR failureFlag[],
                                             ViPInt32 returnedValues);
ViStatus _VI_FUNC rszvl_FetchWiMAXBurstAll (ViSession instrumentHandle,
                                            ViReal64 _VI_FAR results[],
                                            ViPInt32 returnedValues);
ViStatus _VI_FUNC rszvl_FetchWiMAXBurstPower (ViSession instrumentHandle,
                                              ViInt32 powerMeasurement,
                                              ViInt32 mode, ViPReal64 value);
ViStatus _VI_FUNC rszvl_FetchWiMAXBurstError (ViSession instrumentHandle,
                                              ViInt32 errorType, ViInt32 mode,
                                              ViPReal64 value);
ViStatus _VI_FUNC rszvl_FetchWiMAXIQImp (ViSession instrumentHandle,
                                         ViInt32 IQImpairment, ViInt32 mode,
                                         ViPReal64 value);
ViStatus _VI_FUNC rszvl_FetchWiMAXEVM (ViSession instrumentHandle, ViInt32 EVM,
                                       ViInt32 mode, ViPReal64 value);
ViStatus _VI_FUNC rszvl_FetchWiMAXCINRValues (ViSession instrumentHandle,
                                              ViReal64 _VI_FAR values[]);
ViStatus _VI_FUNC rszvl_FetchWiMAXRSSIValues (ViSession instrumentHandle,
                                              ViReal64 _VI_FAR values[]);
ViStatus _VI_FUNC rszvl_FetchWiMAXCount (ViSession instrumentHandle, ViInt32 type,
                                         ViPInt32 count);
ViStatus _VI_FUNC rszvl_FetchWiMAXTimeDomainPreamble (ViSession instrumentHandle,
                                                      ViInt32 result,
                                                      ViPReal64 minimum,
                                                      ViPReal64 average,
                                                      ViPReal64 maximum);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXListResultUnit (ViSession instrumentHandle,
                                                      ViInt32 unit);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXRSSICINRAvg (ViSession instrumentHandle,
                                                   ViReal64 averagingParameter);
ViStatus _VI_FUNC rszvl_GetWiMAXLimitCheckResult (ViSession instrumentHandle,
                                                  ViInt32 window, ViInt32 limit,
                                                  ViInt32 limitType, ViInt32 mode,
                                                  ViPInt32 result);
ViStatus _VI_FUNC rszvl_GetWiMAXAllLimitCheckResults (ViSession instrumentHandle,
                                                      ViInt32 window, ViInt32 limit,
                                                      ViInt32 _VI_FAR results[]);
ViStatus _VI_FUNC rszvl_GetWiMAXLimitCheck (ViSession instrumentHandle,
                                            ViInt32 window, ViInt32 limit,
                                            ViPInt32 result);
ViStatus _VI_FUNC rszvl_SelectWlanScreen (ViSession instrumentHandle, ViInt32 window);
ViStatus _VI_FUNC rszvl_SelectWiMAXScreen (ViSession instrumentHandle, ViInt32 window);
ViStatus _VI_FUNC rszvl_ConfigureWiMAXSEMTTA (ViSession instrumentHandle, ViInt32 TTA);
ViStatus _VI_FUNC rszvl_Configure3GPPBSESP (ViSession instrumentHandle,
                                            ViInt32 limitLineType,
                                            ViInt32 limitLineValue,
                                            ViReal64 _30kHz1MHzTransition);
ViStatus _VI_FUNC rszvl_Restore3GPPBSESPLimitLines (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_Adjust3GPPBSLevel (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_ConfigureNoiseMarkerEnabled (ViSession instrumentHandle,
                                                     ViBoolean markerEnabled,
                                                     ViInt32 trace);
ViStatus _VI_FUNC rszvl_NoiseMoveMarker (ViSession instrumentHandle,
                                         ViReal64 markerPosition);
ViStatus _VI_FUNC rszvl_SetNoiseMarkerPeak (ViSession instrumentHandle,
                                            ViInt32 position);
ViStatus _VI_FUNC rszvl_QueryNoiseMarker (ViSession instrumentHandle,
                                          ViReal64 *markerAmplitude);
ViStatus _VI_FUNC rszvl_NoiseDisableAllMarkers (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_WiMAXGetMarkerTTC (ViSession instrumentHandle,
                                           ViInt32 window, ViInt32 markerNumber,
                                           ViReal64 *time);
ViStatus _VI_FUNC rszvl_GetWiMAXMarkAmplitude (ViSession instrumentHandle,
                                               ViInt32 window, ViInt32 marker,
                                               ViReal64 *amplitude);
ViStatus _VI_FUNC rszvl_ConfigureReferenceOscillator (ViSession instrumentHandle,
                                                      ViInt32 reference);
ViStatus _VI_FUNC rszvl_setStatusRegister (ViSession instrumentHandle,
                                           ViInt32 registerOperation,
                                           ViInt32 questionableRegister,
                                           ViInt32 enable, ViInt32 PTransition,
                                           ViInt32 NTransition);
ViStatus _VI_FUNC rszvl_getStatusRegister (ViSession instrumentHandle,
                                           ViInt32 statusRegistersQuery,
                                           ViPInt32 registerValue);
ViStatus _VI_FUNC rszvl_FileManagerOperations (ViSession instrumentHandle,
                                               ViInt32 operation, ViString source,
                                               ViString destination);
ViStatus _VI_FUNC rszvl_GetCurrentDirectory (ViSession instrumentHandle,
                                             ViChar _VI_FAR currentDirectory[]);
ViStatus _VI_FUNC rszvl_DataSetFileOperations (ViSession instrumentHandle,
                                               ViInt32 operation, ViString path);
ViStatus _VI_FUNC rszvl_DataSetFileClearAll (ViSession instrumentHandle,
                                             ViString directory);
ViStatus _VI_FUNC rszvl_SetAttributeViInt32 (ViSession instrumentHandle,
                                             ViString repeatedCapabilityName,
                                             ViUInt32 attributeID,
                                             ViInt32 attributeValue);
ViStatus _VI_FUNC rszvl_SetAttributeViReal64 (ViSession instrumentHandle,
                                              ViString repeatedCapabilityName,
                                              ViUInt32 attributeID,
                                              ViReal64 attributeValue);
ViStatus _VI_FUNC rszvl_SetAttributeViString (ViSession instrumentHandle,
                                              ViString repeatedCapabilityName,
                                              ViUInt32 attributeID,
                                              ViString attributeValue);
ViStatus _VI_FUNC rszvl_SetAttributeViBoolean (ViSession instrumentHandle,
                                               ViString repeatedCapabilityName,
                                               ViUInt32 attributeID,
                                               ViBoolean attributeValue);
ViStatus _VI_FUNC rszvl_SetAttributeViSession (ViSession instrumentHandle,
                                               ViString repeatedCapabilityName,
                                               ViUInt32 attributeID,
                                               ViSession attributeValue);
ViStatus _VI_FUNC rszvl_GetAttributeRepeatedCapabilityIds
             (ViSession instrumentHandle, ViUInt32 attributeID, ViInt32 bufferSize,
              ViChar _VI_FAR repeatedCapabilityId_s[]);
ViStatus _VI_FUNC rszvl_GetAttributeRepeatedCapabilityIdNames
             (ViSession instrumentHandle, ViUInt32 attributeID,
              ViString repeatedCapabilityId, ViInt32 bufferSize,
              ViChar _VI_FAR repeatedCapabilityIdName_s[]);
ViStatus _VI_FUNC rszvl_GetAttributeViInt32 (ViSession instrumentHandle,
                                             ViString repeatedCapabilityName,
                                             ViUInt32 attributeID,
                                             ViPInt32 attributeValue);
ViStatus _VI_FUNC rszvl_GetAttributeViReal64 (ViSession instrumentHandle,
                                              ViString repeatedCapabilityName,
                                              ViUInt32 attributeID,
                                              ViPReal64 attributeValue);
ViStatus _VI_FUNC rszvl_GetAttributeViString (ViSession instrumentHandle,
                                              ViString repeatedCapabilityName,
                                              ViUInt32 attributeID,
                                              ViInt32 bufferSize,
                                              ViChar _VI_FAR attributeValue[]);
ViStatus _VI_FUNC rszvl_GetAttributeViBoolean (ViSession instrumentHandle,
                                               ViString repeatedCapabilityName,
                                               ViUInt32 attributeID,
                                               ViPBoolean attributeValue);
ViStatus _VI_FUNC rszvl_GetAttributeViSession (ViSession instrumentHandle,
                                               ViString repeatedCapabilityName,
                                               ViUInt32 attributeID,
                                               ViPSession attributeValue);
ViStatus _VI_FUNC rszvl_CheckAttributeViInt32 (ViSession instrumentHandle,
                                               ViString repeatedCapabilityName,
                                               ViUInt32 attributeID,
                                               ViInt32 attributeValue);
ViStatus _VI_FUNC rszvl_CheckAttributeViReal64 (ViSession instrumentHandle,
                                                ViString repeatedCapabilityName,
                                                ViUInt32 attributeID,
                                                ViReal64 attributeValue);
ViStatus _VI_FUNC rszvl_CheckAttributeViString (ViSession instrumentHandle,
                                                ViString repeatedCapabilityName,
                                                ViUInt32 attributeID,
                                                ViString attributeValue);
ViStatus _VI_FUNC rszvl_CheckAttributeViBoolean (ViSession instrumentHandle,
                                                 ViString repeatedCapabilityName,
                                                 ViUInt32 attributeID,
                                                 ViBoolean attributeValue);
ViStatus _VI_FUNC rszvl_CheckAttributeViSession (ViSession instrumentHandle,
                                                 ViString repeatedCapabilityName,
                                                 ViUInt32 attributeID,
                                                 ViSession attributeValue);
ViStatus _VI_FUNC rszvl_setCheckOption (ViSession instrumentHandle,
                                        ViBoolean optionChecking);
ViStatus _VI_FUNC rszvl_setCheckRange (ViSession instrumentHandle,
                                       ViBoolean rangeChecking);
ViStatus _VI_FUNC rszvl_reset (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_self_test (ViSession instrumentHandle,
                                   ViPInt16 selfTestResult,
                                   ViChar _VI_FAR selfTestMessage[]);
ViStatus _VI_FUNC rszvl_SelfTestResults (ViSession instrumentHandle,
                                         ViInt32 bufferSize,
                                         ViChar _VI_FAR results[]);
ViStatus _VI_FUNC rszvl_revision_query (ViSession instrumentHandle,
                                        ViChar _VI_FAR instrumentDriverRevision[],
                                        ViChar _VI_FAR firmwareRevision[]);
ViStatus _VI_FUNC rszvl_SystemVersion (ViSession instrumentHandle, ViInt32 length,
                                       ViChar _VI_FAR systemVersion[]);
ViStatus _VI_FUNC rszvl_error_query (ViSession instrumentHandle, ViPInt32 errorCode,
                                     ViChar _VI_FAR errorMessage[]);
ViStatus _VI_FUNC rszvl_ErrorList (ViSession instrumentHandle, ViInt32 bufferSize,
                                   ViChar _VI_FAR errors[]);
ViStatus _VI_FUNC rszvl_error_message (ViSession instrumentHandle,
                                       ViStatus errorCode,
                                       ViChar _VI_FAR errorMessage[]);
ViStatus _VI_FUNC rszvl_ClearInstrumentErrors (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_Preset (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_GetError (ViSession instrumentHandle, ViPStatus code,
                                  ViInt32 bufferSize, ViChar _VI_FAR description[]);
ViStatus _VI_FUNC rszvl_ClearError (ViSession instrumentHandle);
ViStatus _VI_FUNC rszvl_WriteInstrData (ViSession instrumentHandle,
                                        ViString writeBuffer);
ViStatus _VI_FUNC rszvl_WriteFromFileToInstrument (ViSession instrumentHandle,
                                                   ViString source,
                                                   ViString destination);
ViStatus _VI_FUNC rszvl_ReadInstrData (ViSession instrumentHandle,
                                       ViInt32 number_ofBytesToRead,
                                       ViChar _VI_FAR readBuffer[],
                                       ViPUInt32 numBytesRead);
ViStatus _VI_FUNC rszvl_ReadToFileFromInstrument (ViSession instrumentHandle,
                                                  ViString source,
                                                  ViString destination);
ViStatus _VI_FUNC rszvl_QueryViBoolean (ViSession instrumentHandle,
                                        ViString command, ViPBoolean value);
ViStatus _VI_FUNC rszvl_QueryViInt32 (ViSession instrumentHandle, ViString command,
                                      ViPInt32 value);
ViStatus _VI_FUNC rszvl_QueryViReal64 (ViSession instrumentHandle, ViString command,
                                       ViPReal64 value);
ViStatus _VI_FUNC rszvl_QueryViString (ViSession instrumentHandle, ViString command,
                                       ViInt32 bufferSize, ViChar _VI_FAR value[]);
ViStatus _VI_FUNC rszvl_close (ViSession instrumentHandle);

     /*********************************************************
        Functions reserved for class driver use only.
        End-users should not call these functions.
     *********************************************************/

ViStatus _VI_FUNC  rszvl_RsClose (ViSession vi);

/****************************************************************************
 *------------------------ Error And Completion Codes ----------------------*
 ****************************************************************************/

#define RSZVL_WARN_MEASURE_UNCALIBRATED    (RS_CLASS_WARN_BASE + 0x0001L)
#define RSZVL_WARN_OVER_RANGE              (RS_CLASS_WARN_BASE + 0x0002L)

#define RSZVL_ERROR_MARKER_NOT_ENABLED     (RS_CLASS_ERROR_BASE + 0x0003L)
#define RSZVL_ERROR_NOT_DELTA_MARKER       (RS_CLASS_ERROR_BASE + 0x0002L)
#define RSZVL_ERROR_MAX_TIME_EXCEEDED      (RS_CLASS_ERROR_BASE + 0x0001L)

#define RSZVL_WARNMSG_MEASURE_UNCALIBRATED "Uncalibrated measurement"
#define RSZVL_WARNMSG_OVER_RANGE           "Measurement overrange"

#define RSZVL_ERRMSG_MARKER_NOT_ENABLED    "Marker Not Enabled"
#define RSZVL_ERRMSG_NOT_DELTA_MARKER      "Not Delta Marker"
#define RSZVL_ERRMSG_MAX_TIME_EXCEEDED     "Max Time Exceeded"

#define RSZVL_ERROR_CODES_AND_MSGS \
    {RSZVL_WARN_MEASURE_UNCALIBRATED,  RSZVL_WARNMSG_MEASURE_UNCALIBRATED},\
    {RSZVL_WARN_OVER_RANGE,            RSZVL_WARNMSG_OVER_RANGE},\
    {RSZVL_ERROR_MARKER_NOT_ENABLED,   RSZVL_ERRMSG_MARKER_NOT_ENABLED},\
    {RSZVL_ERROR_NOT_DELTA_MARKER,     RSZVL_ERRMSG_NOT_DELTA_MARKER},\
    {RSZVL_ERROR_MAX_TIME_EXCEEDED,    RSZVL_ERRMSG_MAX_TIME_EXCEEDED}


/****************************************************************************
 *---------------------------- End Include File ----------------------------*
 ****************************************************************************/

#if defined(__cplusplus) || defined(__cplusplus__)
}
#endif
#endif /* __RSZVL_HEADER */



